#include "flagbuilder.h"
#include <string.h>
#include <stdlib.h>

#define NUM_FLAGS 64

struct string
{
	size_t length;
	char   data[1]; /* one additional character for the terminating zero */
};
typedef struct _string string;

struct FlagBuilder
{
	struct string* flags[NUM_FLAGS];
	struct string* replacement;
};

static struct string* flagbld_strcpy(const char* str)
{
	const size_t         len = strlen(str);
	struct string* const cpy = (struct string*)malloc(sizeof(struct string) + sizeof(char)*(len + 1));

	if(NULL!=cpy)
	{
		cpy->length = len;
		memcpy(cpy->data, str, sizeof(char)*(len + 1));
	}

	return cpy;
}

struct FlagBuilder* FlagBuilder_Create(const char* const none_replacement)
{
	struct FlagBuilder* const builder = (struct FlagBuilder*)malloc(sizeof(struct FlagBuilder));

	if(NULL!=builder)
	{
		memset(builder, 0, sizeof(struct FlagBuilder));

		if(NULL==(builder->replacement=flagbld_strcpy(none_replacement?none_replacement:"")))
		{
			free(builder);
			return NULL;
		}
	}

	return builder;
}

void FlagBuilder_Release(struct FlagBuilder* const builder)
{
	if(NULL!=builder)
	{
		size_t i;

		if(builder->replacement)
		{
			free(builder->replacement);
		}

		for(i=0;i<NUM_FLAGS;++i)
		{
			if(builder->flags[i])
			{
				free(builder->flags[i]);
			}
		}

		free(builder);
	}
}

static size_t ctz64(wt_u64 x)
{
	size_t n = 1;

	if(0u==x) return 64;
	if(0u==(x & 0xFFFFFFFFul)) { n+=32; x>>=32;}
	if(0u==(x & 0x0000FFFFul)) { n+=16; x>>=16;}
	if(0u==(x & 0x000000FFul)) { n+= 8; x>>= 8;}
	if(0u==(x & 0x0000000Ful)) { n+= 4; x>>= 4;}
	if(0u==(x & 0x00000003ul)) { n+= 2; x>>= 2;}

	return n - (x & 1ul);
}

int FlagBuilder_Add(struct FlagBuilder* const builder, wt_u64 flag, const char* const str)
{
	int res = 0;

	if(0==(flag&(flag-1)))
	{
		struct string* const  cpy  = flagbld_strcpy(str);
		const size_t          idx  = ctz64(flag);
		struct string** const slot = builder->flags + idx;

		if(NULL!=cpy)
		{
			if(NULL!=*slot)
			{
				free(*slot);
			}

			*slot = cpy;
		}
		else
		{
			res = -2;
		}
	}
	else
	{
		res = -1;
	}

	return res;
}

char* FlagBuilder_Parse(struct FlagBuilder* const builder, wt_u64 flags, char* buf, size_t size)
{
	size_t          idx  = 0;
	wt_u64          flag = 1;
	struct string** slot = builder->flags;
	char*           out  = buf;

	for(idx=0;idx<NUM_FLAGS;flag<<=1 && *slot,++slot,++idx)
	{
		if(flags & flag)
		{
			if(buf<out)
			{
				if(size<2) { return NULL; }
				size -= 2;
				*out++ = ',';
				*out++ = ' ';
			}

			if(size<(*slot)->length) { return NULL; }

			memcpy(out, (*slot)->data, (*slot)->length);
			size -= (*slot)->length;
			out  += (*slot)->length;

			flags &= ~flag;
		}
	}

	if(flags)
	{
		static const char   others[]   = "others(0x";
		static const size_t others_len = sizeof(others) - sizeof(char);

		size_t n = NUM_FLAGS / 4;

		if(buf>out)
		{
			if(size<2) { return NULL; }
			size -= 2;
			*out++ = ',';
			*out++ = ' ';
		}

		if(size<others_len+1+n) { return NULL; }

		memcpy(out, others, others_len);
		size -= others_len;
		out  += others_len;

		for(idx=0;idx<n;flags<<=4,++idx)
		{
			wt_u64 nibble = flags >> (NUM_FLAGS - 4);

			if(nibble>=10)
			{
				*out++ = 'A' + (nibble - 10);
			}
			else
			{
				*out++ = '0' + nibble;
			}
		}

		*out++ = ')';
	}
	else
	if(out==buf)
	{
		if(size<builder->replacement->length) { return NULL; }

		memcpy(out, builder->replacement->data, builder->replacement->length);
		size -= builder->replacement->length;
		out  += builder->replacement->length;
	}

	*out = '\0';

	return buf;
}
