#include <stdio.h>
#include "arg.h"
#include "log.h"
#include "dep.h"
#include "tmpasm_scconfig.h"
#include "../wtruntime/scconfig_hooks.h"
#include "../readwtb/scconfig_hooks.h"
#include "../wtparser/scconfig_hooks.h"
#include "../wtc/scconfig_hooks.h"
#include "../wt/scconfig_hooks.h"
#include "../libwtref/scconfig_hooks.h"

/* defines a single dependency -- if Proj_A depends on Proj_B, then the */
/* dependency record is: { .dep_of = "Proj_A", .dep_is = "Proj_B" } */
struct dep
{
	const char* dep_of;
	const char* dep_is;
};

static const struct dep deps[] =
{
	{ "wtruntime", NULL },
	{ "libwtref", "wtruntime" },
	{ "readwtb", "wtruntime" },
	{ "readwtb", "libwtref" },
	{ "wtparser", "wtruntime" },
	{ "wt", "libwtref" },
	{ "wt", "wtruntime" },
	{ "wtc", "wtruntime" },
	{ "wtc", "libwtref" },
	{ "wtc", "wtparser" },
	{ 0 }
};

static void deps_with(const char* const proj)
{
	const int is_all = (0==strcmp("ALL", proj));
	const struct dep* dep = deps;
	int match = 0;
	char keybuf[128];

	for(;NULL!=dep->dep_of;++dep)
	{
		if(0==strcmp(dep->dep_of, proj))
		{
			if(NULL!=dep->dep_is)
			{
				deps_with(dep->dep_is);
			}

			++match;
		}
		else
		if(is_all)
		{
			deps_with(dep->dep_of);
			++match;
		}
	}

	if(0==match)
	{
		fprintf(stderr, "ERROR: Unknown project name in 'With' option: "
			"%s\n", proj);
		exit(2);
	}

	sprintf(keybuf, "/local/with/%s", proj);
	put(keybuf, strue);
}

static void deps_without(const char* const proj)
{
	const int is_all = (0==strcmp("ALL", proj));
	const struct dep* dep = deps;
	int match = 0;
	char keybuf[128];

	for(;NULL!=dep->dep_of;++dep)
	{
		if(is_all || (NULL!=dep->dep_is && 0==strcmp(dep->dep_is, proj)))
		{
			deps_without(dep->dep_of);
			++match;
		}
		else
		if(0==strcmp(dep->dep_of, proj))
		{
			++match;
		}
	}

	if(0==match)
	{
		fprintf(stderr, "ERROR: Unknown project name in 'Without' option: "
			"%s\n", proj);
		exit(2);
	}

	sprintf(keybuf, "/local/with/%s", proj);
	put(keybuf, sfalse);
}

/* Runs when a custom command line argument is found */
/* returns true if no further argument processing should be done */
int hook_custom_arg(const char* key, const char* value)
{
	if(0==strcmp(key, "With"))
	{
		deps_with(value);
		return 1;
	}

	if(0==strcmp(key, "Without"))
	{
		deps_without(value);
		return 1;
	}

	if(0==strcmp(key, "debug"))
	{
		wtr_set_debug(strue);
		wt_set_debug(strue);
		libwtref_set_debug(strue);
		return 1;
	}

	if(0==strcmp(key, "prefix"))
	{
		put("/local/wt_suite/prefix", value);
		wtr_set_prefix(value);
		libwtref_set_prefix(value);
		return 1;
	}

	if(0==strcmp(key, "vmmu-inc"))
	{
		put("/local/extern/vmmu/inc", value);
		return 1;
	}

	if(0==strcmp(key, "vmmu-lib"))
	{
		put("/local/extern/vmmu/lib", value);
		return 1;
	}

	if(0==strcmp(key, "buddy-inc"))
	{
		put("/local/extern/buddy/inc", value);
		return 1;
	}

	if(0==strcmp(key, "buddy-lib"))
	{
		put("/local/extern/buddy/lib", value);
		return 1;
	}

	if(wtr_custom_arg(key, value))
	{
		return 1;
	}

	return 0;
}


/* Runs before anything else */
int hook_preinit()
{
	return 0;
}

/* Runs after initialization */
int hook_postinit()
{
	db_mkdir("/local");
	put("/local/wt_suite/prefix", "/usr");
	put("/local/wt_suite/wt", "");
	put("/local/wt_suite/wt_dir", "");
	put("/local/wt_suite/wtc", "");
	put("/local/wt_suite/wtc_dir", "");
	put("/local/extern/vmmu/inc", "/usr/include");
	put("/local/extern/vmmu/lib", "-lvmmu");
	put("/local/extern/buddy/inc", "/usr/include");
	put("/local/extern/buddy/lib", "-lrlbuddy");

	/* by default, enable everything */
	deps_with("ALL");

	wtr_hook_postinit();
	readwtb_hook_postinit();
	wtparser_hook_postinit();
	wtc_hook_postinit();
	wt_hook_postinit();
	libwtref_hook_postinit();
	return 0;
}

/* Runs after all arguments are read and parsed */
int hook_postarg()
{
	return 0;
}

/* Runs when things should be detected for the host system */
int hook_detect_host()
{
	int err = 0;

	err |= require("sys/path_sep", 0, 0);

	err |= wtr_hook_detect_host();
	err |= readwtb_hook_detect_host();
	err |= wtparser_hook_detect_host();
	err |= wtc_hook_detect_host();
	err |= wt_hook_detect_host();
	err |= libwtref_hook_detect_host();

	return err;
}

/* Runs when things should be detected for the target system */
int hook_detect_target()
{
	int err = 0;

	require("cc/cc", 0, 0);
	require("cc/cflags", 0, 0);

	err |= wtr_hook_detect_target();
	err |= readwtb_hook_detect_target();
	err |= wtparser_hook_detect_target();
	err |= wtc_hook_detect_target();
	err |= wt_hook_detect_target();
	err |= libwtref_hook_detect_target();

	return err;
}

/* Runs after detection hooks, should generate the output (Makefiles, etc.) */
static int generate(const char* const root, const char* const outfn, const char* const infn)
{
	const int generr = tmpasm(root, infn, outfn);
	fprintf(stderr, "Generating %s/%s (%d)\n", root, outfn, generr);
	return generr;
}

static int wrap_generate(int (*hook_generate)(const char* root),
	const char* const root, const char* const proj)
{
	char keybuf[128];
	sprintf(keybuf, "/local/with/%s", proj);

	if(istrue(get(keybuf)))
	{
		return hook_generate(root);
	}
	else
	{
		return generate(root, "Makefile", "../scconfig/without-Makefile.in");
	}
}

int hook_generate()
{
	int generr = 0;

	generr |= wrap_generate(wtr_hook_generate, "../wtruntime", "wtruntime");
	generr |= wrap_generate(readwtb_hook_generate, "../readwtb", "readwtb");
	generr |= wrap_generate(wtparser_hook_generate, "../wtparser", "wtparser");
	generr |= wrap_generate(wtc_hook_generate, "../wtc", "wtc");
	generr |= wrap_generate(wt_hook_generate, "../wt", "wt");
	generr |= wrap_generate(libwtref_hook_generate, "../libwtref", "libwtref");

	generr |= generate("../doc", "Makefile", "Makefile.in");
	generr |= generate("../test-src", "Makefile", "Makefile.in");

	return generr;
}

/* Runs before everything is uninitialized */
void hook_preuninit()
{
}

/* Runs at the very end, when everything is already uninitialized */
void hook_postuninit()
{
}
