//================================================================================================
// $Id: impincwriter.cpp 6308 2023-01-16 20:16:41Z aron $
//
// Project:
//		Welltype Compiler
// Contents:
//		Import include writer
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "impincwriter.h"
#include "codegen.h"
#include "version.h"
#include "parserext.h"
#include <cctype>
#include <sstream>
#include "wtparser/ast/ast_program.h"
#include "wtparser/ast/ast_functionsignature.h"
#include "wtparser/ast/ast_functionargument.h"
#include "wtparser/ast/ast_type.h"
#include "wtparser/ast/ast_recordtype.h"
#include "wtparser/ast/ast_recordfield.h"
#include "wtparser/ast/ast_operatordeclaration.h"
#include "wtparser/ast/ast_datatype.h"
#include "wtparser/ast/ast_datatypecons.h"
#include "wtparser/ast/ast_functiontype.h"
#include "wtparser/ast/ast_enumerator.h"
#include "wtparser/ast/ast_enumtype.h"

namespace wtc
{
namespace codegen
{

ImportIncludeWriter::ImportIncludeWriter(wtc::ProblemHandler* const probhan,
	wtast::Program* const program, const std::string & sourcename,
	const std::string & filename)
	: mProblemHandler(probhan)
	, mProgram(program)
	, mSourceName(sourcename)
	, mFilename(filename)
	, mIsOpenFailed(false)
{
	// ok
}

ImportIncludeWriter::~ImportIncludeWriter()
{
	if(mOutFile.is_open())
	{
		mOutFile << "}" << std::endl;
		mOutFile << std::endl;

		mOutFile.close();
	}
}

bool ImportIncludeWriter::IsOpen()
{
	if(mOutFile.is_open())
	{
		return true;
	}

	if(mIsOpenFailed)
	{
		return false;
	}

	mOutFile.open(mFilename.c_str());

	if(!mOutFile.is_open())
	{
		mIsOpenFailed = true;
		mProblemHandler->AddError("?", 0, "Failed to create file \"" +
			mFilename + "\"");
		return false;
	}

	mOutFile << "// Generated import include for \"" << mSourceName << "\"." << std::endl;
	mOutFile << "// Welltype Compiler v" <<
		version::Major << '.' <<
		version::Minor << '.' <<
		version::Patch << std::endl;
	mOutFile << "// Date: " << mProgram->GetCompileDate() << ' ' <<
		mProgram->GetCompileTime() << std::endl;
	mOutFile << std::endl;
	Once(mOutFile);
	mOutFile << std::endl;
	mOutFile << "import" << std::endl;
	mOutFile << "{" << std::endl;

	return true;
}

void ImportIncludeWriter::Once(std::ostream & out)
{
	out << "#once _";

	for(std::string::const_iterator it=mSourceName.begin(),end=mSourceName.end();it!=end;++it)
	{
		char ch = *it;

		if(isalnum(ch))
		{
			out << ch;
		}
		else
		{
			out << '_';
		}
	}

	out << "_ImportInclude" << std::endl;
}

void ImportIncludeWriter::WriteFunctionSignature(std::ostream & os, wtast::FunctionSignature* sig)
{
	if(!sig->IsPure())
	{
		os << "impure ";
	}

	os << "function ";

	if(isalpha(sig->GetName()[0]) && !mOperators.count(sig->GetName()))
	{
		os << sig->GetName();
	}
	else
	{
		os << '\"' << sig->GetName() << '\"';
	}

	os << '(';

	for(std::size_t i=0,n=sig->GetArgc();i!=n;++i)
	{
		if(0<i)
		{
			os << ", ";
		}

		wtast::FunctionArgument* const arg = sig->GetArg(i);

		if(arg->GetType()->IsMutable())
		{
			os << "mutable ";
		}

		os << arg->GetType()->GetName();
		os << ' ';
		os << arg->GetName();
	}

	os << ')';

	for(std::size_t i=0,n=sig->GetRetc();i!=n;++i)
	{
		if(0<i)
		{
			os << ", ";
		}
		else
		{
			os << " : ";
		}

		os << sig->GetRet(i)->GetType(0)->GetName();
	}
}

void ImportIncludeWriter::WriteFunction(wtast::FunctionSignature* sig)
{
	if(!IsOpen()) { return; }

	std::ostringstream ss;
	WriteFunctionSignature(ss, sig);
	mOutFile << '\t' << ss.str() << ';' << std::endl;
}

void ImportIncludeWriter::WriteException(const std::string & excname)
{
	if(!IsOpen()) { return; }

	mOutFile << "\texception " << excname << ';' << std::endl;
}

void ImportIncludeWriter::WriteRecord(wtast::RecordType* rec)
{
	if(!IsOpen()) { return; }

	std::ostream & out = mOutFile;

	if(rec->IsLimited())
	{
		out << "\tlimited record " << rec->GetName() << ';' << std::endl;
	}
	else
	if(rec->IsPrivate())
	{
		out << "\tprivate record " << rec->GetName() << ';' << std::endl;
	}
	else
	{
		out << "\trecord " << rec->GetName() << std::endl;
		out << "\t{" << std::endl;
		for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
		{
			wtast::RecordField* field = rec->GetField(i);
			out << "\t\t" << field->GetType()->GetName() << ' ' <<
				field->GetName() << std::endl;
		}
		out << "\t}" << std::endl;
	}
}

void ImportIncludeWriter::WriteDataType(wtast::DataType* dt)
{
	if(!IsOpen()) { return; }

	std::ostream & out = mOutFile;

	out << "\tdatatype " << dt->GetName() << std::endl;
	out << "\t{" << std::endl;

	for(std::size_t i=0,n=dt->GetNumberOfConses();i!=n;++i)
	{
		wtast::DataTypeCons* cons = dt->GetCons(i);
		out << "\t\t" << cons->GetName();

		for(std::size_t i=0,n=cons->GetNumberOfTypes();i!=n;++i)
		{
			if(i>0) { out << ", "; } else { out << " "; }
			out << cons->GetType(i)->GetName();
		}

		out << ";" << std::endl;
	}

	out << "\t}" << std::endl;
}

void ImportIncludeWriter::WriteEnumType(wtast::EnumType* et)
{
	if(!IsOpen()) { return; }

	std::ostream & out = mOutFile;

	out << "\tenum " << et->GetName() << std::endl;
	out << "\t{" << std::endl;

	for(std::size_t i=0,n=et->GetNumberOfEnumerators();i!=n;++i)
	{
		wtast::Enumerator* enumerator = et->GetEnumerator(i);
		out << "\t\t" << enumerator->GetName();

		if(i<n)
		{
			out << ",";
		}

		out << std::endl;
	}

	out << "\t}" << std::endl;
}

void ImportIncludeWriter::WriteFunctionType(wtast::FunctionType* ft)
{
	if(!IsOpen()) { return; }

	std::ostringstream ss;
	WriteFunctionSignature(ss, ft->GetFunctionSignature());
	mOutFile << "\ttype " << ss.str() << ';' << std::endl;
}

void ImportIncludeWriter::WriteOperator(wtast::OperatorDeclaration* op)
{
	if(!IsOpen()) { return; }

	mOperators.insert(op->GetOpName());

	mOutFile << "\toperator " << op->GetOpName() << ' ' <<
		(op->IsLeftAssociative()?"left":"right") << ' ' << op->GetPrecedence()
		<< ';' << std::endl;
}

} // codegen
} // wtc
