//================================================================================================
// $Id: ParserHelper.cpp 6585 2023-04-07 13:40:43Z aron $
//
// Project:
//		Welltype Parser
// Contents:
//		Parser helper
// Authors:
//		Aron Barath, 2014-2023
//================================================================================================
// Copyright (c) Aron Barath
//================================================================================================

#include "ParserHelper.h"
#include "lex.yy.cc.h"
#include "parser.tab.hpp"
#include "builtins.h"
#include "checker.h"
#include "../wtruntime/types.h"
#include "welltype/stdlib.h"
#include "astnode.h"
#include "astvisitor.h"
#include "ast/ast_program.h"
#include "ast/ast_basictype.h"
#include "ast/ast_functioncall.h"
#include "ast/ast_functiondeclaration.h"
#include "ast/ast_simultaneoustype.h"
#include "ast/ast_simultaneousexpression.h"
#include "ast/ast_recorddeclaration.h"
#include "ast/ast_recordtype.h"
#include "ast/ast_recordfield.h"
#include "ast/ast_recordliteral.h"
#include "ast/ast_fieldassign.h"
#include <cstdlib>
#include <ctime>
#include <sstream>
#include <iomanip>

#ifdef _MSC_VER
#  pragma warning ( disable : 4996 )
#  pragma warning ( disable : 4355 ) // warning C4355: 'this' : used in base member initializer list
#endif // _MSC_VER

const std::string wtparser::ParserHelper::GlobalBlockName("global");
const std::string wtparser::ParserHelper::MainBlockName("main");
const std::string wtparser::ParserHelper::CtorBlockName("ctor");
const std::string wtparser::ParserHelper::DtorBlockName("dtor");
const std::string wtparser::ParserHelper::StrLeftAssoc("left");
const std::string wtparser::ParserHelper::StrRightAssoc("right");
const std::string wtparser::ParserHelper::OperatorChars("+-*/%^=!&|<>~");

wtparser::ParserHelper::ParserHelper(wtparser::ISourceManager* ism, wtparser::IProblemHandler* iph, const wtparser::IOptions* opt)
	: FirstUniqueId(wtparser::builtin::builtins.GetLastUniqueId()+1)
	, mLexerDebugger(opt->DebugLexer())
	, mRuleDebugger(opt->DebugRules())
	, mSymTab(this)
	, mSrcMan(this, ism)
	, mProbHan(this, iph)
	, mProgram(0)
	, mDeclEntity(0)
	, mNextUniqueId(FirstUniqueId)
	, mNextPublicTypeId(FirstPublicTypeId)
	, mNextPrivateTypeId(FirstPrivateTypeId)
	, mCurrentVariableClass(wtast::AstNode::Variable::LOCAL)
	, mCurrentDataType(0)
	, mCurrentEnumType(0)
{
	mCurrentFunctionStack.push_back(0);
	mEHandlerModeStack.push_back(false);
	mFinallyModeStack.push_back(false);
	mDistinguishedBlockModeStack.push_back(false);

	// date and time setup

	{
		struct tm *newtime;

#ifdef _MSC_VER
		__time64_t long_time;
		_time64( &long_time );
		newtime = _localtime64( &long_time );
#else // _MSC_VER
		time_t t;
		time(&t);
		newtime = localtime(&t);
#endif // _MSC_VER

		std::ostringstream ss_date;
		std::ostringstream ss_time;

		ss_date << (1900 + newtime->tm_year) << '.';
		ss_date << std::setw(2) << std::setfill('0') << (1 + newtime->tm_mon) << '.';
		ss_date << std::setw(2) << std::setfill('0') << (newtime->tm_mday)    << '.';

		ss_time << std::setw(2) << std::setfill('0') << newtime->tm_hour << ':';
		ss_time << std::setw(2) << std::setfill('0') << newtime->tm_min  << ':';
		ss_time << std::setw(2) << std::setfill('0') << newtime->tm_sec; 

		mCompileDate = ss_date.str();	
		mCompileTime = ss_time.str();
	}

	// adding built-in functions and operators

	_AddOperator("||", LeftAssoc,  5);
	_AddOperator("&&", LeftAssoc,  6);
	_AddOperator("|",  LeftAssoc,  7);
	_AddOperator("^",  LeftAssoc,  8);
	_AddOperator("&",  LeftAssoc,  9);
	_AddOperator("==", LeftAssoc, 10);
	_AddOperator("!=", LeftAssoc, 10);
	_AddOperator("<",  LeftAssoc, 11);
	_AddOperator("<=", LeftAssoc, 11);
	_AddOperator(">",  LeftAssoc, 11);
	_AddOperator(">=", LeftAssoc, 11);
	_AddOperator("<<", LeftAssoc, 12);
	_AddOperator(">>", LeftAssoc, 12);
	_AddOperator("+",  LeftAssoc, 13);
	_AddOperator("-",  LeftAssoc, 13);
	_AddOperator("*",  LeftAssoc, 14);
	_AddOperator("/",  LeftAssoc, 14);
	_AddOperator("%",  LeftAssoc, 14);
	_AddOperator("~",  LeftAssoc, 15);
	_AddOperator("!",  LeftAssoc, 15);

	for(builtin::Builtins::const_iterator it=builtin::builtins.begin(),end=builtin::builtins.end();it!=end;++it)
	{
		const builtin::Builtin & bf = *it;

		if(bf.IsGenerated())
		{
			_AddGenerated(bf.GetName(), RawTypes(bf.GetArgs()), RawTypes(bf.GetRets()), bf.IsPure(), bf.GetUniqueId());
		}
		else
		{
			_AddBuiltin(bf.GetName(), RawTypes(bf.GetArgs()), RawTypes(bf.GetRets()), bf.IsPure(), bf.GetUniqueId());
		}
	}
}

wtparser::ParserHelper::~ParserHelper()
{
	while(!mAllocatedAstNodes.empty())
	{
		delete mAllocatedAstNodes.back();
		mAllocatedAstNodes.pop_back();
	}
}

int yylex(wtparser::WTParser::semantic_type* value, wtparser::WTParser::location_type* location, wtparser::ParserHelper & helper)
{
	int          ret;
	WTFlexLexer* lexer;

	if(location)
	{
		location->begin.initialize();
		location->end.initialize();
	}

restart:

	lexer = helper.GetLexer();

	if(0==lexer)
	{
		value->str = new std::string("");
		ret = -1;
	}
	else
	{
		ret = lexer->yylex();

		if(location)
		{
			location->begin.line   = lexer->GetStartLine();
			location->begin.column = lexer->GetStartColumn();
			location->end.line     = lexer->GetEndLine();
			location->end.column   = lexer->GetEndColumn();
		}

		value->str = new std::string(lexer->YYText());
	}

	if(CTL_INCLUDE==wt_u32(ret))
	{
		if(lexer==0)
		{
			helper.PostError("Unexpected end-of-file.");
			return -1;
		}

		int tok = lexer->yylex();

		if(tok!=wtparser::WTParser::token::STRINGLIT)
		{
			helper.PostError("\"#include\" requires one string literal argument.");
		}
		else
		{
			std::string filename = helper.ParseString(lexer->YYText());

			switch(helper.GetSourceManager().OpenFile(filename))
			{
			case wtparser::SourceManager::OK:
				break;

			case wtparser::SourceManager::OUTOFMEMORY:
				helper.PostError("Internal error: out of memory.");
				break;

			case wtparser::SourceManager::FILEOPENERROR:
				helper.PostError(std::string("Cannot open include file: '") + filename + "': No such file or directory.");
				break;

			case wtparser::SourceManager::DEPTHLIMITEXCEEDED:
				helper.PostError("Include depth limit exceeded.");
				break;
			}
		}

		goto restart;
	}
	else
	if(CTL_ONCE==wt_u32(ret))
	{
		if(lexer==0)
		{
			helper.PostError("Unexpected end-of-file.");
			return -1;
		}

		if(wtparser::WTParser::token::IDENTIFIER==lexer->yylex())
		{
			std::string ident = lexer->YYText();

			if(helper.IsLabelDefined(ident))
			{
				helper.GetSourceManager().CloseTop();
			}

			goto restart;
		}
		else
		{
			helper.PostError("\"#once\" requires one identifier argument.");
			helper.GetSourceManager().CloseTop();
			goto restart;
		}
	}
	else
	if(CTL_EOF==wt_u32(ret) && 0!=lexer)
	{
		helper.GetSourceManager().CloseTop();

		goto restart;
	}
	else
	if(CTL_UTF8_BOM==wt_u32(ret))
	{
		goto restart;
	}

	helper.mLexerDebugger.PrintLex(ret, lexer?lexer->YYText():"");

    return ret;
}

wtast::Program* wtparser::ParserHelper::GetProgram()
{
	if(mProgram)
	{
		wtast::Program* prog = mProgram;
		mProgram = 0;

		wtast::DeclareBlock* std_exc = new wtast::DeclareBlock(true, false);
#define ADD_EXCEPTION(exc_name) std_exc->AddDeclarationBack(new wtast::ExceptionDeclaration(#exc_name, CreateUniqueId()));
		WTR_STDEXCEPTIONS(ADD_EXCEPTION);
#undef ADD_EXCEPTION
		prog->mEntities.push_front(std_exc);

		prog->mAllocatedAstNodes.swap(mAllocatedAstNodes);
		prog->mLastUniqueId = mNextUniqueId;
		prog->mCompileDate  = mCompileDate;
		prog->mCompileTime  = mCompileTime;

		mNametypeFuncsig.clear();
		mNameFuncsig.clear();
		mTypenameToType.clear();
		mAliasToId.clear();
		mAutoAliases.clear();
		mOpnameToOp.clear();
		mExceptionToId.clear();

		mNextUniqueId         = FirstUniqueId;
		mNextPublicTypeId     = FirstPublicTypeId;
		mNextPrivateTypeId    = FirstPrivateTypeId;
		mCurrentVariableClass = wtast::AstNode::Variable::LOCAL;

		return prog;
	}

	return 0;
}

int wtparser::ParserHelper::GetOperatorToken(bool left_assoc, wt_i32 prec)
{
	if(MinPrecedence>prec || MaxPrecedence<prec)
	{
		return 0;
	}

	switch(prec)
	{
	case  3: return left_assoc ? WTParser::token::OP_LEFT03 : WTParser::token::OP_RIGHT03;
	case  4: return left_assoc ? WTParser::token::OP_LEFT04 : WTParser::token::OP_RIGHT04;
	case  5: return left_assoc ? WTParser::token::OP_LEFT05 : WTParser::token::OP_RIGHT05;
	case  6: return left_assoc ? WTParser::token::OP_LEFT06 : WTParser::token::OP_RIGHT06;
	case  7: return left_assoc ? WTParser::token::OP_LEFT07 : WTParser::token::OP_RIGHT07;
	case  8: return left_assoc ? WTParser::token::OP_LEFT08 : WTParser::token::OP_RIGHT08;
	case  9: return left_assoc ? WTParser::token::OP_LEFT09 : WTParser::token::OP_RIGHT09;
	case 10: return left_assoc ? WTParser::token::OP_LEFT10 : WTParser::token::OP_RIGHT10;
	case 11: return left_assoc ? WTParser::token::OP_LEFT11 : WTParser::token::OP_RIGHT11;
	case 12: return left_assoc ? WTParser::token::OP_LEFT12 : WTParser::token::OP_RIGHT12;
	case 13: return left_assoc ? WTParser::token::OP_LEFT13 : WTParser::token::OP_RIGHT13;
	case 14: return left_assoc ? WTParser::token::OP_LEFT14 : WTParser::token::OP_RIGHT14;
	case 15: return left_assoc ? WTParser::token::OP_LEFT15 : WTParser::token::OP_RIGHT15;
	}

	return 0;
}

bool wtparser::ParserHelper::_AddOperator(const std::string & opname, bool left_assoc, wt_i32 prec)
{
	wtast::OperatorDeclaration* opdecl = new wtast::OperatorDeclaration(opname, left_assoc, prec);

	if(!opdecl) { return false; }

	opdecl->SetBuiltin();
	mOpnameToOp[opname] = OperatorTokenPair(opdecl, GetOperatorToken(left_assoc, prec));
	return true;
}

bool wtparser::ParserHelper::_AddBuiltin(const std::string & name, const RawTypes & args, const RawTypes & rets, bool pure, UniqueId unique_id)
{
	return _AddFunction(name, args, rets, pure, true, false, unique_id);
}

bool wtparser::ParserHelper::_AddGenerated(const std::string & name, const RawTypes & args, const RawTypes & rets, bool pure, UniqueId unique_id)
{
	return _AddFunction(name, args, rets, pure, false, true, unique_id);
}

bool wtparser::ParserHelper::_AddFunction(const std::string & name, const RawTypes & args, const RawTypes & rets, bool pure, bool builtin, bool generated, UniqueId unique_id)
{
	wtast::FunctionSignature* sig = new wtast::FunctionSignature(name, pure);
	unsigned long counter = 1;

	if(!sig) { return false; }

	if(builtin)
	{
		sig->SetBuiltin();
	}

	if(generated)
	{
		sig->SetGenerated();
	}

	for(RawTypes::const_iterator it=args.begin(),end=args.end();it!=end;++it)
	{
		bool is_mutable = !!(WTTF_MUTABLE & (*it));
		wtast::Type* type = GetResolvedType((*it) & ~WTTF_MUTABLE);

		if(!type)
		{
			delete sig;
			return false;
		}

		if(is_mutable)
		{
			type = type->Copy();
			type->SetMutable();

			mTypeidToType[type->GetTypeId()] = type;
		}

		std::ostringstream ss;
		ss << "a" << (counter++);
		wtast::FunctionArgument* fa = new wtast::FunctionArgument(type, ss.str());

		if(!fa)
		{
			delete sig;
			return false;
		}

		sig->AddArgBack(fa);
	}

	for(RawTypes::const_iterator it=rets.begin(),end=rets.end();it!=end;++it)
	{
		wtast::Type* type = GetResolvedType(*it);

		if(!type)
		{
			delete sig;
			return false;
		}

		sig->AddRetBack(type);
	}

	wtast::TypeInfo* typeinfo = 0;

	if(1==args.size())
	{
		typeinfo = GetResolvedType(*args.begin());
	}
	else
	if(1<args.size())
	{
		wtast::SimultaneousType* st = new wtast::SimultaneousType();
		typeinfo = st;

		for(RawTypes::const_iterator it=args.begin(),end=args.end();it!=end;++it)
		{
			wtast::Type* type = GetResolvedType(*it);

			if(!type)
			{
				delete sig;
				delete typeinfo;
				return false;
			}

			st->AddType(type);
		}
	}

	sig->SetUniqueId(unique_id);
	mAllocatedAstNodes.push_back(sig);
	if(typeinfo && typeinfo->IsSimultaneous()) { mAllocatedAstNodes.push_back(typeinfo); }
	mNameFuncsig[name].push_back(sig);
	mNametypeFuncsig.insert(std::make_pair(NameTypePair(name, typeinfo), sig));

	return true;
}

bool wtparser::ParserHelper::_AddSequenceOperations(wtast::Type* base_type, wt_u32 seq_typeid)
{
	if(!base_type->IsCopyable())
	{
		PostError("Uncopyable type cannot form a sequence.");
		// we don't return now, because the registred error message is enough
	}

	wt_u32 elem_typeid = seq_typeid & ~WTTF_SEQ;

	if(!_AddGenerated("empty", RawTypes(seq_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("seqlen", RawTypes(seq_typeid), RawTypes(WTT_U32), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("seqfill", RawTypes(elem_typeid).Add(WTT_U32), RawTypes(seq_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("subseq", RawTypes(seq_typeid).Add(WTT_U32).Add(WTT_U32), RawTypes(seq_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("lov", RawTypes(seq_typeid), RawTypes(elem_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("hiv", RawTypes(seq_typeid), RawTypes(elem_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("lorem", RawTypes(seq_typeid), RawTypes(seq_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("hirem", RawTypes(seq_typeid), RawTypes(seq_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("loext", RawTypes(seq_typeid).Add(elem_typeid), RawTypes(seq_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("hiext", RawTypes(seq_typeid).Add(elem_typeid), RawTypes(seq_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("lopop", RawTypes(seq_typeid), RawTypes(seq_typeid).Add(elem_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("hipop", RawTypes(seq_typeid), RawTypes(seq_typeid).Add(elem_typeid), true, CreateUniqueId())) { return false; }

	return true;
}

bool wtparser::ParserHelper::_AddEnumOperations(wtast::EnumType* enum_type, wt_u32 enum_typeid)
{
	std::string to_enum = std::string("to") + enum_type->GetName();

	if(!_AddGenerated("succ", RawTypes(enum_typeid), RawTypes(enum_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("pred", RawTypes(enum_typeid), RawTypes(enum_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("ordinal", RawTypes(enum_typeid), RawTypes(WTT_U32), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("tos", RawTypes(enum_typeid), RawTypes(WTT_STRING), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated(to_enum, RawTypes(WTT_STRING), RawTypes(enum_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated(to_enum, RawTypes(WTT_U32), RawTypes(enum_typeid), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("==", RawTypes(enum_typeid).Add(enum_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("!=", RawTypes(enum_typeid).Add(enum_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("<", RawTypes(enum_typeid).Add(enum_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated("<=", RawTypes(enum_typeid).Add(enum_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated(">", RawTypes(enum_typeid).Add(enum_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }
	if(!_AddGenerated(">=", RawTypes(enum_typeid).Add(enum_typeid), RawTypes(WTT_BOOL), true, CreateUniqueId())) { return false; }

	return true;
}

bool wtparser::ParserHelper::NameTypePair::less(const wtparser::ParserHelper::NameTypePair & lhs, const wtparser::ParserHelper::NameTypePair & rhs)
{
	if(lhs.name < rhs.name)
	{
		return true;
	}

	if(lhs.name == rhs.name)
	{
		if(!lhs.type && rhs.type) { return true; }
		if(!rhs.type) { return false; }

		bool lsim = lhs.type->IsSimultaneous();
		bool rsim = rhs.type->IsSimultaneous();

		if(!lsim && rsim) { return true; }
		if(lsim && !rsim) { return false; }

		if(lsim)
		{
			wtast::SimultaneousType* ls = static_cast<wtast::SimultaneousType*>(lhs.type);
			wtast::SimultaneousType* rs = static_cast<wtast::SimultaneousType*>(rhs.type);

			if(ls->GetNumberOfTypes() < rs->GetNumberOfTypes()) { return true; }
			if(ls->GetNumberOfTypes() > rs->GetNumberOfTypes()) { return false; }

			for(wt_u32 i=0,num=ls->GetNumberOfTypes();i<num;++i)
			{
				wt_u32 lid = ls->GetType(i)->GetTypeId() & ~WTTF_MUTABLE;
				wt_u32 rid = rs->GetType(i)->GetTypeId() & ~WTTF_MUTABLE;

				if(lid < rid) { return true; }
				if(lid > rid) { return false; }
			}
		}
		else
		{
			return
				(static_cast<wtast::Type*>(lhs.type)->GetTypeId() & ~WTTF_MUTABLE)
				<
				(static_cast<wtast::Type*>(rhs.type)->GetTypeId() & ~WTTF_MUTABLE);
		}
	}

	return false;
}

bool wtparser::ParserHelper::IsLabelDefined(const std::string & once_label)
{
	if(mOnceLabels.find(once_label)==mOnceLabels.end())
	{
		mOnceLabels.insert(once_label);
		return false;
	}

	return true;
}

bool wtparser::ParserHelper::HaveLocalType(const std::string & str)
{
	return mTypenameToType.find(str) != mTypenameToType.end();
}

bool wtparser::ParserHelper::IsSpecialIdent(const std::string & str)
{
	{
		map_name_datatypeconses_t::iterator it = mNameToDataTypeConses.find(str);

		if(mNameToDataTypeConses.end()!=it && !it->second.empty())
		{
			return true;
		}
	}

	{
		map_name_enumerators_t::iterator it = mNameToEnumerators.find(str);

		if(mNameToEnumerators.end()!=it && !it->second.empty())
		{
			return true;
		}
	}

	return false;
}

int wtparser::ParserHelper::OperatorLookup(const std::string & str)
{
	map_name_optok_t::iterator it = mOpnameToOp.find(str);

	if(mOpnameToOp.end()!=it)
	{
		return it->second.tok;
	}

	return wtparser::WTParser::token::UNBOUNDOPERATOR;
}

int wtparser::ParserHelper::TransformIdent(const std::string & str)
{
	if(IsSpecialIdent(str))
	{
		return wtparser::WTParser::token::SPECIALIDENT;
	}

	if(HaveLocalType(str))
	{
		return wtparser::WTParser::token::TYPENAME;
	}

	int op = OperatorLookup(str);
	if(wtparser::WTParser::token::UNBOUNDOPERATOR!=op)
	{
		return op;
	}

	return wtparser::WTParser::token::IDENTIFIER;
}

unsigned int wtparser::ParserHelper::GetLine()
{
	if(GetLexer())
	{
		return GetLexer()->GetStartLine();
	}

	return 0;
}

void wtparser::WTParser::error(const wtparser::location &, const std::string & msg)
{
	helper.GetProblemHandler().AddError(helper.GetFilename(), helper.GetLine(), msg);
}

void wtparser::ParserHelper::PostError(const std::string & msg, unsigned int line)
{
	mProbHan.AddError(GetFilename(), line, msg);
}

void wtparser::ParserHelper::PostWarning(const std::string & msg, unsigned int line)
{
	mProbHan.AddWarning(GetFilename(), line, msg);
}

void wtparser::ParserHelper::PostNote(const std::string & msg, unsigned int line)
{
	mProbHan.AddNote(GetFilename(), line, msg);
}

int wtparser::ParserHelper::Lexer_Operator(void* lexer)
{
	WTFlexLexer* flex = reinterpret_cast<WTFlexLexer*>(lexer);
	wtparser::ParserHelper* helper = flex->GetHelper();
	return helper->OperatorLookup(flex->YYText());
}

bool wtparser::ParserHelper::Lexer_HaveLocalType(void* lexer)
{
	WTFlexLexer* flex = reinterpret_cast<WTFlexLexer*>(lexer);
	wtparser::ParserHelper* helper = flex->GetHelper();
	return helper->HaveLocalType(flex->YYText());
}

int wtparser::ParserHelper::Lexer_TransformIdent(void* lexer)
{
	WTFlexLexer* flex = reinterpret_cast<WTFlexLexer*>(lexer);
	wtparser::ParserHelper* helper = flex->GetHelper();
	return helper->TransformIdent(flex->YYText());
}

std::string wtparser::ParserHelper::GetNumberCore(const std::string & str)
{
	std::size_t offset = 0;
	std::size_t count  = str.length();

	if(count==0)
	{
		return "";
	}

	if(str.length()>=2 && str[0]=='0' && IsAnyOf(str[1], "xXbBoO"))
	{
		offset += 2;
		count  -= 2;
	}

	if(IsAnyOf(str[offset+count-1], "bB"))
	{
		--count;
	}

	if(IsAnyOf(str[offset+count-1], "hH"))
	{
		--count;
	}

	if(IsAnyOf(str[offset+count-1], "lL"))
	{
		--count;
	}

	if(IsAnyOf(str[offset+count-1], "iI"))
	{
		--count;
	}

	if(IsAnyOf(str[offset+count-1], "uU"))
	{
		--count;
	}

	return str.substr(offset, count);
}

wt_u32 wtparser::ParserHelper::BinToI32(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto32(s.c_str(), 2);
}

wt_u32 wtparser::ParserHelper::OctToI32(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto32(s.c_str(), 8);
}

wt_u32 wtparser::ParserHelper::DecToI32(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto32(s.c_str(), 10);
}

wt_u32 wtparser::ParserHelper::HexToI32(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto32(s.c_str(), 16);
}

template<typename IntegerType>
const char* string_to_integer(const char* str, wt_u32 base, IntegerType & value)
{
	value = 0;

	while(*str)
	{
		if('_'==*str)
		{
			++str;
		}
		else
		{
			const IntegerType value_before = value;

			value *= base;

			if(value_before!=(value/base))
			{
				return "Number literal overflow.";
			}

			value += wtparser::ParserHelper::HexValue(*str++);

			if(value_before>value)
			{
				return "Number literal overflow.";
			}
		}
	}

	return 0;
}

wt_u8 wtparser::ParserHelper::strto8(const char* str, wt_u32 base)
{
	wt_u8 value = 0;

	const char* const msg = string_to_integer(str, base, value);

	if(msg)
	{
		PostError(msg);
		return 0;
	}

	return value;
}

wt_u16 wtparser::ParserHelper::strto16(const char* str, wt_u32 base)
{
	wt_u16 value = 0;

	const char* const msg = string_to_integer(str, base, value);

	if(msg)
	{
		PostError(msg);
		return 0;
	}

	return value;
}

wt_u32 wtparser::ParserHelper::strto32(const char* str, wt_u32 base)
{
	wt_u32 value = 0;

	const char* const msg = string_to_integer(str, base, value);

	if(msg)
	{
		PostError(msg);
		return 0;
	}

	return value;
}

wt_u64 wtparser::ParserHelper::strto64(const char* str, wt_u32 base)
{
	wt_u64 value = 0;

	const char* const msg = string_to_integer(str, base, value);

	if(msg)
	{
		PostError(msg);
		return 0;
	}

	return value;
}

wt_u64 wtparser::ParserHelper::BinToI64(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto64(s.c_str(), 2);
}

wt_u64 wtparser::ParserHelper::OctToI64(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto64(s.c_str(), 8);
}

wt_u64 wtparser::ParserHelper::DecToI64(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto64(s.c_str(), 10);
}

wt_u64 wtparser::ParserHelper::HexToI64(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto64(s.c_str(), 16);
}

wt_f32 wtparser::ParserHelper::ParseF32(const std::string & str)
{
	return (wt_f32)atof(str.c_str());
}

wt_f64 wtparser::ParserHelper::ParseF64(const std::string & str)
{
	return atof(str.c_str());
}

wt_u8 wtparser::ParserHelper::BinToI8(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto8(s.c_str(), 2);
}

wt_u8 wtparser::ParserHelper::OctToI8(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto8(s.c_str(), 8);
}

wt_u8 wtparser::ParserHelper::DecToI8(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto8(s.c_str(), 10);
}

wt_u8 wtparser::ParserHelper::HexToI8(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto8(s.c_str(), 16);
}

wt_u16 wtparser::ParserHelper::BinToI16(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto16(s.c_str(), 2);
}

wt_u16 wtparser::ParserHelper::OctToI16(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto16(s.c_str(), 8);
}

wt_u16 wtparser::ParserHelper::DecToI16(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto16(s.c_str(), 10);
}

wt_u16 wtparser::ParserHelper::HexToI16(const std::string & str)
{
	std::string s = GetNumberCore(str);
	return strto16(s.c_str(), 16);
}

wt_u32 wtparser::ParserHelper::HexValue(char ch)
{
	if('0'<=ch && ch<='9')
	{
		return ch - '0';
	}
	else
	if('a'<=ch && ch<='f')
	{
		return 10 + (ch - 'a');
	}
	else
	if('A'<=ch && ch<='F')
	{
		return 10 + (ch - 'A');
	}

	return 0;
}

wt_u32 wtparser::ParserHelper::ParseChar(const char* & ptr)
{
	if(ptr[0]=='\\')
	{
		char ch = ptr[1];
		ptr += 2;

		switch(ch)
		{
		case '\\' : return '\\';
		case 'b'  : return '\b';
		case 'v'  : return '\v';
		case 'n'  : return '\n';
		case 'r'  : return '\r';
		case 't'  : return '\t';
		case '0'  : return '\0';
		case '\'' : return '\'';
		case '\"' : return '\"';
		case 'x'  :
			{
				if(!IsHexDigit(ptr[0]) || !IsHexDigit(ptr[1]))
				{
					ptr += 2;
					return 0xFFFFFFFFul;
				}

				wt_u32 ret = HexValue(ptr[0])*16 + HexValue(ptr[1]);
				ptr += 2;

				return ret;
			}
			break;

		default: return 0xFFFFFFFFul;
		}
	}
	else
	{
		return *ptr++;
	}
}

char wtparser::ParserHelper::ParseChar(const std::string & str)
{
	const char* s = &str[1];

	wt_u32 cu = ParseChar(s);

	if(cu==0xFFFFFFFFul)
	{
		PostWarning(std::string("Unrecognised character sequence: ") + std::string(&str[1],s-&str[1]));
		cu = 0;
	}

	return char(cu);
}

std::string wtparser::ParserHelper::ParseString(const std::string & str)
{
	if(str.empty() || str.length()<2)
	{
		return "";
	}

	std::ostringstream ss;
	const char*        it  = &str[1];
	const char*        end = &str[0] + str.length() - 1;

	while(it<end)
	{
		const char* p0 = it;
		wt_u32      cu = ParseChar(it);

		if(cu==0 || cu==0xFFFFFFFFul)
		{
			PostWarning(std::string("Unrecognised character sequence: ") + std::string(p0, it-p0));
			cu = 0;
		}

		ss << ((char)cu);
	}

	return ss.str();
}

wtast::Type* wtparser::ParserHelper::GetStockType(wt_u32 basictype_id)
{
	map_u32_type_t::iterator it = mTypeidToType.find(basictype_id);

	if(mTypeidToType.end()!=it)
	{
		return it->second;
	}

	const char* type_name = 0;
	wt_u32      load_size = 0;

	switch(basictype_id)
	{
	case WTT_CHAR:    type_name = "char";    load_size = sizeof(wt_char);   break;
	case WTT_BOOL:    type_name = "bool";    load_size = sizeof(wt_bool);   break;
	case WTT_HANDLE:  type_name = "handle";  load_size = sizeof(wt_handle); break;
	case WTT_STRING:  type_name = "string";  load_size = sizeof(wt_vmptr);  break; // object
	case WTT_FILE:    type_name = "file";    load_size = sizeof(wt_vmptr);  break; // object
	case WTT_DIR:     type_name = "dir";     load_size = sizeof(wt_vmptr);  break; // object
	case WTT_I8:      type_name = "i8";      load_size = sizeof(wt_i8);     break;
	case WTT_U8:      type_name = "u8";      load_size = sizeof(wt_u8);     break;
	case WTT_I16:     type_name = "i16";     load_size = sizeof(wt_i16);    break;
	case WTT_U16:     type_name = "u16";     load_size = sizeof(wt_u16);    break;
	case WTT_I32:     type_name = "i32";     load_size = sizeof(wt_i32);    break;
	case WTT_U32:     type_name = "u32";     load_size = sizeof(wt_u32);    break;
	case WTT_I64:     type_name = "i64";     load_size = sizeof(wt_i64);    break;
	case WTT_U64:     type_name = "u64";     load_size = sizeof(wt_u64);    break;
	case WTT_F32:     type_name = "f32";     load_size = sizeof(wt_f32);    break;
	case WTT_F64:     type_name = "f64";     load_size = sizeof(wt_f64);    break;
	default: return 0;
	}

	if(!type_name)
	{
		return 0;
	}

	wtast::BasicType* basic_type = new wtast::BasicType(type_name, basictype_id, load_size);
	mTypeidToType[basictype_id] = basic_type;
	mAllocatedAstNodes.push_back(basic_type);
	return basic_type;
}

wtast::Type* wtparser::ParserHelper::GetResolvedType(wt_u32 type_id)
{
	type_id = ResolveTypeIdAliasing(type_id);

	map_u32_type_t::iterator it = mTypeidToType.find(type_id);

	if(mTypeidToType.end()!=it)
	{
		return it->second;
	}

	if(WTTF_SEQ & type_id)
	{
		wtast::SeqType* seq_type = new wtast::SeqType(GetResolvedType(type_id & ~WTTF_SEQ), type_id);
		mTypeidToType[type_id] = seq_type;
		mAllocatedAstNodes.push_back(seq_type);
		return seq_type;
	}

	return GetStockType(type_id);
}

wtast::SeqType* wtparser::ParserHelper::GetSeqType(wtast::Type* type)
{
	wt_u32 type_id = GetSeqTypeId(type);
	return static_cast<wtast::SeqType*>(GetResolvedType(type_id));
}

wt_u32 wtparser::ParserHelper::ResolveTypeIdAliasing(wt_u32 type_id)
{
	for(;;)
	{
		map_u32_u32_t::iterator it = mAliasToId.find(type_id);

		if(mAliasToId.end()==it)
		{
			return type_id;
		}

		type_id = it->second;
	}

	return WTT_VOID;
}

wtast::Type* wtparser::ParserHelper::ResolveType(const std::string & type_name)
{
	map_name_type_t::iterator it = mTypenameToType.find(type_name);

	if(mTypenameToType.end()!=it)
	{
		return it->second;
	}

	std::ostringstream ss;
	ss << "Typename '" << type_name << "' can not be resolved.";
	PostError(ss.str());

	return 0;
}

wtast::FunctionSignature* wtparser::ParserHelper::ResolveFunction(const std::string & funcname, wtast::TypeInfo* typeinfo)
{
	if("seqfill"==funcname && typeinfo && 2==typeinfo->GetNumberOfTypes())
	{
		wt_u32 elem_typeid = typeinfo->GetType(0)->GetTypeId();
		_AddGenerated("seqfill", RawTypes(elem_typeid).Add(WTT_U32), RawTypes(elem_typeid|WTTF_SEQ), true, CreateUniqueId());
	}

	nametype_funcsig_t::iterator it = mNametypeFuncsig.find(NameTypePair(funcname, typeinfo));

	if(it==mNametypeFuncsig.end())
	{
		name_funcsig_t::iterator overloads_it = mNameFuncsig.find(funcname);

		if(overloads_it!=mNameFuncsig.end())
		{
			PostError("Function argument mismatch.");
			PostNote("Possible overloads:");

			for(std::size_t i=0,n=overloads_it->second.size();i<n;++i)
			{
				wtast::FunctionSignature* sig = overloads_it->second[i];

				PostNote(std::string("   ") + sig->GetCanonicalName(true, true));
			}
		}
		else
		{
			std::ostringstream ss;

			ss << "'" << funcname << "' is not a name of a function.";

			PostError(ss.str());
		}

		return 0;
	}

	return it->second;
}

wtast::Expression* wtparser::ParserHelper::ResolveFunctionCall(wtast::Expression* funcexpr, wtast::Expression* argexpr)
{
	if(!funcexpr)
	{
		PostError("Failed to resolve function call.");
		return 0;
	}

	wtast::FunctionSignature* funcsig = 0;
	wtast::TypeInfo* typeinfo = 0;
	wtast::FunctionType* functype = 0;

	if(argexpr)
	{
		typeinfo = argexpr->GetType();
	}

	if(wtast::AstNode::UNRESOLVEDFUNCTIONNAME==funcexpr->GetKind() || wtast::AstNode::STRINGLITERAL==funcexpr->GetKind())
	{
		std::string funcname;

		if(wtast::AstNode::UNRESOLVEDFUNCTIONNAME==funcexpr->GetKind())
		{
			funcname = static_cast<wtast::UnresolvedFunctionName*>(funcexpr)->GetName();
		}
		else
		{
			funcname = static_cast<wtast::StringLiteral*>(funcexpr)->GetLiteral();
			funcname = ValidateFunctionName(funcname, false);

			if(funcname.empty())
			{
				return 0;
			}
		}

		funcsig = ResolveFunction(funcname, typeinfo);
		delete funcexpr;
		funcexpr = 0;
	}
	else
	{
		if(!IsFunctionType(funcexpr->GetType()))
		{
			PostError("Type of the expression is not a function type.");
			return 0;
		}

		functype = static_cast<wtast::FunctionType*>(funcexpr->GetType());
		funcsig = functype->GetFunctionSignature();
	}

	if(!funcsig)
	{
		return 0;
	}

	if(!ExpressionCanBeArgument(funcsig, argexpr))
	{
		return 0;
	}

	if(functype)
	{
		return MakeFunctionPointerCall(functype, funcexpr, argexpr);
	}

	return new wtast::FunctionCall(funcsig, argexpr);
}

wtast::UnresolvedFunctionName* wtparser::ParserHelper::ResolveGlobalFunctionName(const std::string & funcname)
{
	if(mNameFuncsig.end()!=mNameFuncsig.find(funcname))
	{
		return new wtast::UnresolvedFunctionName(funcname);
	}

	std::ostringstream ss;
	ss << "\'" << funcname << "' is not a name of a function.";
	PostError(ss.str());

	return 0;
}

wtparser::UniqueId wtparser::ParserHelper::GetFuncUniqueId(wtast::FunctionSignature* sig)
{
	std::string name = sig->GetName();
	wtast::TypeInfo* typeinfo = sig->GetArgType();
	nametype_funcsig_t::iterator it = mNametypeFuncsig.find(NameTypePair(name, typeinfo));

	if(it==mNametypeFuncsig.end())
	{
		// deal with 'typeinfo'
		if(typeinfo)
		{
			mAllocatedAstNodes.push_back(typeinfo);
		}

		// generate a new UniqueId
		sig->SetUniqueId(CreateUniqueId());
		mNameFuncsig[name].push_back(sig);
		mNametypeFuncsig[NameTypePair(name, typeinfo)] = sig;
		return sig->GetUniqueId();
	}
	else
	{
		// deal with 'typeinfo'
		delete typeinfo;

		// return with existent UniqueId
		return it->second->GetUniqueId();
	}
}

wtparser::UniqueId wtparser::ParserHelper::CreateUniqueId()
{
	return mNextUniqueId++;
}

bool wtparser::ParserHelper::ExpressionHasNoType(wtast::Expression* expr)
{
	if(!expr)
	{
		return false;
	}

	wtast::TypeInfo* type = expr->GetType();

	if(!type)
	{
		return true;
	}

	if(type->IsSimultaneous())
	{
		wtast::SimultaneousType* st = static_cast<wtast::SimultaneousType*>(type);

		return 0==st->GetNumberOfTypes();
	}

	return false;
}

wt_u32 wtparser::ParserHelper::GetTypeId(const std::string & name)
{
	// check whether the type name exists

	map_name_type_t::const_iterator it = mTypenameToType.find(name);

	if(it!=mTypenameToType.end())
	{
		return it->second->GetTypeId();
	}

	std::ostringstream ss;
	ss << "Unknown typename '" << name << "'.";
	PostError(ss.str());

	return WTT_VOID;
}

wt_u32 wtparser::ParserHelper::GetSeqTypeId(wtast::Type* type)
{
	wt_u32 type_id = type->GetTypeId();

	if(WTTF_SEQ & type_id)
	{
		// create a new automatic type alias if needed

		// TODO, nincs AliasDeclaration
		return WTT_VOID;
	}

	type_id = WTTF_SEQ | type_id;

	if(!_AddSequenceOperations(type, type_id))
	{
		return WTT_VOID;
	}

	return type_id;
}

std::string wtparser::ParserHelper::FileAsString()
{
	return mSrcMan.GetFilename(false);
}

std::string wtparser::ParserHelper::LineAsString()
{
	std::ostringstream ss;
	ss << GetLexer()->GetStartLine();
	return ss.str();
}

std::string wtparser::ParserHelper::DateAsString()
{
	return mCompileDate;
}

std::string wtparser::ParserHelper::TimeAsString()
{
	return mCompileTime;
}

bool wtparser::ParserHelper::IsBoolType(wtast::TypeInfo* typeinfo)
{
	return wtast::AstNode::BASICTYPE==typeinfo->GetKind()
		&& WTT_BOOL==static_cast<wtast::BasicType*>(typeinfo)->GetTypeId();
}

bool wtparser::ParserHelper::IsFunctionType(wtast::TypeInfo* typeinfo)
{
	return typeinfo && (!typeinfo->IsSimultaneous()) &&
		typeinfo->GetType(0)->IsFunctionType();
}

bool wtparser::ParserHelper::IsSameType(wtast::Type* lhs, wtast::Type* rhs, Usage::Kind /*usage*/)
{
	return lhs->GetTypeId()==(rhs->GetTypeId() & ~WTTF_MUTABLE);
}

void wtparser::ParserHelper::EnterCurrentFunction(wtast::FunctionSignature* funcsig)
{
	mCurrentFunctionStack.push_back(funcsig);
}

void wtparser::ParserHelper::LeaveCurrentFunction()
{
	mCurrentFunctionStack.pop_back();
}

wtast::FunctionSignature* wtparser::ParserHelper::_GetCurrentFunction()
{
	return mCurrentFunctionStack.back();
}

void wtparser::ParserHelper::EnterEHandlerMode(bool en)
{
	mEHandlerModeStack.push_back(en);
}

void wtparser::ParserHelper::LeaveEHandlerMode()
{
	mEHandlerModeStack.pop_back();
}

bool wtparser::ParserHelper::_GetEHandlerMode()
{
	return mEHandlerModeStack.back();
}

void wtparser::ParserHelper::EnterFinallyMode(bool en)
{
	mFinallyModeStack.push_back(en);
}

void wtparser::ParserHelper::LeaveFinallyMode()
{
	mFinallyModeStack.pop_back();
}

bool wtparser::ParserHelper::_GetFinallyMode()
{
	return mFinallyModeStack.back();
}

void wtparser::ParserHelper::EnterDistinguishedBlockMode(bool en)
{
	mDistinguishedBlockModeStack.push_back(en);
}

void wtparser::ParserHelper::LeaveDistinguishedBlockMode()
{
	mDistinguishedBlockModeStack.pop_back();
}

bool wtparser::ParserHelper::_GetDistinguishedBlockMode()
{
	return mDistinguishedBlockModeStack.back();
}

wtast::RecordField* wtparser::ParserHelper::MakeRecordField(wtast::Type* type, const std::string & name)
{
	return new wtast::RecordField(type, name);
}

wtast::FunctionDeclaration* wtparser::ParserHelper::MakeFunctionDeclaration(wtast::FunctionSignature* funcsig)
{
	RegisterFunctionSignature(funcsig);
	return new wtast::FunctionDeclaration(funcsig);
}

wtast::FunctionDefinition* wtparser::ParserHelper::MakeFunctionDefinition(wtast::FunctionSignature* funcsig, wtast::PreCondition* precond, wtast::PostCondition* postcond, wtast::ExceptionCompound* ec)
{
	RegisterFunctionSignature(funcsig);
	return new wtast::FunctionDefinition(funcsig, precond, postcond, ec);
}

void wtparser::ParserHelper::RegisterFunctionSignature(wtast::FunctionSignature* funcsig)
{
	nametype_funcsig_t::iterator nt_it = mNametypeFuncsig.find(NameTypePair(funcsig->GetName(), funcsig->GetArgType()));
	if(mNametypeFuncsig.end()!=nt_it && nt_it->second->IsBuiltin() && !funcsig->IsBuiltin())
	{
		std::ostringstream ss;
		ss << "Function '";
		ss << funcsig->GetCanonicalName(true, true);
		ss << "' try to hide a builtin function.";

		PostError(ss.str());
		return;
	}

	map_uniqueid_funcsig_t::iterator it = mUniqueidFuncsig.find(funcsig->GetUniqueId());

	if(mUniqueidFuncsig.end()==it)
	{
		mUniqueidFuncsig[funcsig->GetUniqueId()] = funcsig;
	}
	else
	if(funcsig->IsPure()!=it->second->IsPure())
	{
		std::ostringstream ss;
		ss << "Function '";
		ss << funcsig->GetCanonicalName(true, true);
		ss << "' declared with different pure property.";

		PostError(ss.str());
	}
}

wt_u32 wtparser::ParserHelper::GetNextContextSpecificTypeId()
{
	if(!mDeclEntity)
	{
		PostError("Internal error: GetNextContextSpecificTypeId() called without a declaration entity.");
		return -1;
	}

	if(mDeclEntity->IsExport() || mDeclEntity->IsImport())
	{
		if(mNextPublicTypeId==FirstPublicTypeId+TypeIdRange)
		{
			PostError("Out of types identifiers.");
			return -1;
		}

		return mNextPublicTypeId++;
	}
	else
	{
		if(mNextPrivateTypeId==FirstPrivateTypeId+TypeIdRange)
		{
			PostError("Out of types identifiers.");
			return -1;
		}

		return mNextPrivateTypeId++;
	}
}

wtast::RecordDeclaration* wtparser::ParserHelper::MakeRecordDeclaration(const std::string & name, recordfields_t & fields, bool priv, bool lim, bool restricted)
{
	if(mTypenameToType.end()!=mTypenameToType.find(name))
	{
		std::ostringstream ss;
		ss << "Internal error: the name already defined '";
		ss << name;
		ss << "'.";
		PostError(ss.str());
		return 0;
	}

	wt_u32 type_id = GetNextContextSpecificTypeId() | WTTF_LOCAL;
	wtast::RecordType* rec = new wtast::RecordType(name, type_id, priv, lim, restricted);
	mTypenameToType[name] = rec;
	mTypeidToType[type_id] = rec;

	for(recordfields_t::iterator it=fields.begin(),end=fields.end();it!=end;++it)
	{
		if(!rec->AddField(*it))
		{
			std::ostringstream ss;
			ss << "The field '" << (*it)->GetName();
			ss << "' already exists in record '" << name;
			ss << "'.";
			PostError(ss.str());
		}
	}

	return new wtast::RecordDeclaration(rec);
}

wtast::FunctionTypeDeclaration* wtparser::ParserHelper::MakeFunctionTypeDeclaration(wtast::FunctionSignature* sig)
{
	const std::string & name = sig->GetName();

	if(mTypenameToType.end()!=mTypenameToType.find(name))
	{
		std::ostringstream ss;
		ss << "Internal error: the name already defined '";
		ss << name;
		ss << "'.";
		PostError(ss.str());
		return 0;
	}

	wt_u32 type_id = GetNextContextSpecificTypeId() | WTTF_LOCAL;
	wtast::FunctionType* functype = new wtast::FunctionType(sig, type_id);

	mTypenameToType[name] = functype;
	mTypeidToType[type_id] = functype;

	return new wtast::FunctionTypeDeclaration(functype);
}

wtast::FunctionPointerCall* wtparser::ParserHelper::MakeFunctionPointerCall(wtast::FunctionType* functype, wtast::Expression* funcptraccess, wtast::Expression* argexpr)
{
	if(!funcptraccess)
	{
		PostError("Invalid function pointer access expression.");
		return 0;
	}

	wtast::TypeInfo* const type_info = funcptraccess->GetType();

	if(type_info->IsSimultaneous())
	{
		PostError("Type of the function pointer access expression cannot be simultaneous.");
		return 0;
	}

	wtast::Type* const pfn_type = type_info->GetType(0);

	if(!pfn_type->IsFunctionType())
	{
		PostError("Type of the function pointer access expression must be a function type.");
		return 0;
	}

	if(pfn_type->GetTypeId()!=functype->GetTypeId())
	{
		PostError("Expected and actual function types not match.");
		return 0;
	}

	if(!ExpressionCanBeArgument(functype->GetFunctionSignature(), argexpr))
	{
		return 0;
	}

	if(argexpr && argexpr->GetType()->HasFunctionType()
		&& argexpr->OriginateFrom(wtast::AstNode::Variable::GLOBAL)
		&& funcptraccess->OriginateFrom(wtast::AstNode::Variable::FORMAL))
	{
		wtast::FunctionSignature* sig = functype->GetFunctionSignature();

		for(std::size_t i=0,n=sig->GetArgc();i<n;++i)
		{
			wtast::Type* const type = sig->GetArg(i)->GetType();

			if(type->HasFunctionType() && type->IsMutable())
			{
				PostError("Global data with function pointers cannot be passed as mutable data to argument function pointer.");
				return 0;
			}
		}
	}

	return new wtast::FunctionPointerCall(functype, funcptraccess, argexpr);
}

wtast::MakeFunctionPointer* wtparser::ParserHelper::MakeMakeFunctionPointer(const std::string & functype_name, const std::string & funcname)
{
	map_name_type_t::iterator type_it = mTypenameToType.find(functype_name);

	if(mTypenameToType.end()==type_it)
	{
		std::ostringstream ss;

		ss << "'" << functype_name << "' is not a name of a type.";

		PostError(ss.str());
		return 0;
	}

	wtast::Type* const type = type_it->second;

	if(!type->IsFunctionType())
	{
		std::ostringstream ss;

		ss << "'" << functype_name << "' is not a function type.";

		PostError(ss.str());
		return 0;
	}

	wtast::FunctionType* functype = static_cast<wtast::FunctionType*>(type);
	const bool func_is_pure = functype->GetFunctionSignature()->IsPure();

	wtast::FunctionSignature* funcsig = ResolveFunction(funcname, functype->GetFunctionSignature()->GetArgType());

	if(!funcsig)
	{
		return 0;
	}

	if(func_is_pure && !funcsig->IsPure())
	{
		std::ostringstream ss;

		ss << "Function '" << funcname << " is not pure, but pure was expected.";

		PostError(ss.str());
		return 0;
	}

	bool bad = false;

	if(functype->GetFunctionSignature()->GetRetc()!=funcsig->GetRetc())
	{
		bad = true;
	}
	else
	{
		for(std::size_t i=0,n=funcsig->GetRetc();!bad && i<n;++i)
		{
			if(!IsSameType(functype->GetFunctionSignature()->GetRet(i), funcsig->GetRet(i)))
			{
				bad = true;
			}
		}

		for(std::size_t i=0,n=funcsig->GetArgc();!bad && i<n;++i)
		{
			if(functype->GetFunctionSignature()->GetArg(i)->GetType()->IsMutable()
				&& !funcsig->GetArg(i)->GetType()->IsMutable())
			{
				bad = true;
			}
		}
	}

	if(bad)
	{
		std::ostringstream ss;

		ss << "The signature of the function '" << funcname;
		ss << "' do not match with the signature of the function type.";

		PostError(ss.str());
		return 0;
	}

	return new wtast::MakeFunctionPointer(functype, funcsig);
}

wtast::DataType* wtparser::ParserHelper::MakeDataType(const std::string & name)
{
	if(mTypenameToType.end()!=mTypenameToType.find(name))
	{
		std::ostringstream ss;
		ss << "Internal error: the name already defined '";
		ss << name;
		ss << "'.";
		PostError(ss.str());
		return 0;
	}

	wt_u32 type_id = GetNextContextSpecificTypeId() | WTTF_LOCAL;
	wtast::DataType* datatype = new wtast::DataType(name, type_id);

	mTypenameToType[name] = datatype;
	mTypeidToType[type_id] = datatype;

	return datatype;
}

wtast::DataTypeCons* wtparser::ParserHelper::MakeDataTypeCons(const std::string & name)
{
	wtast::DataTypeCons* cons = new wtast::DataTypeCons(name);

	mNameToDataTypeConses[name].insert(cons);
	mConsToDataType[cons] = mCurrentDataType;

	return cons;
}

wtast::DataTypeDeclaration* wtparser::ParserHelper::MakeDataTypeDeclaration(wtast::DataType* datatype, datatypeconses_t & conses)
{
	bool error = false;

	for(datatypeconses_t::iterator it=conses.begin(),end=conses.end();it!=end;++it)
	{
		wtast::DataTypeCons* cons = *it;

		if(!cons) { continue; }

		if(datatype->GetCons(cons->GetName()))
		{
			std::ostringstream ss;
			ss << "Datatype '" << datatype->GetName();
			ss << "' already has the '";
			ss << cons->GetName();
			ss << "' data constructor.";
			PostError(ss.str());
			error = true;
			delete cons;
			continue;
		}

		if(cons->IsDefault())
		{
			if(datatype->GetDefaultCons())
			{
				std::ostringstream ss;
				ss << "'" << cons->GetName() << "'";
				ss << " could not be set as default constructor, because '";
				ss << datatype->GetName();
				ss << "' already has one.";
				PostError(ss.str());
				error = true;
				delete cons;
				continue;
			}

			datatype->SetDefaultCons(cons);
		}

		datatype->AddCons(cons);
	}

	if(error)
	{
		delete datatype;
		return 0;
	}

	return new wtast::DataTypeDeclaration(datatype);
}

void wtparser::ParserHelper::SetCurrentDataType(wtast::DataType* datatype)
{
	mCurrentDataType = datatype;
}

void wtparser::ParserHelper::SetCurrentEnumType(wtast::EnumType* enumtype)
{
	mCurrentEnumType = enumtype;
}

wtast::EnumType* wtparser::ParserHelper::MakeEnumType(const std::string & name)
{
	if(mTypenameToType.end()!=mTypenameToType.find(name))
	{
		std::ostringstream ss;
		ss << "Internal error: the name already defined '";
		ss << name;
		ss << "'.";
		PostError(ss.str());
		return 0;
	}

	if(name=="s")
	{
		PostError("'s' is not a valid enumeration type name.");
		return 0;
	}

	wt_u32 type_id = GetNextContextSpecificTypeId() | WTTF_LOCAL;
	wtast::EnumType* enumtype = new wtast::EnumType(name, type_id);

	mTypenameToType[name] = enumtype;
	mTypeidToType[type_id] = enumtype;

	if(!_AddEnumOperations(enumtype, type_id))
	{
		delete enumtype;
		return 0;
	}

	return enumtype;
}

wtast::Enumerator* wtparser::ParserHelper::MakeEnumerator(const std::string & name)
{
	wtast::Enumerator* enumerator = new wtast::Enumerator(name);

	mNameToEnumerators[name].insert(enumerator);
	mEnumeratorToEnumType[enumerator] = mCurrentEnumType;

	return enumerator;
}

wtast::EnumTypeDeclaration* wtparser::ParserHelper::MakeEnumTypeDeclaration(wtast::EnumType* enumtype, enumerators_t & enumerators)
{
	bool error = false;

	for(enumerators_t::iterator it=enumerators.begin(),end=enumerators.end();it!=end;++it)
	{
		wtast::Enumerator* enumerator = *it;

		if(!enumerator) { continue; }

		if(enumtype->GetEnumerator(enumerator->GetName()))
		{
			std::ostringstream ss;
			ss << "Enum '" << enumtype->GetName();
			ss << "' already has the '";
			ss << enumerator->GetName();
			ss << "' enumerator.";
			PostError(ss.str());
			error = true;
			delete enumerator;
			continue;
		}

		enumtype->AddEnumerator(enumerator);
	}

	if(error)
	{
		delete enumtype;
		return 0;
	}

	return new wtast::EnumTypeDeclaration(enumtype);
}

wtast::ExceptionDeclaration* wtparser::ParserHelper::MakeExceptionDeclaration(const std::string & str)
{
	UniqueId unique_id = CreateUniqueId();
	mExceptionToId[str] = unique_id;
	return new wtast::ExceptionDeclaration(str, unique_id);
}

wtast::OperatorDeclaration* wtparser::ParserHelper::MakeOperatorDeclaration(const std::string & opname, const std::string & assoc, wt_i32 prec)
{
	bool left_assoc = (StrLeftAssoc==assoc);

	if(StrLeftAssoc!=assoc && StrRightAssoc!=assoc)
	{
		std::ostringstream ss;
		ss << "The operator associativity must be '";
		ss << StrLeftAssoc << "' or '";
		ss << StrRightAssoc << "'.";
		PostError(ss.str());
		return 0;
	}

	if(MinPrecedence>prec || MaxPrecedence<prec)
	{
		std::ostringstream ss;
		ss << "The operator recedence must be between ";
		ss << MinPrecedence << " and ";
		ss << MaxPrecedence << ".";
		PostError(ss.str());
		return 0;
	}

	if(mOpnameToOp.find(opname)!=mOpnameToOp.end())
	{
		std::ostringstream ss;
		ss << "The operator '" << opname << "' already exists.";
		PostError(ss.str());
		return 0;
	}

	wtast::OperatorDeclaration* opdecl = new wtast::OperatorDeclaration(opname, left_assoc, prec);
	mOpnameToOp[opname] = OperatorTokenPair(opdecl, GetOperatorToken(left_assoc, prec));
	return opdecl;
}

wtast::/*Alias*/Declaration* wtparser::ParserHelper::MakeAliasDeclaration(const std::string & alias, wtast::Type* type)
{
	// TODO
	return 0;
}

void wtparser::ParserHelper::AddDeclaration(wtast::Declaration* decl)
{
	if(!mDeclEntity)
	{
		mDeclEntity = new wtast::DeclareBlock(false, false);
	}

	mDeclEntity->AddDeclarationBack(decl);

	if(wtast::AstNode::RECORDDECLARATION==decl->GetKind())
	{
		wtast::RecordType* rec = static_cast<wtast::RecordDeclaration*>(decl)->GetRecord();

		if(mDeclEntity->IsExport() && !rec->IsPrivate() && !rec->IsLimited())
		{
			for(std::size_t i=0,n=rec->GetNumberOfFields();i!=n;++i)
			{
				wtast::Type* type = rec->GetField(i)->GetType();

				if(mLocalTypenames.count(type->GetName()))
				{
					std::ostringstream ss;
					ss << "Record '";
					ss << rec->GetName();
					ss << "' can not be exported, because type '";
					ss << type->GetName();
					ss << "' is local.";
					PostError(ss.str());
				}
			}
		}
		else
		if(!mDeclEntity->IsExport() && !mDeclEntity->IsImport()) // nem export es nem import -> local
		{
			mLocalTypenames.insert(rec->GetName());
		}
	}
	else
	if(wtast::AstNode::FUNCTIONDECLARATION==decl->GetKind() && mDeclEntity->IsExport())
	{
		wtast::FunctionSignature* sig = static_cast<wtast::FunctionDeclaration*>(decl)->GetFunctionSignature();

		for(std::size_t i=0,n=sig->GetArgc();i!=n;++i)
		{
			wtast::Type* type = sig->GetArg(i)->GetType();

			if(mLocalTypenames.count(type->GetName()))
			{
				std::ostringstream ss;
				ss << "Function '";
				ss << sig->GetCanonicalName(true, true);
				ss << "' can not be exported, because argument type '";
				ss << type->GetName();
				ss << "' is local.";
				PostError(ss.str());
			}
		}

		for(std::size_t i=0,n=sig->GetRetc();i!=n;++i)
		{
			wtast::Type* type = sig->GetRet(i);

			if(mLocalTypenames.count(type->GetName()))
			{
				std::ostringstream ss;
				ss << "Function '";
				ss << sig->GetCanonicalName(true, true);
				ss << "' can not be exported, because return type '";
				ss << type->GetName();
				ss << "' is local.";
				PostError(ss.str());
			}
		}
	}
}

wtast::LValue* wtparser::ParserHelper::ResolveMemberAccess(wtast::LValue* lvalue, const std::string & fieldname)
{
	if(!lvalue)
	{
		std::ostringstream ss;
		ss << "Member access of '";
		ss << fieldname;
		ss << "' cannot be resolved, because the lvalue is invalid.";
		PostError(ss.str());
		return 0;
	}

	wtast::TypeInfo* typeinfo = lvalue->GetType();

	if(wtast::AstNode::RECORDTYPE!=typeinfo->GetKind())
	{
		std::ostringstream ss;
		ss << "The type '" << typeinfo->GetName();
		ss << "' does not have members.";
		PostError(ss.str());
		return 0;
	}

	wtast::RecordType*  rec   = static_cast<wtast::RecordType*>(typeinfo);
	wtast::RecordField* field = rec->GetField(fieldname);

	if(field)
	{
		return new wtast::MemberAccess(lvalue, fieldname, field->GetType(), field->GetOffset());
	}
	else
	{
		std::ostringstream ss;
		ss << "'" << fieldname;
		ss << "' is not a member of '";
		ss << rec->GetName() << "'";
		PostError(ss.str());
	}

	return 0;
}

wtast::Expression* wtparser::ParserHelper::MakeOrElse(wtast::Expression* lhs, wtast::Expression* rhs)
{
	if(lhs && rhs && IsBoolType(lhs->GetType()) && IsBoolType(rhs->GetType()))
	{
		return new wtast::OrElse(lhs, rhs, GetStockType(WTT_BOOL));
	}
	else
	{
		PostError("The or-else operator works only on 'bool' types.");
		return 0;
	}
}

wtast::Expression* wtparser::ParserHelper::MakeAndThen(wtast::Expression* lhs, wtast::Expression* rhs)
{
	if(lhs && rhs && IsBoolType(lhs->GetType()) && IsBoolType(rhs->GetType()))
	{
		return new wtast::AndThen(lhs, rhs, GetStockType(WTT_BOOL));
	}
	else
	{
		PostError("The and-then operator works only on 'bool' types.");
		return 0;
	}
}

wtast::Expression* wtparser::ParserHelper::ApplyBinaryOperator(wtast::Expression* lhs, const std::string & op, wtast::Expression* rhs)
{
	if(!lhs || !lhs->GetType() || !rhs || !rhs->GetType())
	{
		return 0;
	}

	wtast::SimultaneousExpression* expr = new wtast::SimultaneousExpression(lhs, rhs);
	wtast::Expression* result = ResolveFunctionCall(new wtast::UnresolvedFunctionName(op), expr);

	if(lhs->IsSimultaneous())
	{
		static_cast<wtast::SimultaneousExpression*>(lhs)->mExpressions.clear();
		delete lhs;
	}

	if(rhs->IsSimultaneous())
	{
		static_cast<wtast::SimultaneousExpression*>(rhs)->mExpressions.clear();
		delete rhs;
	}

	return result;
}

wtast::Expression* wtparser::ParserHelper::ApplyUnaryOperator(wtast::Expression* lhs, const std::string & op)
{
	return ResolveFunctionCall(new wtast::UnresolvedFunctionName(op), lhs);
}

wtast::ReturnStatement* wtparser::ParserHelper::MakeReturnStatement(wtast::Expression* expr)
{
	if(_GetFinallyMode())
	{
		PostError("Finally can not contain return statement.");
		return 0;
	}

	wtast::FunctionSignature* current_function = _GetCurrentFunction();
	const bool distinguished_block_mode = _GetDistinguishedBlockMode();

	if(current_function && distinguished_block_mode)
	{
		PostError("Internal error: both CurrentFunction and DistinguishedBlockMode are set.");
		return 0;
	}

	if(!current_function && !distinguished_block_mode)
	{
		PostError("This return statement must be placed into a function or main/ctor/dtor block.");
		return 0;
	}

	if(distinguished_block_mode)
	{
		if(expr)
		{
			PostError("Only the empty return statement can be accepted here.");
			return 0;
		}
	}
	else
	{
		if((0<current_function->GetRetc() && !expr) || (0==current_function->GetRetc() && expr))
		{
print_mismatch_error:
			PostError("Type of the function return type and the type of the returned value mismatch.");

			wtast::TypeInfo* typeinfo = current_function->GetRetType();

			std::ostringstream ss;

			if(typeinfo)
			{
				ss << "Accepted return value type is: '";
				ss << typeinfo->GetName();
				ss << "'.";

				delete typeinfo;
			}
			else
			{
				ss << "The function has no return type, and the return statement must reflect this.";
			}

			PostNote(ss.str());

			return 0;
		}
		else
		if(expr)
		{
			wtast::TypeInfo* typeinfo = expr->GetType();

			if(1==current_function->GetRetc() && !typeinfo->IsSimultaneous())
			{
				if(current_function->GetRet(0)->GetTypeId()!=(static_cast<wtast::Type*>(typeinfo)->GetTypeId() & ~WTTF_MUTABLE))
				{
					goto print_mismatch_error;
				}

				// OK otherwise
			}
			else
			{
				wtast::SimultaneousType* simtype = static_cast<wtast::SimultaneousType*>(typeinfo);

				if(current_function->GetRetc()!=simtype->GetNumberOfTypes())
				{
					goto print_mismatch_error;
				}

				for(std::size_t i=0,n=current_function->GetRetc();i<n;++i)
				{
					if(current_function->GetRet(i)->GetTypeId()!=(simtype->GetType(i)->GetTypeId() & ~WTTF_MUTABLE))
					{
						goto print_mismatch_error;
					}
				}

				// OK otherwise
			}
		}
	}

	return new wtast::ReturnStatement(expr);
}

wtast::RaiseStatement* wtparser::ParserHelper::MakeRaiseStatement(const std::string & exc, bool near)
{
	map_name_uniqueid_t::iterator it = mExceptionToId.find(exc);

	if(mExceptionToId.end()==it)
	{
		std::ostringstream ss;
		ss << "The exception '" << exc << "' is not a known exception.";
		PostError(ss.str());
		return 0;
	}

	return new wtast::RaiseStatement(exc, it->second, near);
}

wtast::WhileStatement* wtparser::ParserHelper::MakeWhileStatement(wtast::Expression* cond, wtast::Compound* body)
{
	if(!cond || !IsBoolType(cond->GetType()))
	{
		PostError("Type of the condition of the 'while' statement must be 'bool'.");
		return 0;
	}

	return new wtast::WhileStatement(cond, body);
}

wtast::DoWhileStatement* wtparser::ParserHelper::MakeDoWhileStatement(wtast::Compound* body, wtast::Expression* cond)
{
	if(!cond || !IsBoolType(cond->GetType()))
	{
		PostError("Type of the condition of the 'do-while' statement must be 'bool'.");
		return 0;
	}

	return new wtast::DoWhileStatement(body, cond);
}

wtast::ForStatement* wtparser::ParserHelper::MakeForStatement(wtast::Assignment* init, wtast::Expression* cond, wtast::Assignment* step, wtast::Compound* body)
{
	if(cond && !IsBoolType(cond->GetType()))
	{
		PostError("Type of the condition of the 'for' statement must be 'bool'.");
		return 0;
	}

	return new wtast::ForStatement(init, cond, step, body);
}

wtast::VariableDeclaration* wtparser::ParserHelper::EnterForeachBlock(const std::string & varname, wtast::Expression* from, bool range_based)
{
	if(!from)
	{
		std::ostringstream ss;
		ss << "Error in foreach ";
		if(range_based) { ss << "interval"; } else { ss << "container"; }
		ss << ".";
		PostError(ss.str());
		return 0;
	}

	if(from->GetType()->IsSimultaneous())
	{
		std::ostringstream ss;
		ss << "Type of the ";
		if(range_based) { ss << "interval"; } else { ss << "container"; }
		ss << " of the 'foreach' statement can not be simultaneous.";
		PostError(ss.str());
		return 0;
	}

	wtast::Type* type = static_cast<wtast::Type*>(from->GetType());
	wtast::Type* new_type = 0;

	if(range_based)
	{
		// check range type

		if(wtast::AstNode::BASICTYPE!=type->GetKind())
		{
			PostError("Range type must be a basic type.");
			return 0;
		}

		switch(static_cast<wtast::BasicType*>(type)->GetTypeId())
		{
		case WTT_I8:
		case WTT_U8:
		case WTT_I16:
		case WTT_U16:
		case WTT_I32:
		case WTT_U32:
		case WTT_I64:
		case WTT_U64:
		case WTT_CHAR:
//		case WTT_WCHAR:
			new_type = type->Copy();
			break;

		default:
			PostError("Unsupported range type.");
			return 0;
		}
	}
	else
	{
		// get element type from 'type'

		if(wtast::AstNode::SEQTYPE==type->GetKind())
		{
			new_type = static_cast<wtast::SeqType*>(type)->GetElementType()->Copy();
		}
		else
		if(wtast::AstNode::BASICTYPE==type->GetKind() && WTT_STRING==static_cast<wtast::BasicType*>(type)->GetTypeId())
		{
			new_type = GetStockType(WTT_CHAR);
		}
//		else
//		if(wtast::AstNode::BASICTYPE==type->GetKind() && WTT_WSTRING==static_cast<wtast::BasicType*>(type)->GetTypeId())
//		{
//			new_type = GetStockType(WTT_WCHAR);
//		}
		else
		{
			PostError("Accepted containers are: 'seq<T>' and 'string'.");
			return 0;
		}
	}

	if(!new_type || !mSymTab.EnterBlock())
	{
		PostError("Out of memory.");
		return 0;
	}

	UniqueId unique_id = CreateUniqueId();

	if(!mSymTab.AddVariable(new_type, varname, unique_id, _GetEHandlerMode(), wtast::AstNode::Variable::READONLY))
	{
		std::ostringstream ss;
		ss << "Variable '" << varname << "' can not be created.";
		PostError(ss.str());
		return 0;
	}

	wtast::VariableDeclaration* vardecl = new wtast::VariableDeclaration(new_type);
	vardecl->AddFragment(new wtast::VariableDeclFragment(vardecl, varname, unique_id));
	return vardecl;
}

wtast::ForeachStatement* wtparser::ParserHelper::MakeForeachStatement(wtast::VariableDeclaration* vardecl, wtast::Expression* from, wtast::Expression* to, bool inclusive, bool reverse, wtast::Expression* whil, wtast::Expression* whic, wtast::Compound* body)
{
	if(!from || !from->GetType() || !to || !to->GetType())
	{
		PostError("Invalid foreach range.");
		return 0;
	}

	if(from->GetType()->IsSimultaneous() || to->GetType()->IsSimultaneous())
	{
		PostError("Type of the interval start/end of the 'foreach' statement can not be simultaneous.");
		return 0;
	}

	if(static_cast<wtast::Type*>(from->GetType())->GetTypeId()!=static_cast<wtast::Type*>(to->GetType())->GetTypeId())
	{
		PostError("Type of the interval start/end of the 'foreach' statement must be the same.");
		return 0;
	}

	if(whil && !IsBoolType(whil->GetType()))
	{
		PostError("Type of the 'while' condition of the 'foreach' statement must be 'bool'.");
		return 0;
	}

	if(whic && !IsBoolType(whic->GetType()))
	{
		PostError("Type of the 'which' condition of the 'foreach' statement must be 'bool'.");
		return 0;
	}

	return new wtast::ForeachStatement(vardecl, from, to, whil, whic, body, inclusive, reverse);
}

wtast::ForeachStatement* wtparser::ParserHelper::MakeForeachStatement(wtast::VariableDeclaration* vardecl, wtast::Expression* container, bool reverse, wtast::Expression* whil, wtast::Expression* whic, wtast::Compound* body)
{
	if(!container)
	{
		return 0;
	}

	if(container->GetType()->IsSimultaneous())
	{
		PostError("Type of the container of the 'foreach' statement can not be simultaneous.");
		return 0;
	}

	if(whil && !IsBoolType(whil->GetType()))
	{
		PostError("Type of the 'while' condition of the 'foreach' statement must be 'bool'.");
		return 0;
	}

	if(whic && !IsBoolType(whic->GetType()))
	{
		PostError("Type of the 'which' condition of the 'foreach' statement must be 'bool'.");
		return 0;
	}

	return new wtast::ForeachStatement(vardecl, container, whil, whic, body, reverse);
}

wtast::FunctionArgument* wtparser::ParserHelper::MakeFunctionArgument(wtast::Type* type, const std::string & name, bool is_mutable)
{
	if(is_mutable)
	{
		map_u32_type_t::iterator it = mTypeidToType.find(WTTF_MUTABLE|type->GetTypeId());

		if(mTypeidToType.end()!=it)
		{
			type = it->second;
		}
		else
		{
			type = type->Copy();
			type->SetMutable();

			mTypeidToType[type->GetTypeId()] = type;
		}
	}

	return new wtast::FunctionArgument(type, name);
}

wtast::IfStatement* wtparser::ParserHelper::MakeIfStatement(wtast::Expression* cond, wtast::Compound* then, wtast::Statement* els)
{
	if(!cond || !IsBoolType(cond->GetType()))
	{
		PostError("Type of the 'if' condition must be 'bool'.");
		return 0;
	}

	return new wtast::IfStatement(cond, then, els);
}

wtast::AssertStatement* wtparser::ParserHelper::MakeAssertStatement(wtast::Expression* assertion)
{
	if(!assertion || !IsBoolType(assertion->GetType()))
	{
		PostError("Type of the 'assert' condition must be 'bool'.");
		return 0;
	}

	return new wtast::AssertStatement(assertion);
}

wtast::PreCondition* wtparser::ParserHelper::MakePrecondition(wtast::Expression* cond)
{
	if(!cond || !IsBoolType(cond->GetType()))
	{
		PostError("Type of the 'pre' condition must be 'bool'.");
		return 0;
	}

	if(!checker::SubtreeIsPure(cond))
	{
		PostError("Condition must be pure.");
		return 0;
	}

	return new wtast::PreCondition(cond);
}

variabledecls_t* wtparser::ParserHelper::AddPostcondRets(identlist_t* rets)
{
	wtast::FunctionSignature* current_function = _GetCurrentFunction();

	if(!current_function)
	{
		PostError("Post-condition must be placed in a function.");
		return 0;
	}

	if(!rets)
	{
		return 0;
	}

	if(rets->size()!=current_function->GetRetc())
	{
		PostError("Number of return values in post-condition does not match.");
		return 0;
	}

	variabledecls_t* vardecls = new variabledecls_t;
	std::size_t      retidx   = 0;

	for(identlist_t::iterator it=rets->begin(),end=rets->end();it!=end;++retidx,++it)
	{
		wtast::Type*        ret_type  = current_function->GetRet(retidx);
		UniqueId            unique_id = CreateUniqueId();
		const std::string & ident     = *it;

		if(mSymTab.AddVariable(ret_type, ident, unique_id, _GetEHandlerMode(), wtast::AstNode::Variable::READONLY))
		{
			wtast::VariableDeclaration* decl = new wtast::VariableDeclaration(ret_type->Copy());
			decl->AddFragment(new wtast::VariableDeclFragment(decl, ident, unique_id));
			vardecls->push_back(decl);
		}
		else
		{
			PostError("Failed to add return variable.");
		}
	}

	return vardecls;
}

variabledecls_t* wtparser::ParserHelper::AddPostcondOlds(identlist_t* olds)
{
	if(!_GetCurrentFunction())
	{
		PostError("Post-condition must be placed in a function.");
		return 0;
	}

	if(!olds)
	{
		return 0;
	}

	variabledecls_t* vardecls = new variabledecls_t;

	for(identlist_t::iterator it=olds->begin(),end=olds->end();it!=end;++it)
	{
		UniqueId            unique_id = CreateUniqueId();
		const std::string & ident     = *it;
		const std::string & ident2    = std::string("old ") + ident;
		SymTab::Resolver    resolver;

		if(!mSymTab.Resolve(resolver, ident, SymTab::UnnamedFrame) || wtast::AstNode::Variable::FORMAL!=resolver.varclass)
		{
			PostError(std::string("Identifier '") + ident + "' is not a formal parameter.");
			continue;
		}

		if(mSymTab.AddVariable(resolver.type, ident2, unique_id, _GetEHandlerMode(), wtast::AstNode::Variable::READONLY))
		{
			wtast::VariableDeclaration* decl = new wtast::VariableDeclaration(resolver.type->Copy());
			decl->AddFragment(new wtast::VariableDeclFragment(decl, ident2, unique_id));
			vardecls->push_back(decl);
		}
		else
		{
			PostError("Failed to add old value.");
		}
	}

	return vardecls;
}

wtast::PostCondition* wtparser::ParserHelper::MakePostcondition(variabledecls_t* rets, variabledecls_t* olds, wtast::Expression* cond)
{
	if(!cond || !IsBoolType(cond->GetType()))
	{
		PostError("Type of the 'post' condition must be 'bool'.");
		return 0;
	}

	if(!checker::SubtreeIsPure(cond))
	{
		PostError("Condition must be pure.");
		return 0;
	}

	wtast::PostCondition* postcond = new wtast::PostCondition(cond);
	if(!postcond)
	{
		return 0;
	}

	if(rets)
	{
		postcond->AddReturnValues(rets->begin(), rets->end());

		for(variabledecls_t::iterator it=rets->begin(),end=rets->end();it!=end;++it)
		{
			HideVariables(*it);
		}
	}

	if(olds)
	{
		postcond->AddOldValues(olds->begin(), olds->end());

		for(variabledecls_t::iterator it=olds->begin(),end=olds->end();it!=end;++it)
		{
			HideVariables(*it);
		}
	}

	return postcond;
}

void wtparser::ParserHelper::HideVariables(wtast::VariableDeclaration* decl)
{
	for(std::size_t i=0,n=decl->GetNumberOfFragments();i!=n;++i)
	{
		mSymTab.HideVariable(decl->GetFragment(i)->GetUniqueId());
	}
}

wtast::LValue* wtparser::ParserHelper::ResolveVariableAccess(const std::string & varname, const std::string & scopename)
{
	SymTab::Resolver resolver;

	if(mSymTab.Resolve(resolver, varname, scopename))
	{
		if(_GetEHandlerMode() && !resolver.tagged && (wtast::AstNode::Variable::LOCAL==resolver.varclass))
		{
			std::ostringstream ss;

			ss << "Exception cannot access local variable '";
			ss << varname << "', because it is defined outside the ";

			// EHandlerMode and FinallyMode can be combined
			if(_GetFinallyMode())
			{
				ss << "finally block.";
			}
			else
			{
				ss << "exception handler.";
			}

			PostError(ss.str());

			return 0;
		}

		return new wtast::VariableAccess(resolver.type, resolver.name, resolver.frame_name, resolver.unique_id, resolver.varclass);
	}
	else
	{
		if(SymTab::UnnamedFrame==scopename &&
			(
				mNameFuncsig.end()!=mNameFuncsig.find(varname)
				||
				"seqfill"==varname
			)
		)
		{
			return new wtast::UnresolvedFunctionName(varname);
		}

		std::ostringstream ss;

		ss << "Undeclared identifier: ";
		ss << "'" << varname << "'";

		if(scopename!=SymTab::UnnamedFrame)
		{
			ss << " in '" << scopename << "'";
		}

		ss << '.';

		PostError(ss.str());
		return 0;
	}
}

wtast::SeqLiteral* wtparser::ParserHelper::MakeSeqLiteral(wtast::Type* elemtype, wtast::Expression* init)
{
	if(!elemtype && !init)
	{
		PostError("Invalid sequence literal.");
		return 0;
	}

	if(init)
	{
		wtast::TypeInfo* typeinfo   = init->GetType();
		wtast::Type*     first_type = typeinfo->GetType(0);

		if(!first_type)
		{
			PostError("Invalid sequence literal.");
			return 0;
		}

		if(!elemtype)
		{
			elemtype = first_type->Copy();

			if(!elemtype)
			{
				PostError("Failed to allocate type information: out of memory.");
				return 0;
			}
		}

		const wt_u32 first_typeid = first_type->GetTypeId();

		if(elemtype->GetTypeId()!=first_typeid)
		{
			PostError("Type of the initializer and the type of the sequence element mismatch.");
			return 0;
		}

		for(std::size_t i=1,n=typeinfo->GetNumberOfTypes();i<n;++i)
		{
			if(first_typeid!=typeinfo->GetType(i)->GetTypeId())
			{
				PostError("Type of the members of the sequence literal initializer mismatch.");
				return 0;
			}
		}
	}

	wt_u32 seq_typeid = GetSeqTypeId(elemtype);

	if(WTT_VOID==seq_typeid)
	{
		PostError("Can not construct sequence type.");
		return 0;
	}

	wtast::SeqType* seqtype = new wtast::SeqType(elemtype, seq_typeid);

	if(!seqtype)
	{
		PostError("Internal error: can not create type for sequence literal");
		return 0;
	}

	return new wtast::SeqLiteral(seqtype, init);
}

wtast::RecordLiteral* wtparser::ParserHelper::MakeRecordLiteral(const std::string & type_name, wtast::Expression* initexpr, fieldassigns_t * assigns)
{
	if(initexpr && assigns)
	{
		PostError("Record literal can not be created from initexpr and assigns.");
		return 0;
	}

	wtast::Type* raw_type = ResolveType(type_name);

	if(!raw_type || wtast::AstNode::RECORDTYPE!=raw_type->GetKind())
	{
		PostError(std::string("'") + type_name + "' is not a type name or not a record type.");
		return 0;
	}

	wtast::RecordType* rec = static_cast<wtast::RecordType*>(raw_type);
	std::size_t num_fields = rec->GetNumberOfFields();

	if(initexpr)
	{
		wtast::TypeInfo* expr_type = initexpr->GetType();
		std::size_t num_types  = expr_type->GetNumberOfTypes();

		bool match = num_fields==num_types;

		for(std::size_t i=0;match && i<num_fields;++i)
		{
			match = IsSameType(rec->GetField(i)->GetType(), expr_type->GetType(i), Usage::Assign);
		}

		if(!match)
		{
			PostError("Type of the initializer expression does not match.");
			return 0;
		}

		wtast::FieldAssign* fa = new wtast::FieldAssign(initexpr);

		for(std::size_t i=0;match && i<num_fields;++i)
		{
			fa->AddFieldNameBack(rec->GetField(i)->GetName());
		}

		wtast::RecordLiteral* reclit = new wtast::RecordLiteral(rec);
		reclit->AddAssignment(fa);
		return reclit;
	}

	wtast::RecordLiteral* reclit = new wtast::RecordLiteral(rec);

	std::set<std::string> available_fields;
	bool bad = false; // atmasoljuk az osszes mutatot a 'reclit'-be, es egyutt toroljuk le, ha kell

	for(std::size_t i=0;i<num_fields;++i)
	{
		available_fields.insert(rec->GetField(i)->GetName());
	}

	if(assigns)
	{
		for(fieldassigns_t::iterator it=assigns->begin(),end=assigns->end();it!=end;++it)
		{
			wtast::FieldAssign* fa = *it;
			wtast::TypeInfo* ft;

			if(!fa || !fa->GetExpression() ||
				!(ft=fa->GetExpression()->GetType()))
			{
				PostError("Badly formatted record literal field assign.");
				bad = true;
			}
			else
			{
				for(std::size_t i=0,n=fa->GetNumberOfFieldNames();i!=n;++i)
				{
					const std::string & field_name = fa->GetFieldName(i);
					wtast::RecordField* field = rec->GetField(field_name);

					if(!field)
					{
						PostError(std::string("Field assignment contains unknown field: '") + field_name + "'");
						bad = true;
						continue;
					}

					if(!available_fields.erase(field_name))
					{
						PostError(std::string("Field assignment contains duplicate field: '") + field_name + "'");
						bad = true;
						continue;
					}

					if(!ft->GetType(i))
					{
						PostError(std::string("Invalid assignment of field '") + field_name + "'.");
						bad = true;
						continue;
					}

					if(!IsSameType(field->GetType(), ft->GetType(i)))
					{
						PostError(std::string("Type of field assignment and field type mismatch: '") + field_name + "'");
						bad = true;
						continue;
					}
				}

				reclit->AddAssignment(fa);
			}
		}
	}

	if(bad)
	{
		delete reclit;
		return 0;
	}

	return reclit;
}

wtast::RecordUpdate* wtparser::ParserHelper::MakeRecordUpdate(wtast::Expression* expr, fieldassigns_t * assigns)
{
	if(!expr)
	{
		PostError("Bad expression in record update.");
		return 0;
	}

	if(expr->IsSimultaneous())
	{
		PostError("Simultaneous expression in record update.");
		return 0;
	}

	wtast::Type* const raw_type = expr->GetType()->GetType(0);

	if(wtast::AstNode::RECORDTYPE!=raw_type->GetKind())
	{
		PostError(std::string("Only record types are allowed in record update, but \"") + raw_type->GetName() + "\" is not a record type.");
		return 0;
	}

	wtast::RecordType* rec = static_cast<wtast::RecordType*>(raw_type);
	std::size_t num_fields = rec->GetNumberOfFields();

	wtast::RecordUpdate* recupd = new wtast::RecordUpdate(expr);

	std::set<std::string> available_fields;
	bool bad = false; // atmasoljuk az osszes mutatot a 'reclit'-be, es egyutt toroljuk le, ha kell

	for(std::size_t i=0;i<num_fields;++i)
	{
		available_fields.insert(rec->GetField(i)->GetName());
	}

	for(fieldassigns_t::iterator it=assigns->begin(),end=assigns->end();it!=end;++it)
	{
		wtast::FieldAssign* fa = *it;

		if(!fa || !fa->GetExpression())
		{
			PostError("Badly formatted record update field assign.");
			bad = true;
		}
		else
		{
			wtast::TypeInfo* ft = fa->GetExpression()->GetType();

			for(std::size_t i=0,n=fa->GetNumberOfFieldNames();i!=n;++i)
			{
				const std::string & field_name = fa->GetFieldName(i);
				wtast::RecordField* field = rec->GetField(field_name);

				if(!field)
				{
					PostError(std::string("Field assignment contains unknown field: '") + field_name + "'");
					bad = true;
					continue;
				}

				if(!available_fields.erase(field_name))
				{
					PostError(std::string("Field assignment contains duplicate field: '") + field_name + "'");
					bad = true;
					continue;
				}

				if(!ft->GetType(i))
				{
					PostError(std::string("Invalid assignment of field '") + field_name + "'.");
					bad = true;
					continue;
				}

				if(!IsSameType(field->GetType(), ft->GetType(i)))
				{
					PostError(std::string("Type of field assignment and field type mismatch: '") + field_name + "'");
					bad = true;
					continue;
				}
			}

			recupd->AddAssignment(fa);
		}
	}

	if(bad)
	{
		delete recupd;
		return 0;
	}

	return recupd;
}

wtast::DataTypeLiteral* wtparser::ParserHelper::MakeDataTypeLiteral(const std::string & cons_name, wtast::Expression* expr)
{
	map_name_datatypeconses_t::iterator cons_it = mNameToDataTypeConses.find(cons_name);

	if(mNameToDataTypeConses.end()==cons_it)
	{
		PostError(std::string("Data type constructor not found: ") + cons_name);
		return 0;
	}

	if(1!=cons_it->second.size())
	{
		PostError(std::string("Ambiguous data type constructor: ") + cons_name);
		return 0;
	}

	wtast::DataTypeCons* cons = *cons_it->second.begin();

	map_cons_datatype_t::iterator dt_it = mConsToDataType.find(cons);

	if(mConsToDataType.end()==dt_it || dt_it->second->GetCons(cons_name)!=cons)
	{
		PostError(std::string("Failed to map data type constructor to data type: ") + cons_name);
		return 0;
	}

	wtast::DataType* dt = dt_it->second;
	return MakeDataTypeLiteral(dt, cons, expr);
}

wtast::DataTypeLiteral* wtparser::ParserHelper::MakeDataTypeLiteral(const std::string & dt_name, const std::string & cons_name, wtast::Expression* expr)
{
	map_name_type_t::iterator dt_it = mTypenameToType.find(dt_name);

	if(mTypenameToType.end()==dt_it)
	{
		PostError(std::string("Data type not found: ") + dt_name);
		return 0;
	}

	if(wtast::AstNode::DATATYPE!=dt_it->second->GetKind())
	{
		PostError(std::string("Data type name expected here: '") + dt_name + "'.");
		return 0;
	}

	wtast::DataType*     dt   = static_cast<wtast::DataType*>(dt_it->second);
	wtast::DataTypeCons* cons = dt->GetCons(cons_name);

	if(!cons)
	{
		PostError(std::string("Data type has no constructor with name '") + cons_name + "'.");
		return 0;
	}

	return MakeDataTypeLiteral(dt, cons, expr);
}

wtast::DataTypeLiteral* wtparser::ParserHelper::MakeDataTypeLiteral(wtast::DataType* dt, wtast::DataTypeCons* cons, wtast::Expression* expr)
{
	// tipus ellenorzese

	bool bad = false;

	if(!expr || 0==expr->GetNumberOfExpressions())
	{
		if(0!=cons->GetNumberOfTypes())
		{
			bad = true;
		}
	}
	else
	{
		wtast::TypeInfo* type_info = expr->GetType();

		if(cons->GetNumberOfTypes()!=type_info->GetNumberOfTypes())
		{
			bad = true;
		}
		else
		{
			for(std::size_t i=0,n=cons->GetNumberOfTypes();!bad && i<n;++i)
			{
				bad = !IsSameType(cons->GetType(i), type_info->GetType(i));
			}
		}
	}

	if(bad)
	{
		std::ostringstream ss;
		ss << "Type of the expression and argument type of the '";
		ss << dt->GetName() << "::" << cons->GetName();
		ss << "' constructor mismatch.";
		PostError(ss.str());
		return 0;
	}

	// osszeallitas, es visszateres

	return new wtast::DataTypeLiteral(dt, cons, expr);
}

wtast::SwitchStatement* wtparser::ParserHelper::MakeSwitchStatement(wtast::Expression* cond, switchcases_t* cases, wtast::DefaultStatement* defstmt)
{
	if(!cond)
	{
		PostError("Invalid switch condition.");
		delete cases;
		delete defstmt;
		return 0;
	}

	if(cond->GetType()->IsSimultaneous())
	{
		PostError("Argument of the switch statement cannot be simultaneous.");
		delete cases;
		delete defstmt;
		return 0;
	}

	wtast::Type* const         cond_type = cond->GetType()->GetType(0);
	const wtast::AstNode::Kind case_kind = DecodeCaseKind(cond_type);

	if(wtast::AstNode::UNKNOWN==case_kind)
	{
		PostError("Argument of the switch statement must be an enum or a data type.");
		delete cases;
		delete defstmt;
		return 0;
	}

	std::set<std::string> remaining;
	std::string element_str;
	std::string Element_str;
	std::string part_of_this_str;

	wtast::DataType* dt = 0;
	wtast::EnumType* et = 0;

	if(wtast::AstNode::DATATYPE==cond_type->GetKind())
	{
		dt = static_cast<wtast::DataType*>(cond_type);

		element_str = "data constructor";
		Element_str = "Data constructor";
		part_of_this_str = std::string("data type '") + dt->GetName() + "'.";
	
		for(std::size_t i=0,n=dt->GetNumberOfConses();i<n;++i)
		{
			remaining.insert(dt->GetCons(i)->GetName());
		}
	}
	else
	if(wtast::AstNode::ENUMTYPE==cond_type->GetKind())
	{
		et = static_cast<wtast::EnumType*>(cond_type);

		element_str = "enumerator";
		Element_str = "Enumerator";
		part_of_this_str = std::string("enum '") + et->GetName() + "'.";

		for(std::size_t i=0,n=et->GetNumberOfEnumerators();i<n;++i)
		{
			remaining.insert(et->GetEnumerator(i)->GetName());
		}
	}
	else
	{
		PostError("Argument of the switch statement must be an enum or a data type.");
		delete cases;
		delete defstmt;
		return 0;
	}

	const std::set<std::string> all_names = remaining;

	wtast::SwitchStatement* sw = new wtast::SwitchStatement(cond);

	if(cases)
	{
		for(switchcases_t::iterator it=cases->begin(),end=cases->end();it!=end;++it)
		{
			wtast::SwitchCase* sc = *it;

			if(case_kind!=sc->GetKind())
			{
				PostError("Invalid case kind. Enum and data type cases cannot be mixed.");
				continue;
			}

			const std::string & name = DecodeCaseName(sc);

			if(0==all_names.count(name))
			{
				PostError(Element_str + " '" + name + "' is not part of " + part_of_this_str);
			}
			else
			if(0==remaining.erase(name))
			{
				PostError(Element_str + " '" + name + "' is already used.");
				delete sc;
			}
			else
			{
				sw->AddCase(sc);
			}
		}

		delete cases;
	}

	if(defstmt)
	{
		sw->AddCase(defstmt);
	}
	else
	if(!remaining.empty())
	{
		PostError(std::string("Not all ") + element_str + " were used.");
	}

	return sw;
}

wtast::AstNode::Kind wtparser::ParserHelper::DecodeCaseKind(wtast::Type* type)
{
	switch(type->GetKind())
	{
	case wtast::AstNode::DATATYPE:
		return wtast::AstNode::DATATYPECASESTATEMENT;

	case wtast::AstNode::ENUMTYPE:
		return wtast::AstNode::ENUMCASESTATEMENT;

	default:
		return wtast::AstNode::UNKNOWN;
	}
}

const std::string & wtparser::ParserHelper::DecodeCaseName(wtast::SwitchCase* sc)
{
	static const std::string empty = "";

	switch(sc->GetKind())
	{
	case wtast::AstNode::DATATYPECASESTATEMENT:
		return static_cast<wtast::DataTypeCaseStatement*>(sc)->GetCons()->GetName();

	case wtast::AstNode::ENUMCASESTATEMENT:
		return static_cast<wtast::EnumCaseStatement*>(sc)->GetEnumerator()->GetName();

	default:
		PostError("Internal error: invalid switch case found");
		return empty;
	}
}

variabledecls_t* wtparser::ParserHelper::AddCaseVariables(const std::string & cons_name, identlist_t* ids)
{
	map_name_datatypeconses_t::iterator cons_it = mNameToDataTypeConses.find(cons_name);

	if(mNameToDataTypeConses.end()==cons_it)
	{
		PostError(std::string("Data type constructor not found: ") + cons_name);
		return 0;
	}

	if(1!=cons_it->second.size())
	{
		PostError(std::string("Ambiguous data type constructor: ") + cons_name);
		return 0;
	}

	wtast::DataTypeCons* cons = *cons_it->second.begin();

	map_cons_datatype_t::iterator dt_it = mConsToDataType.find(cons);

	if(mConsToDataType.end()==dt_it || dt_it->second->GetCons(cons_name)!=cons)
	{
		PostError(std::string("Failed to map data type constructor to data type: ") + cons_name);
		return 0;
	}

	wtast::DataType* dt = dt_it->second;
	return AddCaseVariables(dt, cons, ids);
}

variabledecls_t* wtparser::ParserHelper::AddCaseVariables(const std::string & dt_name, const std::string & cons_name, identlist_t* ids)
{
	map_name_type_t::iterator it = mTypenameToType.find(dt_name);

	if(mTypenameToType.end()==it)
	{
		PostError(std::string("Data type not found: ") + dt_name);
		return 0;
	}

	if(wtast::AstNode::DATATYPE!=it->second->GetKind())
	{
		PostError(std::string("Data type name expected here: '") + dt_name + "'.");
		return 0;
	}

	wtast::DataType*     dt   = static_cast<wtast::DataType*>(it->second);
	wtast::DataTypeCons* cons = dt->GetCons(cons_name);

	if(!cons)
	{
		PostError(std::string("Data type has no constructor with name '") + cons_name + "'.");
		return 0;
	}

	return AddCaseVariables(dt, cons, ids);
}

variabledecls_t* wtparser::ParserHelper::AddCaseVariables(wtast::DataType* dt, wtast::DataTypeCons* cons, identlist_t* ids)
{
	bool bad = false;
	variabledecls_t vars;

	if(0==cons->GetNumberOfTypes())
	{
		if(ids && !ids->empty())
		{
			bad = true;
		}
	}
	else
	{
		if(!ids || cons->GetNumberOfTypes()!=ids->size())
		{
			bad = true;
		}
		else
		{
			wtast::AstNode::Variable::Class varclass = wtast::AstNode::Variable::READONLY;

			if(!mSwitchExpressionStack.back())
			{
				PostError("Badly formatted switch statement.");
				bad = true;
			}
			else
			{
				if(mSwitchExpressionStack.back()->OriginateFrom(wtast::AstNode::Variable::FORMAL))
				{
					varclass = wtast::AstNode::Variable::FORMAL;
				}

				for(std::size_t i=0,n=cons->GetNumberOfTypes();!bad && i<n;++i)
				{
					wtast::Type*        type      = cons->GetType(i);
					const std::string & id        = (*ids)[i];
					const UniqueId      unique_id = CreateUniqueId();

					if(mSymTab.AddVariable(type, id, unique_id, _GetEHandlerMode(), varclass))
					{
						wtast::VariableDeclaration* decl = new wtast::VariableDeclaration(type->Copy());
						decl->AddFragment(new wtast::VariableDeclFragment(decl, id, unique_id));
						vars.push_back(decl);
					}
					else
					{
						bad = true;
					}
				}
			}
		}
	}

	if(bad)
	{
		while(!vars.empty())
		{
			delete vars.back();
			vars.pop_back();
		}

		PostError(std::string("Number of arguments and number of types of constructor '") + dt->GetName() + "::" + cons->GetName() + "' mismatch.");
		return 0;
	}

	return new variabledecls_t(vars);
}

wtast::DataTypeCaseStatement* wtparser::ParserHelper::MakeDataTypeCaseStatement(const std::string & cons_name, variabledecls_t* vars, wtast::Compound* comp)
{
	map_name_datatypeconses_t::iterator cons_it = mNameToDataTypeConses.find(cons_name);

	if(mNameToDataTypeConses.end()==cons_it)
	{
		PostError(std::string("Data type constructor not found: ") + cons_name);
		return 0;
	}

	if(1!=cons_it->second.size())
	{
		PostError(std::string("Ambiguous data type constructor: ") + cons_name);
		return 0;
	}

	wtast::DataTypeCons* cons = *cons_it->second.begin();

	map_cons_datatype_t::iterator dt_it = mConsToDataType.find(cons);

	if(mConsToDataType.end()==dt_it || dt_it->second->GetCons(cons_name)!=cons)
	{
		PostError(std::string("Failed to map data type constructor to data type: ") + cons_name);
		return 0;
	}

	wtast::DataType* dt = dt_it->second;
	return MakeDataTypeCaseStatement(dt, cons, vars, comp);
}

wtast::DataTypeCaseStatement* wtparser::ParserHelper::MakeDataTypeCaseStatement(const std::string & dt_name, const std::string & cons_name, variabledecls_t* vars, wtast::Compound* comp)
{
	map_name_type_t::iterator it = mTypenameToType.find(dt_name);

	if(mTypenameToType.end()==it)
	{
		PostError(std::string("Data type not found: ") + dt_name);
		return 0;
	}

	if(wtast::AstNode::DATATYPE!=it->second->GetKind())
	{
		PostError(std::string("Data type name expected here: '") + dt_name + "'.");
		return 0;
	}

	wtast::DataType*     dt   = static_cast<wtast::DataType*>(it->second);
	wtast::DataTypeCons* cons = dt->GetCons(cons_name);

	if(!cons)
	{
		PostError(std::string("Data type has no constructor with name '") + cons_name + "'.");
		return 0;
	}

	return MakeDataTypeCaseStatement(dt, cons, vars, comp);
}

wtast::DataTypeCaseStatement* wtparser::ParserHelper::MakeDataTypeCaseStatement(wtast::DataType* dt, wtast::DataTypeCons* cons, variabledecls_t* vars, wtast::Compound* comp)
{
	wtast::DataTypeCaseStatement* dtcs = new wtast::DataTypeCaseStatement(cons, comp);

	if(vars)
	{
		dtcs->AddVariables(vars->begin(), vars->end());
	}

	return dtcs;
}

wtast::SwitchCase* wtparser::ParserHelper::MakeSpecialCaseStatement(const std::string & special_name, wtast::Compound* comp)
{
	map_name_enumerators_t::iterator    enumerator_it = mNameToEnumerators.find(special_name);
	map_name_datatypeconses_t::iterator dtcons_it     = mNameToDataTypeConses.find(special_name);

	const bool is_enum = mNameToEnumerators.end()!=enumerator_it;
	const bool is_data = mNameToDataTypeConses.end()!=dtcons_it;

	if(!is_enum && !is_data)
	{
		PostError(std::string("Identifier '") + special_name + "' is not an enumerator nor a data type constructor");
		return 0;
	}

	if((is_enum && is_data) || (is_enum && 1!=enumerator_it->second.size()) || (is_data && 1!=dtcons_it->second.size()))
	{
		PostError(std::string("Ambiguous enumerator or data type constructor: ") + special_name);
		return 0;
	}

	if(is_enum)
	{
		wtast::Enumerator* enumerator = *enumerator_it->second.begin();

		map_enumerator_enumtype_t::iterator enum_it = mEnumeratorToEnumType.find(enumerator);

		if(mEnumeratorToEnumType.end()==enum_it || enum_it->second->GetEnumerator(special_name)!=enumerator)
		{
			PostError(std::string("Failed to map enumerator to enum: ") + special_name);
			return 0;
		}

		return new wtast::EnumCaseStatement(enumerator, comp);
	}
	else
	if(is_data)
	{
		wtast::DataTypeCons* cons = *dtcons_it->second.begin();

		map_cons_datatype_t::iterator datatype_it = mConsToDataType.find(cons);

		if(mConsToDataType.end()==datatype_it || datatype_it->second->GetCons(special_name)!=cons)
		{
			PostError(std::string("Failed to map constructor to data type: ") + special_name);
			return 0;
		}

		if(0!=cons->GetNumberOfTypes())
		{
			PostError(std::string("Number of arguments and number of types of constructor '") + datatype_it->second->GetName() + "::" + cons->GetName() + "' mismatch.");
			return 0;
		}

		return new wtast::DataTypeCaseStatement(cons, comp);
	}

	return 0;
}

wtast::SwitchCase* wtparser::ParserHelper::MakeSpecialCaseStatement(const std::string & type_name, const std::string & special_name, wtast::Compound* comp)
{
	map_name_type_t::iterator it = mTypenameToType.find(type_name);

	if(mTypenameToType.end()==it)
	{
		PostError(std::string("Type '") + type_name + "' is not an enum nor a data type");
		return 0;
	}

	switch(it->second->GetKind())
	{
	case wtast::AstNode::DATATYPE:
		{
			wtast::DataType*     dt   = static_cast<wtast::DataType*>(it->second);
			wtast::DataTypeCons* cons = dt->GetCons(special_name);

			if(!cons)
			{
				PostError(std::string("Data type has no constructor with name '") + special_name + "'.");
				return 0;
			}

			return new wtast::DataTypeCaseStatement(cons, comp);
		}
		break;

	case wtast::AstNode::ENUMTYPE:
		{
			wtast::EnumType*   enum_type  = static_cast<wtast::EnumType*>(it->second);
			wtast::Enumerator* enumerator = enum_type->GetEnumerator(special_name);

			if(!enumerator)
			{
				PostError(std::string("Enum has no enumerator with name '") + special_name + "'.");
				return 0;
			}

			return new wtast::EnumCaseStatement(enumerator, comp);
		}
		break;

	default:
		PostError(std::string("Enum type-name or data type-name expected here: '") + type_name + "'.");
		return 0;
	}
}

wtast::DefaultStatement* wtparser::ParserHelper::MakeDefaultStatement(wtast::Compound* comp)
{
	return new wtast::DefaultStatement(comp);
}

wtast::Expression* wtparser::ParserHelper::MakeSpecialLiteral(const std::string & special_name)
{
	map_name_enumerators_t::iterator    enumerator_it = mNameToEnumerators.find(special_name);
	map_name_datatypeconses_t::iterator dtcons_it     = mNameToDataTypeConses.find(special_name);

	const bool is_enum = mNameToEnumerators.end()!=enumerator_it;
	const bool is_data = mNameToDataTypeConses.end()!=dtcons_it;

	if(!is_enum && !is_data)
	{
		PostError(std::string("Identifier '") + special_name + "' is not an enumerator nor a data type constructor");
		return 0;
	}

	if((is_enum && is_data) || (is_enum && 1!=enumerator_it->second.size()) || (is_data && 1!=dtcons_it->second.size()))
	{
		PostError(std::string("Ambiguous enumerator or data type constructor: ") + special_name);
		return 0;
	}

	if(is_enum)
	{
		wtast::Enumerator* enumerator = *enumerator_it->second.begin();

		map_enumerator_enumtype_t::iterator enum_it = mEnumeratorToEnumType.find(enumerator);

		if(mEnumeratorToEnumType.end()==enum_it || enum_it->second->GetEnumerator(special_name)!=enumerator)
		{
			PostError(std::string("Failed to map enumerator to enum: ") + special_name);
			return 0;
		}

		return new wtast::EnumLiteral(enum_it->second, enumerator);
	}
	else
	if(is_data)
	{
		wtast::DataTypeCons* cons = *dtcons_it->second.begin();

		map_cons_datatype_t::iterator datatype_it = mConsToDataType.find(cons);

		if(mConsToDataType.end()==datatype_it || datatype_it->second->GetCons(special_name)!=cons)
		{
			PostError(std::string("Failed to map constructor to data type: ") + special_name);
			return 0;
		}

		if(0!=cons->GetNumberOfTypes())
		{
			PostError(std::string("Number of arguments and number of types of constructor '") + datatype_it->second->GetName() + "::" + cons->GetName() + "' mismatch.");
			return 0;
		}

		return new wtast::DataTypeLiteral(datatype_it->second, cons, 0);
	}

	return 0;
}

wtast::Expression* wtparser::ParserHelper::MakeSpecialLiteral(const std::string & type_name, const std::string & special_name)
{
	map_name_type_t::iterator it = mTypenameToType.find(type_name);

	if(mTypenameToType.end()==it)
	{
		PostError(std::string("Type not found: ") + type_name);
		return 0;
	}

	switch(it->second->GetKind())
	{
	case wtast::AstNode::DATATYPE:
		{
			wtast::DataType*     dt   = static_cast<wtast::DataType*>(it->second);
			wtast::DataTypeCons* cons = dt->GetCons(special_name);

			if(!cons)
			{
				PostError(std::string("Data type has no constructor with name '") + special_name + "'.");
				return 0;
			}

			return new wtast::DataTypeLiteral(dt, cons, 0);
		}
		break;

	case wtast::AstNode::ENUMTYPE:
		{
			wtast::EnumType*   enum_type  = static_cast<wtast::EnumType*>(it->second);
			wtast::Enumerator* enumerator = enum_type->GetEnumerator(special_name);

			if(!enumerator)
			{
				PostError(std::string("Enum has no enumerator with name '") + special_name + "'.");
				return 0;
			}

			return new wtast::EnumLiteral(enum_type, enumerator);
		}
		break;

	default:
		PostError(std::string("Enum type-name or data type-name expected here: '") + type_name + "'.");
		return 0;
	}
}

bool wtparser::ParserHelper::ExpressionCanBeArgument(wtast::FunctionSignature* sig, wtast::Expression* expr)
{
	if(!expr)
	{
		if(0==sig->GetArgc())
		{
			return true;
		}
		else
		{
			PostError("Type of the argument and the function signature do not match.");
			return false;
		}
	}

	for(std::size_t i=0,n=expr->GetNumberOfExpressions();i!=n;++i)
	{
		wtast::FunctionArgument* arg = sig->GetArg(i);

		// mutable helyre immutable-t nem lehet rakni
		// ezert csak akkor kell ellenorizni, ha azt argument nem mutable

		if(!arg->GetType()->IsMutable())
		{
			continue;
		}

		if(!ExpressionIsMutable(expr->GetExpression(i)))
		{
			PostError("Function argument contains access to an immutable variable.");
			return false;
		}
	}

	return true;
}

bool wtparser::ParserHelper::ExpressionIsMutable(wtast::Expression* expr)
{
	if(expr->IsSimultaneous())
	{
		PostError("Internal error: ExpressionIsMutable() called with simultaneous expression.");
		return false;
	}

	switch(expr->GetKind())
	{
	case wtast::AstNode::VARIABLEACCESS:
		{
			wtast::VariableAccess* varacc = static_cast<wtast::VariableAccess*>(expr);

			switch(varacc->GetVarClass())
			{
			case wtast::AstNode::Variable::READONLY:
				{
					return false;
				}
				break;

			case wtast::AstNode::Variable::FORMAL:
				if(!varacc->GetType()->GetType(0)->IsMutable())
				{
					return false;
				}
				break;

			default:
				break;
			}
		}
		break;

	case wtast::AstNode::MEMBERACCESS:
		{
			wtast::MemberAccess* memacc = static_cast<wtast::MemberAccess*>(expr);

			if(!ExpressionIsMutable(memacc->GetLValue()))
			{
				return false;
			}
		}
		break;

	case wtast::AstNode::ELEMENTACCESS:
		{
			wtast::ElementAccess* elemacc = static_cast<wtast::ElementAccess*>(expr);

			if(!ExpressionIsMutable(elemacc->GetLValue()))
			{
				return false;
			}
		}
		break;

	default:
		break;
	}

	return true;
}

bool wtparser::ParserHelper::FunctionHasPfnInArgs(wtast::FunctionSignature* sig)
{
	if(sig)
	{
		for(std::size_t i=0,n=sig->GetArgc();i<n;++i)
		{
			if(sig->GetArg(i)->GetType()->HasFunctionType())
			{
				return true;
			}
		}
	}

	return false;
}

bool wtparser::ParserHelper::ExpressionIsCopyable(wtast::Expression* expr)
{
	bool copyable = true;

	for(std::size_t i=0,n=expr->GetNumberOfExpressions();i!=n;++i)
	{
		wtast::Expression* subexpr = expr->GetExpression(i);

		if(!subexpr->IsTemporary())
		{
			wtast::TypeInfo* etype = subexpr->GetType();

			for(std::size_t j=0,m=etype->GetNumberOfTypes();j!=m;++j)
			{
				wtast::Type* type = etype->GetType(j);

				if(!type->IsCopyable())
				{
					std::ostringstream ss;
					ss << "The type '";
					ss << type->GetName();
					ss << "' is not copyable.";
					PostError(ss.str());
					copyable = false;
				}
			}
		}
	}

	return copyable;
}

void wtparser::ParserHelper::ValidateAssignment(wtast::Assignment* as)
{
	const bool arg_has_pfn = FunctionHasPfnInArgs(_GetCurrentFunction());

	if(!as->GetExpression())
	{
		PostError("Bad right-hand value.");
		return;
	}

	wtast::TypeInfo* typeinfo = as->GetExpression()->GetType();

	if(!typeinfo || as->GetNumberOfLValues()!=typeinfo->GetNumberOfTypes())
	{
		PostError("Number of lvalues in assignment and number of right-hand values mismatch.");
		return;
	}

	if(!ExpressionIsCopyable(as->GetExpression()))
	{
		PostError("Right-hand side cannot be stored, because contains uncopyable elements.");
		return;
	}

	for(std::size_t i=0,n=as->GetNumberOfLValues();i!=n;++i)
	{
		wtast::LValue* lvalue = as->GetLValue(i);

		if(!ExpressionIsMutable(lvalue))
		{
			PostError("Assignment contains an assignment of an immutable variable.");
			return;
		}

		if(lvalue->IsSimultaneous() || lvalue->GetType()->IsSimultaneous())
		{
			PostError("Internal error: lvalue can not be simultaneous.");
			return;
		}

		if(!IsSameType(static_cast<wtast::Type*>(lvalue->GetType()), typeinfo->GetType(i)))
		{
			PostError("Incompatible types in assignment.");
			return;
		}

		if(arg_has_pfn && static_cast<wtast::Type*>(lvalue->GetType())->HasFunctionType())
		{
			if(as->GetExpression()->OriginateFrom(wtast::AstNode::Variable::FORMAL))
			{
				if(lvalue->OriginateFrom(wtast::AstNode::Variable::GLOBAL)
					|| lvalue->OriginateFrom(wtast::AstNode::Variable::LOCAL))
				{
					PostError("Function pointers cannot be assigned when a formal parameter is involved.");
					return;
				}
			}
		}
	}
}

bool wtparser::ParserHelper::ValidateInitialValue(wtast::VariableDeclaration* vardecl, wtast::Expression* init)
{
	if(!init)
	{
		PostError("Wrong initial value.");
		return false;
	}

	wtast::TypeInfo* typeinfo = init->GetType();

	if(!typeinfo)
	{
		PostError("Cannot determine the type of the initial value.");
		return false;
	}

	if(vardecl->GetNumberOfFragments()!=typeinfo->GetNumberOfTypes())
	{
		PostError("Number of the fragments and the arity of the initial value mismatch.");
		return false;
	}

	if(!ExpressionIsCopyable(init))
	{
		PostError("Initializer contains uncopyable elements.");
		return false;
	}

	std::set<UniqueId> uids;
	for(std::size_t i=0,n=vardecl->GetNumberOfFragments();i!=n;++i)
	{
		uids.insert(vardecl->GetFragment(i)->GetUniqueId());
	}

	if(ExpressionHasUniqueId(init, uids))
	{
		PostError("Initializer contains self reference.");
		return false;
	}

	const wt_u32 type_id = vardecl->GetType()->GetTypeId();

	for(std::size_t i=0,n=vardecl->GetNumberOfFragments();i!=n;++i)
	{
		if(type_id!=(typeinfo->GetType(i)->GetTypeId() & ~WTTF_MUTABLE))
		{
			PostError("Incompatible types in assignment.");
			return false;
		}
	}

	if(FunctionHasPfnInArgs(_GetCurrentFunction()) && init->GetType()->HasFunctionType()
		&& init->OriginateFrom(wtast::AstNode::Variable::FORMAL))
	{
		PostError("Function pointers cannot be initialized when a formal parameter is involved.");
		return false;
	}

	return true;
}

bool wtparser::ParserHelper::ValidateUninitializedVariable(wtast::VariableDeclaration* vardecl)
{
	if(!vardecl->GetType()->IsDefaultConstructible())
	{
		std::ostringstream ss;

		ss << "Type '";
		ss << vardecl->GetType()->GetName();
		ss << "' is not default constructible.";

		PostError(ss.str());
		return false;
	}

	return true;
}

std::string wtparser::ParserHelper::ValidateFunctionName(const std::string & funcname, bool raw_string)
{
	if(raw_string && 2>=funcname.length())
	{
		PostError("Function name can not be empty.");
		return "";
	}

	const std::size_t first_index = raw_string ? 1 : 0;

	char first = funcname[1];

	if('_'==first || ('a'<=first && first<='z') || ('A'<=first && first<='Z'))
	{
		for(std::size_t i=first_index,n=funcname.length()-first_index;i<n;++i)
		{
			char ch = funcname[i];

			if(!('_'==ch || ('a'<=ch && ch<='z') || ('A'<=ch && ch<='Z') || ('0'<=ch && ch<='9')))
			{
				PostError("Function name contains one or more invalid characters.");
				return "";
			}
		}
	}
	else
	{
		for(std::size_t i=first_index,n=funcname.length()-first_index;i<n;++i)
		{
			if(std::string::npos==OperatorChars.find(funcname[i]))
			{
				PostError("Function name contains one or more invalid characters.");
				return "";
			}
		}
	}

	if(raw_string)
	{
		return funcname.substr(1, funcname.length()-2);
	}

	return funcname;
}

void wtparser::ParserHelper::PushSwitchExpression(wtast::Expression* expr)
{
	mSwitchExpressionStack.push_back(expr);
}

void wtparser::ParserHelper::PopSwitchExpression()
{
	mSwitchExpressionStack.pop_back();
}

wtast::Expression* wtparser::ParserHelper::ValidateExpression(wtast::Expression* expr)
{
	if(expr && !expr->GetType())
	{
		PostError("Expression has no value.");
		return 0;
	}

	return expr;
}

static bool is_index_type(const wt_u32 type_id)
{
	switch(type_id)
	{
	case WTT_U8:
	case WTT_I8:
	case WTT_U16:
	case WTT_I16:
	case WTT_U32:
	case WTT_I32:
	case WTT_U64:
	case WTT_I64:
		return true;

	default:
		return false;
	}
}

bool wtparser::ParserHelper::CanBeIndexed(wtast::TypeInfo* info, wtast::TypeInfo* eti)
{
	if(!info || info->IsSimultaneous() || !eti || eti->IsSimultaneous())
	{
		return false;
	}

	// csak string-et es seq<T>-t lehet indexelni, es csak u32-vel
	// TODO: esetleg kesobb lehetne u8-cal es u16-tal is indexelni

	wt_u32 type_id = (static_cast<wtast::Type*>(info)->GetTypeId() & ~WTTF_MUTABLE);

	return
		(WTT_STRING==type_id
		|| static_cast<wtast::Type*>(info)->IsSequence()) &&
		is_index_type(static_cast<wtast::Type*>(eti)->GetTypeId());
}

namespace
{

class FindUniqueIdVisitor : public wtast::AstVisitor
{
	const std::set<wtparser::UniqueId> & uids;
	bool                                 has;

public:
	FindUniqueIdVisitor(const std::set<wtparser::UniqueId> & uids) : uids(uids), has(false) { }

	virtual bool Visit(wtast::VariableAccess* va)
	{
		if(uids.count(va->GetUniqueId()))
		{
			has = true;
		}

		return false;
	}

	bool found() const
	{
		return has;
	}
};

} // namespace

bool wtparser::ParserHelper::ExpressionHasUniqueId(wtast::Expression* expr, const std::set<UniqueId> & uids)
{
	FindUniqueIdVisitor vis(uids);
	expr->Accept(&vis);
	return vis.found();
}
