#include "config_private.h"
#include "processor.h"
#include "types.h"
#include "vm.h"
#include "program.h"
#include "ctx.h"
#include "stdlib.h"
#include "microprog.h"
#include "objects.h"
#include "function.h"
#include "funcsig.h"
#include "debug.h"
#include <string.h>
#if WTRUNTIME_DEBUG
#  include <assert.h>
#  include <stdio.h>
#endif /* WTRUNTIME_DEBUG */

#if WTRUNTIME_DEBUG
#  define goto__out_of_memory_exception() do { trap_out_of_memory(); goto out_of_memory_exception; } while(0)
void trap_out_of_memory() { }
#else /* WTRUNTIME_DEBUG */
#  define goto__out_of_memory_exception() goto out_of_memory_exception
#endif /* WTRUNTIME_DEBUG */

#define CHECK_PUSH(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SPACE(_ctx_)>=(_size_)) \
		{ \
			/* skip */ \
		} \
		else \
		{ \
			goto stack_overflow_exception; \
		} \
	} \
	while(0)

#define CHECK_POP(_ctx_,_size_) \
	do \
	{ \
		if(WTR_CTX_DS_SIZE(_ctx_)>=(_size_)) \
		{ \
			/* skip */ \
		} \
		else \
		{ \
			goto stack_underflow_exception; \
		} \
	} \
	while(0)

#define READ_LONGIMM \
	do \
	{ \
		wt_u16 low; \
		if(0!=wtr_vma_read(vm, 2*(ctx->frame.pc++), &low, sizeof(low))) \
		{ \
			goto memory_access_violation_exception; \
		} \
		opdata = (opdata << WT_TOMB_INSTRBITS) | low; \
	} \
	while(0)

#define READ_STACKIMM \
	do \
	{ \
		WTR_CTX_DS_POP(ctx, wt_u32, opdata); \
	} \
	while(0)

#ifndef RotateRight8
#  define RotateRight8(value,shift) (((value)>>(shift)) | ((value)<<(8-(shift))))
#endif /* RotateRight8 */

#ifndef RotateRight16
#  define RotateRight16(value,shift) (((value)>>(shift)) | ((value)<<(16-(shift))))
#endif /* RotateRight16 */

#ifndef RotateRight32
#  define RotateRight32(value,shift) (((value)>>(shift)) | ((value)<<(32-(shift))))
#endif /* RotateRight32 */

#ifndef RotateRight64
#  define RotateRight64(value,shift) (((value)>>(shift)) | ((value)<<(64-(shift))))
#endif /* RotateRight64 */

#define __OP_FUNC_ROR(_lhs_,_rhs_)    (RotateRight32(_lhs_,_rhs_))
#define __OP_FUNC_RORL(_lhs_,_rhs_)   (RotateRight64(_lhs_,_rhs_))
#if WTRUNTIME_HAVE_F_MATH
#define __OP_FUNC_FMOD(_lhs_,_rhs_)   (fmodf((_lhs_), (_rhs_)))
#else /* WTRUNTIME_HAVE_F_MATH */
#define __OP_FUNC_FMOD(_lhs_,_rhs_)   ((wt_f32)fmod((_lhs_), (_rhs_)))
#endif /* WTRUNTIME_HAVE_F_MATH */
#define __OP_FUNC_ADDB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) + ((wt_u8)(_rhs_))))
#define __OP_FUNC_SUBB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) - ((wt_u8)(_rhs_))))
#define __OP_FUNC_MULB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) * ((wt_u8)(_rhs_))))
#define __OP_FUNC_SMULB(_lhs_,_rhs_)  ((wt_u8)(((wt_i8)(wt_u8)(_lhs_)) * ((wt_i8)(wt_u8)(_rhs_))))
#define __OP_FUNC_DIVB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) / ((wt_u8)(_rhs_))))
#define __OP_FUNC_SDIVB(_lhs_,_rhs_)  ((wt_u8)(((wt_i8)(wt_u8)(_lhs_)) / ((wt_i8)(wt_u8)(_rhs_))))
#define __OP_FUNC_MODB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) % ((wt_u8)(_rhs_))))
#define __OP_FUNC_SMODB(_lhs_,_rhs_)  ((wt_u8)(((wt_i8)(wt_u8)(_lhs_)) % ((wt_i8)(wt_u8)(_rhs_))))
#define __OP_FUNC_SHLB(_lhs_,_rhs_)   ((wt_u8)(((wt_u8)(_lhs_)) << ((wt_u8)(_rhs_))))
#define __OP_FUNC_RORB(_lhs_,_rhs_)   ((wt_u8)(RotateRight8((wt_u8)(_lhs_), (wt_u8)(_rhs_))))
#define __OP_FUNC_ADDH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) + ((wt_u16)(_rhs_))))
#define __OP_FUNC_SUBH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) - ((wt_u16)(_rhs_))))
#define __OP_FUNC_MULH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) * ((wt_u16)(_rhs_))))
#define __OP_FUNC_SMULH(_lhs_,_rhs_)  ((wt_u16)(((wt_i16)(wt_u16)(_lhs_)) * ((wt_i16)(wt_u16)(_rhs_))))
#define __OP_FUNC_DIVH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) / ((wt_u16)(_rhs_))))
#define __OP_FUNC_SDIVH(_lhs_,_rhs_)  ((wt_u16)(((wt_i16)(wt_u16)(_lhs_)) / ((wt_i16)(wt_u16)(_rhs_))))
#define __OP_FUNC_MODH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) % ((wt_u16)(_rhs_))))
#define __OP_FUNC_SMODH(_lhs_,_rhs_)  ((wt_u16)(((wt_i16)(wt_u16)(_lhs_)) % ((wt_i16)(wt_u16)(_rhs_))))
#define __OP_FUNC_SHLH(_lhs_,_rhs_)   ((wt_u16)(((wt_u16)(_lhs_)) << ((wt_u16)(_rhs_))))
#define __OP_FUNC_RORH(_lhs_,_rhs_)   ((wt_u16)(RotateRight16((wt_u16)(_lhs_), (wt_u16)(_rhs_))))
#define __OP_FUNC_CMP(_lhs_,_rhs_)    (((_lhs_)<(_rhs_))?(-1):(((_lhs_)>(_rhs_)?1:0)))

static enum wtr_runresult wtr_processor_op_OSF(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	struct wtr_program* const prog = ctx->frame.prog;

	if(WTRCF_CSOS & ctx->frame.flags)
	{
		wt_vmptr csos = ctx->frame.csos;
		wt_vmptr vma  = csos + ctx->frame.csos_sp;
		wt_u32   cnt;
		wt_u32   size;

		while(vma>csos)
		{
			struct wtr_dynstack* stk = &ctx->rtos;
			struct wtr_objentry* top;

			/* the 'vma' points right after the last real element */

			vma -= sizeof(struct wtr_objentry);

			/* add entry into TOS or RTOS */
			/* (add to TOS only if it is empty) */

			if(ctx->tos.sp!=ctx->frame.tos_mark)
			{
				/* TOS is unlikely empty, even if the CSOS has more than one element */
				/* take this empty branch as a branch-prediction optimization */
			}
			else
			{
				stk = &ctx->tos;
			}

			if(NULL==(top=wtr_dynstack_push_v(stk, vma)))
			{
				/* TODO: if the push fails, then it cause a memory leak */
				break;
			}

			/* the 'ent.obj' is a pointer into the DVT */

			if(0!=wtr_vma_read(vm, top->obj, &top->obj, sizeof(wt_vmptr)))
			{
				/* TODO: is this correct? */
				wtr_dynstack_pop(stk);
			}
		}

		WTR_CTX_CS_POP(ctx, wt_u32, size);

		ctx->csp          += size;
		ctx->frame.flags  &= ~WTRCF_CSOS;
		ctx->frame.csos    = 0;
		ctx->frame.csos_sp = 0;

		WTR_CTX_CS_POP(ctx, wt_u32, size);
	}

	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_op_DVTF(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	((void)vm);

	if(0==(WTRCF_CSOS & ctx->frame.flags))
	{
		/* this is the more likely branch */
	}
	else
	{
		ctx->last_exception = BadProgramAccessException;
		return WTRR_RAISE;
	}

	if(WTRCF_DVT & ctx->frame.flags)
	{
		wt_u32 dvt_size;
		WTR_CTX_CS_POP(ctx, wt_u32, dvt_size);

		ctx->csp         += dvt_size;
		ctx->frame.flags &= ~WTRCF_DVT;
		ctx->frame.dvt    = 0;
	}

	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_op_LEAVE(struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	if(WTRCF_CSOS & ctx->frame.flags)
	{
		// TODO: handle return value?
		wtr_processor_op_OSF(vm, ctx);
	}

	if(WTRCF_DVT & ctx->frame.flags)
	{
		// TODO: handle return value?
		wtr_processor_op_DVTF(vm, ctx);
	}

	if(WTRCF_START & ctx->frame.flags)
	{
		WTR_PFN_RETURNACTION return_action = ctx->return_action;

		if(NULL!=return_action)
		{
			ctx->return_action = NULL;
			return_action(ctx);
		}

		return WTRR_RETURNED;
	}

	wtr_vma_read(vm, ctx->csp, &ctx->frame, sizeof(ctx->frame));
	ctx->csp += sizeof(ctx->frame);

	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_unwind_frames(
	struct wtr_vm* const vm, struct wtr_context* const ctx)
{
	enum wtr_runresult rr = WTRR_RUN;

	WTR_CTX_DS_PEEK(ctx, wt_exception, ctx->last_exception);

	while(WTRR_RUN==rr && WTR_NULLPC==ctx->frame.ehandler)
	{
		rr = wtr_processor_op_LEAVE(vm, ctx);
	}

	if(WTRR_RETURNED==rr)
	{
		return WTRR_EXCEPTION;
	}

	if(WTRR_RUN!=rr)
	{
		return rr;
	}

	ctx->dsp            = ctx->frame.dsp_mark;
	ctx->frame.pc       = ctx->frame.ehandler;
	ctx->frame.ehandler = WTR_NULLPC;

	WTR_CTX_DS_PUSH(ctx, wt_exception, ctx->last_exception);
	return WTRR_RUN;
}

static enum wtr_runresult wtr_processor_op_CFP(struct wtr_vm* const vm, struct wtr_context* const ctx, struct wt_pfn* pfn)
{
	if(pfn->ext==vm->external_tag)
	{
		struct wtr_cfunction* const cfunc = (struct wtr_cfunction*)pfn->ptr;
		const wt_u32                argc  = cfunc->funcsig->argc;

		if(!wtr_ctx_resize_value_array(ctx, argc))
		{
			ctx->last_exception = OutOfMemoryException;
			return WTRR_RAISE;
		}

		/* build the extcallinfo */

		ctx->extcallinfo.vm    = vm;
		ctx->extcallinfo.func  = cfunc->extfunc_pfn;
		ctx->extcallinfo.token = cfunc->token;
		ctx->extcallinfo.argc  = argc;
		ctx->extcallinfo.retc  = cfunc->funcsig->retc;
		ctx->extcallinfo.retv  = cfunc->funcsig->retv;

		/* pop function arguments */

		if(0!=(ctx->last_exception=wtr_ctx_pop_values(ctx, argc,
			ctx->value_array, cfunc->funcsig->argv)))
		{
			return WTRR_RAISE;
		}

		ctx->run = WTRR_EXTCALL;

		// TODO: only if context allows it
		if(1)
		{
			ctx->extcallinfo.func(ctx, &ctx->extcallinfo, ctx->value_array);
		}

		return ctx->run;
	}
	else
	{
		/* is there enough space to store the current frame? */
		if(WTR_CTX_CS_SPACE(ctx)>=sizeof(ctx->frame))
		{
			/* store the current frame */
			WTR_CTX_CS_PUSH(ctx, struct wtr_callframe, ctx->frame);

			/* assemble new frame */
			ctx->frame.pc       = pfn->ext;
			ctx->frame.flags    = 0;
			ctx->frame.ehandler = WTR_NULLPC;
			ctx->frame.dvt      = 0;
			ctx->frame.csos     = 0;
			ctx->frame.csos_sp  = 0;
			ctx->frame.tos_mark = ctx->tos.sp;
			ctx->frame.dsp_mark = 0;
			ctx->frame.prog     = (struct wtr_program*)pfn->ptr;
		}
		else
		{
			ctx->last_exception = StackOverflowException;
			return WTRR_RAISE;
		}
	}

	return WTRR_RUN;
}

enum wtr_runresult wtr_processor_tomb_run(struct wtr_vm* const vm,
	struct wtr_context* const ctx, wt_u32 quantum)
{
	/* exception ID buffer for 'raise_exception' */
	wt_exception exc;

	if(WTRR_RAISE==ctx->run)
	{
		exc = ctx->last_exception;
		goto raise_exception;
	}

	while(WTRR_RUN==ctx->run)
	{
		wt_u16 instr;
		wt_u32 opdata;

		wtr_tomb_instructionset opcode;

		if(0==(quantum--))
		{
			ctx->run = WTRR_QUANTUM;
			return WTRR_QUANTUM;
		}

		if(0!=wtr_vma_read(vm, 2*(ctx->frame.pc++), &instr, sizeof(instr)))
		{
			goto memory_access_violation_exception;
		}

		opdata = WT_TOMB_GETOPDATA(instr);
		opcode = (wtr_tomb_instructionset)WT_TOMB_GETOPCODE(instr);

#ifdef DISPLAY_EXECUTED_INSTRUCTIONS
		printf("%08X/%08X %02X %02X %04X\n", (ctx->frame.pc-ctx->frame.prog->pc_base-1)*2,
			(ctx->frame.pc-ctx->frame.prog->pc_base-1)*2, WT_TOMB_GETOPCODE(instr), opdata, code[ctx->frame.pc]);
#endif /* DISPLAY_EXECUTED_INSTRUCTIONS */

#ifdef DISPLAY_DATASTACK_AFTER_INSTRUCTIONS
		{
			wt_u32  i = 0;
			wt_u32* p = (wt_u32*)ctx->dsp;

			for(i=0;p<(wt_u32*)ctx->ds_hi;++i)
			{
				if(0==(i%8))
				{
					if(i)
					{
						printf("\n");
					}

					printf("   ");
				}

				printf(" %08X", *p++);
			}

			printf("\n\n");
		}
#endif /* DISPLAY_DATASTACK_AFTER_INSTRUCTIONS */

		switch(opcode)
		{
		case WT_TOMB_OP_NOP:
			{
				/* no operation */
			}
			break;

		case WT_TOMB_OP_LD_I: /* Load Immediate */
			{
				wt_vmptr addr;
__ld_i:
				CHECK_POP(ctx, sizeof(addr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);

				if(0==(opdata & 0x3))
				{
					if(WTR_CTX_DS_SPACE(ctx)>=opdata)
					{
						ctx->dsp -= opdata;

						if(0!=wtr_vma_memcpy(vm, ctx->dsp, addr, opdata))
						{
							goto memory_access_violation_exception;
						}
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_LD_L: /* Load LongImmediate */
			{
				READ_LONGIMM;
				goto __ld_i;
			}
			break;

		case WT_TOMB_OP_LD_S: /* Load StackImmediate */
			{
				READ_STACKIMM;
				goto __ld_i;
			}
			break;

		case WT_TOMB_OP_ST_I: /* Store Immediate */
			{
				wt_vmptr addr;
__st_i:
				CHECK_POP(ctx, sizeof(addr)+opdata);
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);

				if(0==(0x3 & opdata))
				{
					if(0!=wtr_vma_memcpy(vm, addr, ctx->dsp, opdata))
					{
						goto memory_access_violation_exception;
					}

					ctx->dsp += opdata;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_ST_L: /* Store LongImmediate */
			{
				READ_LONGIMM;
				goto __st_i;
			}
			break;

		case WT_TOMB_OP_ST_S: /* Store StackImmediate */
			{
				READ_STACKIMM;
				goto __st_i;
			}
			break;

		case WT_TOMB_OP_IDRD_I: /* Internal DataStack Dereference Immediate */
			{
__idrd_i:
				if(0==(0x3 & opdata) && WTR_CTX_DS_SIZE(ctx)>=opdata)
				{
					wt_vmptr addr = ctx->dsp + opdata;
					wt_vmptr ptr;

					if(0!=wtr_vma_read(vm, addr, &ptr, sizeof(ptr)))
					{
						goto memory_access_violation_exception;
					}

					if(0!=wtr_vma_read(vm, ptr, &ptr, sizeof(ptr)))
					{
						goto memory_access_violation_exception;
					}

					if(0!=wtr_vma_write(vm, addr, &ptr, sizeof(ptr)))
					{
						goto memory_access_violation_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_IDRD_L: /* Internal DataStack Dereference LongImmediate */
			{
				READ_LONGIMM;
				goto __idrd_i;
			}
			break;

		case WT_TOMB_OP_IDRD_S: /* Internal DataStack Dereference StackImmediate */
			{
				READ_STACKIMM;
				goto __idrd_i;
			}
			break;

		case WT_TOMB_OP_PBA_I: /* Push Base Address Immediate */
			{
				wt_vmptr addr;
__pba_i:
				addr = opdata + ctx->frame.prog->base_address;
				CHECK_PUSH(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, addr);
			}
			break;

		case WT_TOMB_OP_PBA_L: /* Push Base Address LongImmediate */
			{
				READ_LONGIMM;
				goto __pba_i;
			}
			break;

		case WT_TOMB_OP_PBA_S: /* Push Base Address StackImmediate */
			{
				READ_STACKIMM;
				goto __pba_i;
			}
			break;

		case WT_TOMB_OP_PVT_I: /* Push Variable Table Immediate */
			{
__pvt_i:
				if(WTRCF_DVT & ctx->frame.flags)
				{
					wt_vmptr addr = opdata + ctx->frame.dvt;
					CHECK_PUSH(ctx, sizeof(wt_vmptr));
					WTR_CTX_DS_PUSH(ctx, wt_vmptr, addr);
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_PVT_L: /* Push Variable Table LongImmediate */
			{
				READ_LONGIMM;
				goto __pvt_i;
			}
			break;

		case WT_TOMB_OP_PVT_S: /* Push Variable Table StackImmediate */
			{
				READ_STACKIMM;
				goto __pvt_i;
			}
			break;

		case WT_TOMB_OP_PPC: /* Push Program Counter */
			{
				CHECK_PUSH(ctx, sizeof(ctx->frame.pc));
				WTR_CTX_DS_PUSH(ctx, wt_u32, ctx->frame.pc);
			}
			break;

		case WT_TOMB_OP_DUP_I: /* Duplicate Immediate */
			{
				goto __dup_l;
			}
			break;

		case WT_TOMB_OP_DUP_L: /* Duplicate LongImmediate */
			{
				READ_LONGIMM;
__dup_l:
				if(0==(opdata & 0x3))
				{
					wt_vmptr prev_dsp;

					CHECK_POP(ctx, opdata);
					CHECK_PUSH(ctx, opdata);
					prev_dsp = ctx->dsp;
					ctx->dsp -= opdata;

					if(0!=wtr_vma_memcpy(vm, ctx->dsp, prev_dsp, opdata))
					{
						goto memory_access_violation_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_POP_I: /* Pop Immediate */
			{
				if(0==(0x3 & opdata))
				{
					CHECK_POP(ctx, opdata);
					ctx->dsp += opdata;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_POP_L: /* Pop LongImmediate */
			{
				READ_LONGIMM;
				if(0==(0x3 & opdata))
				{
					CHECK_POP(ctx, opdata);
					ctx->dsp += opdata;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_PWZ_I: /* Push Word Zero Immediate */
			{
__pwz_i:
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_PWZ_L: /* Push Word Zero LongImmediate */
			{
				READ_LONGIMM;
				goto __pwz_i;
			}
			break;

		case WT_TOMB_OP_PWO_I: /* Push Word One Immediate */
			{
				opdata |= ~WT_TOMB_OPDATAMASK;
__pwo_i:
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_PWO_L: /* Push Word One LongImmediate */
			{
				READ_LONGIMM;
				opdata |= ~WT_TOMB_MASKLONGIMM;
				goto __pwo_i;
			}
			break;

		case WT_TOMB_OP_PWZL_I: /* Push LongWord Zero Immediate */
			{
				wt_u64 longword;
__pwzl_i:
				longword = opdata;
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_PWZL_L: /* Push LongWord Zero LongImmediate */
			{
				READ_LONGIMM;
				goto __pwzl_i;
			}
			break;

		case WT_TOMB_OP_PWOL_I: /* Push LongWord One Immediate */
			{
				wt_u64 longword = opdata | (~(wt_u64)WT_TOMB_OPDATAMASK);
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_PWOL_L: /* Push LongWord One LongImmediate */
			{
				wt_u64 longword;
				READ_LONGIMM;
				longword = opdata | (~(wt_u64)WT_TOMB_MASKLONGIMM);
				CHECK_PUSH(ctx, sizeof(wt_u64));
				WTR_CTX_DS_PUSH(ctx, wt_u64, longword);
			}
			break;

		case WT_TOMB_OP_DVTA_I: /* Allocate DVT Immediate */
			{
__dvta_i:
				if((ctx->frame.flags & (WTRCF_DVT|WTRCF_CSOS))==0 && (opdata&0x3)==0)
				{
					if(WTR_CTX_CS_SPACE(ctx)>=sizeof(wt_u32)+opdata)
					{
						ctx->csp -= opdata;
						ctx->frame.dvt = ctx->csp;
						wtr_vma_memset(vm, ctx->csp, 0, opdata); /* clear the DVT */
						WTR_CTX_CS_PUSH(ctx, wt_u32, opdata);
						ctx->frame.flags |= WTRCF_DVT;
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_DVTA_L: /* Allocate DVT LongImmediate */
			{
				READ_LONGIMM;
				goto __dvta_i;
			}
			break;

		case WT_TOMB_OP_DVTA_S: /* Allocate DVT StackImmediate */
			{
				READ_STACKIMM;
				goto __dvta_i;
			}
			break;

		case WT_TOMB_OP_DVTF: /* Free DVT */
			{
				ctx->run = wtr_processor_op_DVTF(vm, ctx);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_OSA_I: /* Allocate CSOS Immediate */
			{
__osa_i:
				if(WTRCF_DVT==(ctx->frame.flags & (WTRCF_DVT|WTRCF_CSOS)))
				{
					const wt_u32 cs_space = WTR_CTX_CS_SPACE(ctx);

					const wt_u32 size = opdata * sizeof(struct wtr_objentry);

					if((cs_space/sizeof(struct wtr_objentry))>opdata &&
						cs_space>=sizeof(wt_u32)*2+size)
					{
						WTR_CTX_CS_PUSH(ctx, wt_u32, size);
						ctx->csp -= size;
						ctx->frame.csos = ctx->csp;
						wtr_vma_memset(vm, ctx->csp, 0, size); /* clear the CSOS */
						WTR_CTX_CS_PUSH(ctx, wt_u32, size);
						ctx->frame.flags |= WTRCF_CSOS;
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_OSA_L: /* Allocate CSOS LongImmediate */
			{
				READ_LONGIMM;
				goto __osa_i;
			}
			break;

		case WT_TOMB_OP_OSA_S: /* Allocate CSOS StackImmediate */
			{
				READ_STACKIMM;
				goto __osa_i;
			}
			break;

		case WT_TOMB_OP_OSF: /* Free CSOS */
			{
				ctx->run = wtr_processor_op_OSF(vm, ctx);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_OSPU: /* CSOS Push */
			{
				if(WTRCF_CSOS & ctx->frame.flags)
				{
					wt_u32 csos_size;

					/* CSOS size is just right below the CSOS start addr */

					if(0!=wtr_vma_read(vm, ctx->frame.csos-sizeof(wt_u32),
						&csos_size, sizeof(csos_size)))
					{
						goto memory_access_violation_exception;
					}

					if(ctx->frame.csos_sp<csos_size)
					{
						const wt_vmptr vma =
							ctx->frame.csos + ctx->frame.csos_sp;

						if(0!=wtr_vma_memcpy(vm, vma, ctx->dsp,
							sizeof(struct wtr_objentry)))
						{
							goto memory_access_violation_exception;
						}

						ctx->frame.csos_sp += sizeof(struct wtr_objentry);
						ctx->dsp += sizeof(struct wtr_objentry);
					}
					else
					{
						goto stack_overflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_OSPO: /* CSOS Pop */
			{
				if(WTRCF_CSOS & ctx->frame.flags)
				{
					if(0<ctx->frame.csos_sp)
					{
						const wt_vmptr vma =
							ctx->frame.csos + ctx->frame.csos_sp -
							sizeof(struct wtr_objentry);

						CHECK_PUSH(ctx, sizeof(struct wtr_objentry));

						/* note: we can pre-decrement the DSP, because if */
						/* read fails, we raise an exception anyway */
						ctx->dsp -= sizeof(struct wtr_objentry);

						if(0!=wtr_vma_memcpy(vm, ctx->dsp, vma,
							sizeof(struct wtr_objentry)))
						{
							goto memory_access_violation_exception;
						}

						ctx->frame.csos_sp -= sizeof(struct wtr_objentry);
					}
					else
					{
						goto stack_underflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_OSSP: /* CSOS SP */
			{
				wt_u32 sp = ctx->frame.csos_sp / sizeof(struct wtr_objentry);
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, sp);
			}
			break;

		case WT_TOMB_OP_TSZ: /* TOS Size */
			{
				wt_u32 sz = (ctx->tos.sp - ctx->frame.tos_mark) /
					sizeof(struct wtr_objentry);
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, sz);
			}
			break;

		case WT_TOMB_OP_TPU: /* TOS Push */
			{
				CHECK_POP(ctx, sizeof(struct wtr_objentry));
				if(NULL!=wtr_dynstack_push_v(&ctx->tos, ctx->dsp))
				{
					ctx->dsp += sizeof(struct wtr_objentry);
				}
				else
				{
					goto stack_overflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_TPO: /* TOS Pop */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					struct wtr_objentry* top = wtr_dynstack_top(&ctx->tos);

					CHECK_PUSH(ctx, sizeof(struct wtr_objentry));
					WTR_CTX_DS_PUSH(ctx, struct wtr_objentry, *top);
					wtr_dynstack_pop(&ctx->tos);

					if(ctx->tos.sp==ctx->frame.tos_mark && ctx->rtos.sp)
					{
_put_one_from_rtos_to_tos:
						top = wtr_dynstack_top(&ctx->rtos);
						wtr_dynstack_push_p(&ctx->tos, top);
						wtr_dynstack_pop(&ctx->rtos);

						break;
					}
				}
				else
				{
					goto stack_underflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_TDR: /* TOS Drop */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					wtr_dynstack_pop(&ctx->tos);

					if(ctx->tos.sp==ctx->frame.tos_mark && ctx->rtos.sp)
					{
						goto _put_one_from_rtos_to_tos;
					}
				}
				else
				{
					goto stack_underflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_TCL: /* TOS Clear */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					CHECK_PUSH(ctx, sizeof(wt_u32));
					WTR_CTX_DS_PUSH(ctx, wt_u32, ctx->frame.pc);
					ctx->frame.pc = vm->pc_tcl_ja;
				}
			}
			break;

		case WT_TOMB_OP_PDSP_I: /* Push Data Stack Pointer Immediate */
			{
__pdsp_i:
				CHECK_PUSH(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_PUSH(ctx, wt_vmptr, ctx->dsp+opdata);
			}
			break;

		case WT_TOMB_OP_PDSP_L: /* Push Data Stack Pointer LongImmediate */
			{
				READ_LONGIMM;
				goto __pdsp_i;
			}
			break;

		case WT_TOMB_OP_VAS_I: /* Virtual Address Size Immediate */
			{
__vas_i:
				opdata *= sizeof(wt_vmptr);
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_VAS_L: /* Virtual Address Size LongImmediate */
			{
				READ_LONGIMM;
				goto __vas_i;
			}
			break;

		case WT_TOMB_OP_ENTER_I: /* Enter Immediate */
			{
__enter_i:
				/* is there enough space to store the current frame? */
				if(WTR_CTX_CS_SPACE(ctx)>=sizeof(ctx->frame))
				{
					/* store the current frame */
					WTR_CTX_CS_PUSH(ctx, struct wtr_callframe, ctx->frame);

					/* assemble new frame */
					ctx->frame.pc       = ctx->frame.prog->pc_base + opdata;
					ctx->frame.flags    = 0;
					ctx->frame.ehandler = WTR_NULLPC;
					ctx->frame.dvt      = 0;
					ctx->frame.csos     = 0;
					ctx->frame.csos_sp  = 0;
					ctx->frame.tos_mark = ctx->tos.sp;
					ctx->frame.dsp_mark = 0;
				}
				else
				{
					goto stack_overflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_ENTER_L: /* Enter LongImmediate */
			{
				READ_LONGIMM;
				goto __enter_i;
			}
			break;

		case WT_TOMB_OP_ENTER_S: /* Enter StackImmediate */
			{
				READ_STACKIMM;
				goto __enter_i;
			}
			break;

		case WT_TOMB_OP_LEAVE: /* Leave current frame */
			{
				ctx->run = wtr_processor_op_LEAVE(vm, ctx);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_JMP_I: /* Jump Immediate */
			{
				if(opdata & WT_TOMB_HIBITIMM)
				{
					opdata |= ~WT_TOMB_OPDATAMASK;
				}

				ctx->frame.pc += opdata - 1;
			}
			break;

		case WT_TOMB_OP_JMP_L: /* Jump LongImmediate */
			{
				READ_LONGIMM;
				if(opdata & WT_TOMB_HIBITLONGIMM)
				{
					opdata |= ~WT_TOMB_MASKLONGIMM;
				}

				ctx->frame.pc += opdata - 2;
			}
			break;

		case WT_TOMB_OP_JMP_S: /* Jump StackImmediate */
			{
				READ_STACKIMM;
				ctx->frame.pc += opdata - 1;
			}
			break;

		case WT_TOMB_OP_JT_I: /* JumpTrue Immediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				WTR_CTX_DS_POP(ctx, wt_bool, cond);

				if(cond)
				{
					if(opdata & WT_TOMB_HIBITIMM)
					{
						opdata |= ~WT_TOMB_OPDATAMASK;
					}

					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JT_L: /* JumpTrue LongImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				WTR_CTX_DS_POP(ctx, wt_bool, cond);

				READ_LONGIMM;
				if(cond)
				{
					if(opdata & WT_TOMB_HIBITLONGIMM)
					{
						opdata |= ~WT_TOMB_MASKLONGIMM;
					}

					ctx->frame.pc += opdata - 2;
				}
			}
			break;

		case WT_TOMB_OP_JT_S: /* JumpTrue StackImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				WTR_CTX_DS_POP(ctx, wt_bool, cond);

				READ_STACKIMM;
				if(cond)
				{
					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JF_I: /* JumpFalse Immediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				WTR_CTX_DS_POP(ctx, wt_bool, cond);

				if(!cond)
				{
					if(opdata & WT_TOMB_HIBITIMM)
					{
						opdata |= ~WT_TOMB_OPDATAMASK;
					}

					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JF_L: /* JumpFalse LongImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				WTR_CTX_DS_POP(ctx, wt_bool, cond);

				READ_LONGIMM;
				if(!cond)
				{
					if(opdata & WT_TOMB_HIBITLONGIMM)
					{
						opdata |= ~WT_TOMB_MASKLONGIMM;
					}

					ctx->frame.pc += opdata - 2;
				}
			}
			break;

		case WT_TOMB_OP_JF_S: /* JumpFalse StackImmediate */
			{
				wt_bool cond;
				CHECK_POP(ctx, sizeof(wt_bool));
				WTR_CTX_DS_POP(ctx, wt_bool, cond);

				READ_STACKIMM;
				if(!cond)
				{
					ctx->frame.pc += opdata - 1;
				}
			}
			break;

		case WT_TOMB_OP_JA_I: /* Jump Absolute Immediate */
			{
				ctx->frame.pc = ctx->frame.prog->pc_base + opdata;
			}
			break;

		case WT_TOMB_OP_JA_L: /* Jump Absolute LongImmediate */
			{
				READ_LONGIMM;
				ctx->frame.pc = ctx->frame.prog->pc_base + opdata;
			}
			break;

		case WT_TOMB_OP_JA_S: /* Jump Absolute StackImmediate */
			{
				READ_STACKIMM;
				ctx->frame.pc = opdata;
			}
			break;

		case WT_TOMB_OP_EXEC_I: /* Execute microprogram Immediate */
			{
__exec_i:
				if(opdata<WTR_NUM_MICROPROGRAMS)
				{
					ctx->run = wtr_microprogtab[opdata](vm, ctx);
					goto handle_rr;
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_EXEC_L: /* Execute microprogram LongImmediate */
			{
				READ_LONGIMM;
				goto __exec_i;
			}
			break;

		case WT_TOMB_OP_EXEC_S: /* Execute microprogram StackImmediate */
			{
				READ_STACKIMM;
				goto __exec_i;
			}
			break;

		case WT_TOMB_OP_DLE_I: /* Dynamic Linking Execution Immediate */
			{
				goto __dle_l;
			}
			break;

		case WT_TOMB_OP_DLE_L: /* Dynamic Linking Execution LongImmediate */
			{
				struct wt_pfn pfn;
				READ_LONGIMM;
__dle_l:
				if(0!=wtr_vma_read(vm, ctx->frame.prog->base_address + opdata,
					&pfn, sizeof(pfn)))
				{
					goto memory_access_violation_exception;
				}

				ctx->run = wtr_processor_op_CFP(vm, ctx, &pfn);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_DLE_S: /* Dynamic Linking Execution StackImmediate */
			{
				READ_STACKIMM;
				goto __dle_l;
			}
			break;

		case WT_TOMB_OP_CFP: /* Call Function Pointer */
			{
				struct wt_pfn pfn;
				CHECK_POP(ctx, sizeof(struct wt_pfn));
				WTR_CTX_DS_POP(ctx, struct wt_pfn, pfn);

				ctx->run = wtr_processor_op_CFP(vm, ctx, &pfn);
				goto handle_rr;
			}
			break;

		case WT_TOMB_OP_MKFP_I: /* Make Function Pointer Immediate */
			{
				struct wt_pfn pfn;
__mkfp_i:
				/* check whether the assembled function pointer fits on the stack */
				CHECK_PUSH(ctx, sizeof(struct wt_pfn));

				pfn._pad = 0;
				pfn.ptr  = ctx->frame.prog;
				pfn.ext  = ctx->frame.prog->pc_base + opdata;

				WTR_CTX_DS_PUSH(ctx, struct wt_pfn, pfn);
			}
			break;

		case WT_TOMB_OP_MKFP_L: /* Make Function Pointer LongImmediate */
			{
				READ_LONGIMM;
				goto __mkfp_i;
			}
			break;

		case WT_TOMB_OP_MKFP_S: /* Make Function Pointer StackImmediate */
			{
				READ_STACKIMM;
				goto __mkfp_i;
			}
			break;

		case WT_TOMB_OP_TTD: /* TOS Top Delete */
			{
				if(ctx->tos.sp>ctx->frame.tos_mark)
				{
					struct wtr_objentry ent = *wtr_dynstack_top(&ctx->tos);
					wtr_dynstack_pop(&ctx->tos);

					if(ctx->tos.sp==ctx->frame.tos_mark && ctx->rtos.sp)
					{
						struct wtr_objentry* const top = wtr_dynstack_top(&ctx->rtos);
						wtr_dynstack_push_p(&ctx->tos, top);
						wtr_dynstack_pop(&ctx->rtos);
					}

					CHECK_PUSH(ctx, sizeof(wt_u32));
					WTR_CTX_DS_PUSH(ctx, wt_u32, ent.obj);

					ctx->run = wtr_processor_op_CFP(vm, ctx, &ent.pfn);
					goto handle_rr;
				}
				else
				{
					goto stack_underflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_LDFP: /* Load Function Pointer */
			{
				wt_vmptr addr;
				CHECK_POP(ctx, sizeof(addr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);

				if(WTR_CTX_DS_SPACE(ctx)>=sizeof(struct wt_pfn))
				{
					ctx->dsp -= sizeof(struct wt_pfn);

					if(0!=wtr_vma_memcpy(vm, ctx->dsp, addr,
						sizeof(struct wt_pfn)))
					{
						goto memory_access_violation_exception;
					}
				}
				else
				{
					goto stack_overflow_exception;
				}
			}
			break;

		case WT_TOMB_OP_STFP: /* Store Function Pointer */
			{
				wt_vmptr addr;
				CHECK_POP(ctx, sizeof(addr)+sizeof(struct wt_pfn));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);

				if(0!=wtr_vma_memcpy(vm, addr, ctx->dsp,
					sizeof(struct wt_pfn)))
				{
					goto memory_access_violation_exception;
				}

				ctx->dsp += sizeof(struct wt_pfn);
			}
			break;

		case WT_TOMB_OP_FPS_I: /* Function Pointer Size Immediate */
			{
__fps_i:
				opdata *= sizeof(struct wt_pfn);
				CHECK_PUSH(ctx, sizeof(wt_u32));
				WTR_CTX_DS_PUSH(ctx, wt_u32, opdata);
			}
			break;

		case WT_TOMB_OP_FPS_L: /* Function Pointer Size LongImmediate */
			{
				READ_LONGIMM;
				goto __fps_i;
			}
			break;

		case WT_TOMB_OP_STRN: /* New String */
			{
				wt_vmptr cstr;
				wt_vmptr str;
				wt_u32   len;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, cstr);

				if(0!=cstr)
				{
					if(0!=wtr_vma_strlen(vm, cstr, &len))
					{
						goto memory_access_violation_exception;
					}
				}
				else
				{
					len = 0;
				}

				if(0==(str=wtr_str_create_n(vm, NULL, len)))
				{
					goto__out_of_memory_exception();
				}

				if(0!=cstr)
				{
					wtr_vma_memcpy(vm, str+offsetof(struct wtr_str, data),
						cstr, len);
				}

				WTR_CTX_DS_PUSH(ctx, wt_vmptr, str);
			}
			break;

		case WT_TOMB_OP_STRC: /* Copy String */
			{
				wt_vmptr str;
				wt_vmptr cpy;
				wt_u32   len;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, str);
				ASSERT_OBJECT_EXISTENCE(str);

				if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length),
					&len, sizeof(len)))
				{
					goto memory_access_violation_exception;
				}

				if(0==(cpy=wtr_str_create_n(vm, NULL, len)))
				{
					goto__out_of_memory_exception();
				}

				if(0!=wtr_vma_memcpy(vm,
					cpy+offsetof(struct wtr_str, data),
					str+offsetof(struct wtr_str, data),
					len))
				{
					wtr_str_delete(vm, cpy);
					goto memory_access_violation_exception;
				}

				WTR_CTX_DS_PUSH(ctx, wt_vmptr, cpy);
			}
			break;

		case WT_TOMB_OP_STRL: /* Get String Length */
			{
				wt_vmptr str;
				wt_u32   len;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, str);
				ASSERT_OBJECT_EXISTENCE(str);

				if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length),
					&len, sizeof(len)))
				{
					goto memory_access_violation_exception;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, len);
			}
			break;

		case WT_TOMB_OP_STRD: /* Delete String */
			{
				wt_vmptr vm_addr;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, vm_addr);

				if(0!=vm_addr)
				{
					wtr_str_delete(vm, vm_addr);
				}
			}
			break;

		case WT_TOMB_OP_STRCR_I: /* String Character Reference Immediate */
			{
				goto __strcr_s;
			}
			break;

		case WT_TOMB_OP_STRCR_L: /* String Character Reference LongImmediate */
			{
				READ_LONGIMM;
				goto __strcr_s;
			}
			break;

		case WT_TOMB_OP_STRCR_S: /* String Character Reference StackImmediate */
			{
				wt_vmptr str;
				wt_u32   len;
				READ_STACKIMM;
__strcr_s:
				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, str);
				ASSERT_OBJECT_EXISTENCE(str);
				CHECK_OBJECT_MAGIC12_STR(str);

				if(0!=wtr_vma_read(vm, str+offsetof(struct wtr_str, length),
					&len, sizeof(len)))
				{
					goto memory_access_violation_exception;
				}

				if(opdata<len)
				{
					WTR_CTX_DS_PUSH(ctx, wt_vmptr,
						str+offsetof(struct wtr_str, data) + opdata);
				}
				else
				{
					goto index_out_of_range_exception;
				}
			}
			break;

		case WT_TOMB_OP_SEQN_I: /* New Sequence Skeleton Immediate */
			{
				wt_vmptr seq;
				wt_u32   cnt;
__seqn_i:
				CHECK_POP(ctx, sizeof(wt_u32));
				WTR_CTX_DS_POP(ctx, wt_u32, cnt);

				if(0<opdata)
				{
					if(0!=(seq=wtr_seq_create(vm, opdata, cnt)))
					{
						WTR_CTX_DS_PUSH(ctx, wt_vmptr, seq);
					}
					else
					{
						goto__out_of_memory_exception();
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_SEQN_L: /* New Sequence Skeleton LongImmediate */
			{
				READ_LONGIMM;
				goto __seqn_i;
			}
			break;

		case WT_TOMB_OP_SEQN_S: /* New Sequence Skeleton StackImmediate */
			{
				READ_STACKIMM;
				goto __seqn_i;
			}
			break;

		case WT_TOMB_OP_SEQD: /* Delete Sequence Skeleton */
			{
				wt_vmptr seq;
				struct wtr_program* const prog = ctx->frame.prog;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, seq);
				wtr_seq_delete(vm, seq);
			}
			break;

		case WT_TOMB_OP_SEQIR_I: /* Sequence Item Reference Immediate */
			{
				goto __seqir_s;
			}
			break;

		case WT_TOMB_OP_SEQIR_L: /* Sequence Item Reference LongImmediate */
			{
				READ_LONGIMM;
				goto __seqir_s;
			}
			break;

		case WT_TOMB_OP_SEQIR_S: /* Sequence Item Reference StackImmediate */
			{
				wt_vmptr seq;
				wt_u32   len;
				wt_u32   esz;
				READ_STACKIMM;
__seqir_s:
				CHECK_POP(ctx, sizeof(seq));
				WTR_CTX_DS_POP(ctx, wt_vmptr, seq);
				ASSERT_OBJECT_EXISTENCE(seq);
				CHECK_OBJECT_MAGIC12_SEQ(seq);

				if(0!=wtr_vma_read(vm, seq+offsetof(struct wtr_seq, length),
					&len, sizeof(len)))
				{
					goto memory_access_violation_exception;
				}

				if(0!=wtr_vma_read(vm, seq+offsetof(struct wtr_seq, elemsize),
					&esz, sizeof(esz)))
				{
					goto memory_access_violation_exception;
				}

				if(opdata<len)
				{
					WTR_CTX_DS_PUSH(ctx, wt_vmptr,
						seq+offsetof(struct wtr_seq, data) + esz*opdata);
				}
				else
				{
					goto index_out_of_range_exception;
				}
			}
			break;

		case WT_TOMB_OP_SEQL: /* Get Sequence Length */
			{
				wt_vmptr seq;
				wt_u32   len;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, seq);
				ASSERT_OBJECT_EXISTENCE(seq);

				if(0!=wtr_vma_read(vm, seq+offsetof(struct wtr_seq, length),
					&len, sizeof(len)))
				{
					goto memory_access_violation_exception;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, len);
			}
			break;

		case WT_TOMB_OP_RECN_I: /* New Record Skeleton Immediate */
			{
				wt_vmptr rec;
__recn_i:
				if(0!=(rec=wtr_rec_create(vm, opdata)))
				{
					WTR_CTX_DS_PUSH(ctx, wt_vmptr, rec);
				}
				else
				{
					goto__out_of_memory_exception();
				}
			}
			break;

		case WT_TOMB_OP_RECN_L: /* New Record Skeleton LongImmediate */
			{
				READ_LONGIMM;
				goto __recn_i;
			}
			break;

		case WT_TOMB_OP_RECN_S: /* New Record Skeleton StackImmediate */
			{
				READ_STACKIMM;
				goto __recn_i;
			}
			break;

		case WT_TOMB_OP_RECD: /* Delete Record Skeleton */
			{
				wt_vmptr vm_addr;
				struct wtr_program* const prog = ctx->frame.prog;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, vm_addr);
				wtr_rec_delete(vm, vm_addr);
			}
			break;

		case WT_TOMB_OP_RECFR_I: /* Record Field Reference Immediate */
			{
				wt_vmptr rec;
				wt_u32   size;
__recfr_i:
				CHECK_POP(ctx, sizeof(rec));
				WTR_CTX_DS_POP(ctx, wt_vmptr, rec);
				ASSERT_OBJECT_EXISTENCE(rec);
				CHECK_OBJECT_MAGIC12_REC(rec);

				if(0!=wtr_vma_read(vm, rec+offsetof(struct wtr_rec, size),
					&size, sizeof(size)))
				{
					goto memory_access_violation_exception;
				}

				if(opdata<size)
				{
					WTR_CTX_DS_PUSH(ctx, wt_vmptr,
						rec+offsetof(struct wtr_rec, data)+opdata);
				}
				else
				{
					goto index_out_of_range_exception;
				}
			}
			break;

		case WT_TOMB_OP_RECFR_L: /* Record Field Reference LongImmediate */
			{
				READ_LONGIMM;
				goto __recfr_i;
			}
			break;

		case WT_TOMB_OP_RECFR_S: /* Record Field Reference StackImmediate */
			{
				READ_STACKIMM;
				goto __recfr_i;
			}
			break;

		case WT_TOMB_OP_ERAISE: /* Raise Exception */
			{
				CHECK_POP(ctx, sizeof(ctx->last_exception));
				WTR_CTX_DS_PEEK(ctx, wt_exception, ctx->last_exception);
				goto unwind_frames_until_ehandler;
			}
			break;

		case WT_TOMB_OP_EXHAN_I: /* set Exception Handler Immediate */
			{
				ctx->frame.ehandler = ctx->frame.pc - 1 + opdata;
				ctx->frame.dsp_mark = ctx->dsp;
			}
			break;

		case WT_TOMB_OP_EXHAN_L: /* set Exception Handler LongImmediate */
			{
				READ_LONGIMM;
				ctx->frame.ehandler = ctx->frame.pc - 2 + opdata;
				ctx->frame.dsp_mark = ctx->dsp;
			}
			break;

		case WT_TOMB_OP_EXHAN_S: /* set Exception Handler StackImmediate */
			{
				READ_STACKIMM;
				ctx->frame.ehandler = ctx->frame.pc - 1 + opdata;
				ctx->frame.dsp_mark = ctx->dsp;
			}
			break;

		case WT_TOMB_OP_EHCL: /* Clear Exception Handler */
			{
				ctx->frame.ehandler = WTR_NULLPC;
				ctx->frame.dsp_mark = 0;
			}
			break;

		case WT_TOMB_OP_ADD_S:
		case WT_TOMB_OP_SUB_S:
		case WT_TOMB_OP_MUL_S:
		case WT_TOMB_OP_DIV_S:
		case WT_TOMB_OP_SHL_S:
		case WT_TOMB_OP_SHR_S:
		case WT_TOMB_OP_SAR_S:
		case WT_TOMB_OP_ROR_S:
		case WT_TOMB_OP_SMUL_S:
		case WT_TOMB_OP_SDIV_S:
		case WT_TOMB_OP_ADDB_S:
		case WT_TOMB_OP_SUBB_S:
		case WT_TOMB_OP_ADDH_S:
		case WT_TOMB_OP_SUBH_S:
		case WT_TOMB_OP_SHLB_S:
		case WT_TOMB_OP_SHLH_S:
			READ_STACKIMM;
			/* FALLTHRU */
		case WT_TOMB_OP_ADD_I:
		case WT_TOMB_OP_SUB_I:
		case WT_TOMB_OP_MUL_I:
		case WT_TOMB_OP_DIV_I:
		case WT_TOMB_OP_SHL_I:
		case WT_TOMB_OP_SHR_I:
		case WT_TOMB_OP_SAR_I:
		case WT_TOMB_OP_ROR_I:
		case WT_TOMB_OP_SMUL_I:
		case WT_TOMB_OP_SDIV_I:
		case WT_TOMB_OP_ADDB_I:
		case WT_TOMB_OP_SUBB_I:
		case WT_TOMB_OP_ADDH_I:
		case WT_TOMB_OP_SUBH_I:
		case WT_TOMB_OP_SHLB_I:
		case WT_TOMB_OP_SHLH_I:
			{
				union wtr_genvalue32 tmp;
				CHECK_POP(ctx, sizeof(tmp));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, tmp);

				switch(opcode)
				{
				case WT_TOMB_OP_ADD_I:
				case WT_TOMB_OP_ADD_S:
					tmp.v_u32 += opdata;
					break;

				case WT_TOMB_OP_SUB_I:
				case WT_TOMB_OP_SUB_S:
					tmp.v_u32 -= opdata;
					break;

				case WT_TOMB_OP_MUL_I:
				case WT_TOMB_OP_MUL_S:
					tmp.v_u32 *= opdata;
					break;

				case WT_TOMB_OP_DIV_I:
				case WT_TOMB_OP_DIV_S:
					if(0==opdata) { goto division_by_zero_exception; }
					tmp.v_u32 /= opdata;
					break;

				case WT_TOMB_OP_SHL_I:
				case WT_TOMB_OP_SHL_S:
					if(32>opdata)
					{
						tmp.v_u32 <<= opdata;
					}
					else
					{
						tmp.v_u32 = 0;
					}
					break;

				case WT_TOMB_OP_SHR_I:
				case WT_TOMB_OP_SHR_S:
					if(32>opdata)
					{
						tmp.v_u32 >>= opdata;
					}
					else
					{
						tmp.v_u32 = 0;
					}
					break;

				case WT_TOMB_OP_SAR_I:
				case WT_TOMB_OP_SAR_S:
					if(32>opdata)
					{
						tmp.v_i32 >>= opdata;
					}
					else
					if(0<tmp.v_i32)
					{
						tmp.v_i32 = 0;
					}
					else
					{
						tmp.v_i32 = -1;
					}
					break;

				case WT_TOMB_OP_ROR_I:
				case WT_TOMB_OP_ROR_S:
					opdata &= 0x1f;
					tmp.v_u32 = __OP_FUNC_ROR(tmp.v_u32, opdata);
					break;

				case WT_TOMB_OP_SMUL_I:
				case WT_TOMB_OP_SMUL_S:
					tmp.v_i32 *= (wt_i32)opdata;
					break;

				case WT_TOMB_OP_SDIV_I:
				case WT_TOMB_OP_SDIV_S:
					if(0==opdata) { goto division_by_zero_exception; }
					tmp.v_i32 /= (wt_i32)opdata;
					break;

				case WT_TOMB_OP_ADDB_I:
				case WT_TOMB_OP_ADDB_S:
					tmp.v_u32 = __OP_FUNC_ADDB(tmp.v_u32, opdata);
					break;

				case WT_TOMB_OP_SUBB_I:
				case WT_TOMB_OP_SUBB_S:
					tmp.v_u32 = __OP_FUNC_SUBB(tmp.v_u32, opdata);
					break;

				case WT_TOMB_OP_ADDH_I:
				case WT_TOMB_OP_ADDH_S:
					tmp.v_u32 = __OP_FUNC_ADDH(tmp.v_u32, opdata);
					break;

				case WT_TOMB_OP_SUBH_I:
				case WT_TOMB_OP_SUBH_S:
					tmp.v_u32 = __OP_FUNC_SUBH(tmp.v_u32, opdata);
					break;

				case WT_TOMB_OP_SHLB_I:
				case WT_TOMB_OP_SHLB_S:
					if(8>opdata)
					{
						tmp.v_u32 = __OP_FUNC_SHLB(tmp.v_u32, opdata);
					}
					else
					{
						tmp.v_u32 = 0;
					}
					break;

				case WT_TOMB_OP_SHLH_I:
				case WT_TOMB_OP_SHLH_S:
					if(16>opdata)
					{
						tmp.v_u32 = __OP_FUNC_SHLH(tmp.v_u32, opdata);
					}
					else
					{
						tmp.v_u32 = 0;
					}
					break;

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, tmp.v_u32);
			}
			break;

		case WT_TOMB_OP_LSOR_L:
		case WT_TOMB_OP_RSB_L:
			READ_LONGIMM;
			/* FALLTHRU */
		case WT_TOMB_OP_LSOR_I:
		case WT_TOMB_OP_RSB_I:
			{
				union wtr_genvalue32 tmp;
				CHECK_POP(ctx, sizeof(tmp));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, tmp);

				switch(opcode)
				{
				case WT_TOMB_OP_LSOR_I:
				case WT_TOMB_OP_LSOR_L:
					tmp.v_u32 = (tmp.v_u32 << 16) | opdata;
					break;

				/* reverse subtraction */
				case WT_TOMB_OP_RSB_I:
				case WT_TOMB_OP_RSB_L:
					tmp.v_u32 = opdata - tmp.v_u32;
					break;

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, tmp.v_u32);
			}
			break;

		case WT_TOMB_OP_LSORL_L:
			READ_LONGIMM;
			/* FALLTHRU */
		case WT_TOMB_OP_LSORL_I:
			{
				union wtr_genvalue64 tmp;
				CHECK_POP(ctx, sizeof(tmp));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue64, tmp);

				switch(opcode)
				{
				case WT_TOMB_OP_LSORL_I:
				case WT_TOMB_OP_LSORL_L:
					tmp.v_u64 = (tmp.v_u64 << 24) | opdata;
					break;

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u64, tmp.v_u64);
			}
			break;

		case WT_TOMB_OP_XOR:
		case WT_TOMB_OP_BAND:
		case WT_TOMB_OP_BOR:
		case WT_TOMB_OP_MOD:
		case WT_TOMB_OP_SMOD:
		case WT_TOMB_OP_LAND:
		case WT_TOMB_OP_LOR:
		case WT_TOMB_OP_LXOR:
#if WTRUNTIME_ENABLE_FLOAT32
		case WT_TOMB_OP_FADD:
		case WT_TOMB_OP_FSUB:
		case WT_TOMB_OP_FMUL:
		case WT_TOMB_OP_FDIV:
		case WT_TOMB_OP_FMOD:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
		case WT_TOMB_OP_MULB:
		case WT_TOMB_OP_SMULB:
		case WT_TOMB_OP_DIVB:
		case WT_TOMB_OP_SDIVB:
		case WT_TOMB_OP_MODB:
		case WT_TOMB_OP_SMODB:
		case WT_TOMB_OP_RORB:
		case WT_TOMB_OP_MULH:
		case WT_TOMB_OP_SMULH:
		case WT_TOMB_OP_DIVH:
		case WT_TOMB_OP_SDIVH:
		case WT_TOMB_OP_MODH:
		case WT_TOMB_OP_SMODH:
		case WT_TOMB_OP_RORH:
			{
				union wtr_genvalue32 rhs;
				union wtr_genvalue32 lhs;
				CHECK_POP(ctx, sizeof(lhs)+sizeof(rhs));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, rhs);
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, lhs);

				switch(opcode)
				{
				case WT_TOMB_OP_XOR:
					lhs.v_u32 ^= rhs.v_u32;
					break;

				case WT_TOMB_OP_BAND:
					lhs.v_u32 &= rhs.v_u32;
					break;

				case WT_TOMB_OP_BOR:
					lhs.v_u32 |= rhs.v_u32;
					break;

				case WT_TOMB_OP_MOD:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 %= rhs.v_u32;
					break;

				case WT_TOMB_OP_SMOD:
					if(0==rhs.v_i32) { goto division_by_zero_exception; }
					lhs.v_i32 %= rhs.v_i32;
					break;

				case WT_TOMB_OP_LAND:
					lhs.v_bool = lhs.v_bool && rhs.v_bool;
					break;

				case WT_TOMB_OP_LOR:
					lhs.v_bool = lhs.v_bool || rhs.v_bool;
					break;

				case WT_TOMB_OP_LXOR:
					lhs.v_bool = lhs.v_bool ^ rhs.v_bool;
					break;

#if WTRUNTIME_ENABLE_FLOAT32
				case WT_TOMB_OP_FADD:
					lhs.v_f32 += rhs.v_f32;
					break;

				case WT_TOMB_OP_FSUB:
					lhs.v_f32 -= rhs.v_f32;
					break;

				case WT_TOMB_OP_FMUL:
					lhs.v_f32 *= rhs.v_f32;
					break;

				case WT_TOMB_OP_FDIV:
					if(0==rhs.v_f32) { goto division_by_zero_exception; }
					lhs.v_f32 /= rhs.v_f32;
					break;

				case WT_TOMB_OP_FMOD:
					if(0==rhs.v_f32) { goto division_by_zero_exception; }
					lhs.v_f32 = __OP_FUNC_FMOD(lhs.v_f32, rhs.v_f32);
					break;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */

				/* note: B and H instructions fetch u32, and pass the raw */
				/* u32 to the actual opcode function */

				case WT_TOMB_OP_MULB:
					lhs.v_u32 = __OP_FUNC_MULB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SMULB:
					lhs.v_u32 = __OP_FUNC_SMULB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_DIVB:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_DIVB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SDIVB:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_SDIVB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_MODB:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_MODB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SMODB:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_SMODB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_RORB:
					rhs.v_u32 &= 7;
					lhs.v_u32 = __OP_FUNC_RORB(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_MULH:
					lhs.v_u32 = __OP_FUNC_MULH(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SMULH:
					lhs.v_u32 = __OP_FUNC_SMULH(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_DIVH:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_DIVH(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SDIVH:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_SDIVH(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_MODH:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_MODH(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SMODH:
					if(0==rhs.v_u32) { goto division_by_zero_exception; }
					lhs.v_u32 = __OP_FUNC_SMODH(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_RORH:
					rhs.v_u32 &= 0xf;
					lhs.v_u32 = __OP_FUNC_RORH(lhs.v_u32, rhs.v_u32);
					break;

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, lhs.v_u32);
			}
			break;

		case WT_TOMB_OP_ADDL:
		case WT_TOMB_OP_SUBL:
		case WT_TOMB_OP_MULL:
		case WT_TOMB_OP_DIVL:
		case WT_TOMB_OP_SHLL:
		case WT_TOMB_OP_SHRL:
		case WT_TOMB_OP_SARL:
		case WT_TOMB_OP_RORL:
		case WT_TOMB_OP_XORL:
		case WT_TOMB_OP_BANDL:
		case WT_TOMB_OP_BORL:
		case WT_TOMB_OP_MODL:
		case WT_TOMB_OP_SMODL:
		case WT_TOMB_OP_SMULL:
		case WT_TOMB_OP_SDIVL:
#if WTRUNTIME_ENABLE_FLOAT64
		case WT_TOMB_OP_FADDL:
		case WT_TOMB_OP_FSUBL:
		case WT_TOMB_OP_FMULL:
		case WT_TOMB_OP_FDIVL:
		case WT_TOMB_OP_FMODL:
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
			{
				union wtr_genvalue64 rhs;
				union wtr_genvalue64 lhs;
				CHECK_POP(ctx, sizeof(lhs)+sizeof(rhs));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue64, rhs);
				WTR_CTX_DS_POP(ctx, union wtr_genvalue64, lhs);

				switch(opcode)
				{
				case WT_TOMB_OP_ADDL:
					lhs.v_u64 += rhs.v_u64;
					break;

				case WT_TOMB_OP_SUBL:
					lhs.v_u64 -= rhs.v_u64;
					break;

				case WT_TOMB_OP_MULL:
					lhs.v_u64 *= rhs.v_u64;
					break;

				case WT_TOMB_OP_DIVL:
					if(0==rhs.v_u64) { goto division_by_zero_exception; }
					lhs.v_u64 /= rhs.v_u64;
					break;

				case WT_TOMB_OP_SHLL:
					if(64>rhs.v_u64)
					{
						lhs.v_u64 <<= rhs.v_u64;
					}
					else
					{
						lhs.v_u64 = 0;
					}
					break;

				case WT_TOMB_OP_SHRL:
					if(64>rhs.v_u64)
					{
						lhs.v_u64 >>= rhs.v_u64;
					}
					else
					{
						lhs.v_u64 = 0;
					}
					break;

				case WT_TOMB_OP_SARL:
					if(64>rhs.v_i64)
					{
						lhs.v_i64 >>= rhs.v_i64;
					}
					else
					if(0<lhs.v_i64)
					{
						lhs.v_i64 = 0;
					}
					else
					{
						lhs.v_i64 = -1;
					}
					break;

				case WT_TOMB_OP_RORL:
					opdata &= 0x3f;
					lhs.v_u64 = __OP_FUNC_RORL(lhs.v_u64, rhs.v_u64);
					break;

				case WT_TOMB_OP_XORL:
					lhs.v_u64 ^= rhs.v_u64;
					break;

				case WT_TOMB_OP_BANDL:
					lhs.v_u64 &= rhs.v_u64;
					break;

				case WT_TOMB_OP_BORL:
					lhs.v_u64 |= rhs.v_u64;
					break;

				case WT_TOMB_OP_MODL:
					if(0==rhs.v_u64) { goto division_by_zero_exception; }
					lhs.v_u64 %= rhs.v_u64;
					break;

				case WT_TOMB_OP_SMODL:
					if(0==rhs.v_i64) { goto division_by_zero_exception; }
					lhs.v_i64 %= rhs.v_i64;
					break;

				case WT_TOMB_OP_SMULL:
					lhs.v_i64 *= rhs.v_i64;
					break;

				case WT_TOMB_OP_SDIVL:
					if(0==rhs.v_i64) { goto division_by_zero_exception; }
					lhs.v_i64 /= rhs.v_i64;
					break;

#if WTRUNTIME_ENABLE_FLOAT64
				case WT_TOMB_OP_FADDL:
					lhs.v_f64 += rhs.v_f64;
					break;

				case WT_TOMB_OP_FSUBL:
					lhs.v_f64 -= rhs.v_f64;
					break;

				case WT_TOMB_OP_FMULL:
					lhs.v_f64 *= rhs.v_f64;
					break;

				case WT_TOMB_OP_FDIVL:
					if(0==rhs.v_f64) { goto division_by_zero_exception; }
					lhs.v_f64 /= rhs.v_f64;
					break;

				case WT_TOMB_OP_FMODL:
					if(0==rhs.v_f64) { goto division_by_zero_exception; }
					lhs.v_f64 = fmod(lhs.v_f64, rhs.v_f64);
					break;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u64, lhs.v_u64);
			}
			break;

		case WT_TOMB_OP_BNOT:
		case WT_TOMB_OP_NEG:
#if WTRUNTIME_ENABLE_FLOAT32
		case WT_TOMB_OP_FNEG:
		case WT_TOMB_OP_FABS:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
		case WT_TOMB_OP_LNOT:
		case WT_TOMB_OP_BNOTB:
		case WT_TOMB_OP_BNOTH:
			{
				union wtr_genvalue32 tmp;
				CHECK_POP(ctx, sizeof(tmp));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, tmp);

				switch(opcode)
				{
				case WT_TOMB_OP_BNOT:
					tmp.v_u32 = ~tmp.v_u32;
					break;

				case WT_TOMB_OP_NEG:
					tmp.v_i32 = -tmp.v_i32;
					break;

#if WTRUNTIME_ENABLE_FLOAT32
				case WT_TOMB_OP_FNEG:
					tmp.v_f32 = -tmp.v_f32;
					break;

				case WT_TOMB_OP_FABS:
					if(0.f>tmp.v_f32) { tmp.v_f32 = -tmp.v_f32; }
					break;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */

				case WT_TOMB_OP_LNOT:
					tmp.v_bool = !tmp.v_bool;
					break;

				case WT_TOMB_OP_BNOTB:
					tmp.v_u32 = ~((wt_u8)tmp.v_u32);
					break;

				case WT_TOMB_OP_BNOTH:
					tmp.v_u32 = ~((wt_u16)tmp.v_u32);
					break;

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, tmp.v_u32);
			}
			break;

		case WT_TOMB_OP_BNOTL:
		case WT_TOMB_OP_NEGL:
#if WTRUNTIME_ENABLE_FLOAT64
		case WT_TOMB_OP_FNEGL:
		case WT_TOMB_OP_FABSL:
#endif /* WTRUNTIME_ENABLE_FLOAT64 */
			{
				union wtr_genvalue64 tmp;
				CHECK_POP(ctx, sizeof(tmp));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue64, tmp);

				switch(opcode)
				{
				case WT_TOMB_OP_BNOTL:
					tmp.v_u64 = ~tmp.v_u64;
					break;

				case WT_TOMB_OP_NEGL:
					tmp.v_i64 = -tmp.v_i64;
					break;

#if WTRUNTIME_ENABLE_FLOAT64
				case WT_TOMB_OP_FNEGL:
					tmp.v_f64 = -tmp.v_f64;
					break;

				case WT_TOMB_OP_FABSL:
					if(0.f>tmp.v_f64) { tmp.v_f64 = -tmp.v_f64; }
					break;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u64, tmp.v_u64);
			}
			break;

		case WT_TOMB_OP_CMP:
		case WT_TOMB_OP_SCMP:
		case WT_TOMB_OP_SCMPB:
		case WT_TOMB_OP_SCMPH:
#if WTRUNTIME_ENABLE_FLOAT32
		case WT_TOMB_OP_FCMP:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
			{
				wt_i32 res;
				union wtr_genvalue32 rhs;
				union wtr_genvalue32 lhs;
				CHECK_POP(ctx, sizeof(lhs)+sizeof(rhs));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, rhs);
				WTR_CTX_DS_POP(ctx, union wtr_genvalue32, lhs);

				switch(opcode)
				{
				case WT_TOMB_OP_CMP:
					res = __OP_FUNC_CMP(lhs.v_u32, rhs.v_u32);
					break;

				case WT_TOMB_OP_SCMP:
					res = __OP_FUNC_CMP(lhs.v_i32, rhs.v_i32);
					break;

				case WT_TOMB_OP_SCMPB:
					{
						wt_i8 l = (wt_i8)(wt_u8)lhs.v_u32;
						wt_i8 r = (wt_i8)(wt_u8)rhs.v_u32;
						res = __OP_FUNC_CMP(l, r);
					}
					break;

				case WT_TOMB_OP_SCMPH:
					{
						wt_i16 l = (wt_i16)(wt_u16)lhs.v_u32;
						wt_i16 r = (wt_i16)(wt_u16)rhs.v_u32;
						res = __OP_FUNC_CMP(l, r);
					}
					break;

#if WTRUNTIME_ENABLE_FLOAT32
				case WT_TOMB_OP_FCMP:
					res = __OP_FUNC_CMP(lhs.v_f32, rhs.v_f32);
					break;
#endif /* WTRUNTIME_ENABLE_FLOAT32 */

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_i32, res);
			}
			break;

		case WT_TOMB_OP_CMPL:
		case WT_TOMB_OP_SCMPL:
#if WTRUNTIME_ENABLE_FLOAT32
		case WT_TOMB_OP_FCMPL:
#endif /* WTRUNTIME_ENABLE_FLOAT32 */
			{
				wt_i32 res;
				union wtr_genvalue64 rhs;
				union wtr_genvalue64 lhs;
				CHECK_POP(ctx, sizeof(lhs)+sizeof(rhs));
				WTR_CTX_DS_POP(ctx, union wtr_genvalue64, rhs);
				WTR_CTX_DS_POP(ctx, union wtr_genvalue64, lhs);

				switch(opcode)
				{
				case WT_TOMB_OP_CMPL:
					res = __OP_FUNC_CMP(lhs.v_u64, rhs.v_u64);
					break;

				case WT_TOMB_OP_SCMPL:
					res = __OP_FUNC_CMP(lhs.v_i64, rhs.v_i64);
					break;

#if WTRUNTIME_ENABLE_FLOAT64
				case WT_TOMB_OP_FCMPL:
					res = __OP_FUNC_CMP(lhs.v_f64, rhs.v_f64);
					break;
#endif /* WTRUNTIME_ENABLE_FLOAT64 */

				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_i32, res);
			}
			break;

		case WT_TOMB_OP_QLT:
		case WT_TOMB_OP_QLE:
		case WT_TOMB_OP_QGT:
		case WT_TOMB_OP_QGE:
		case WT_TOMB_OP_QEQ:
		case WT_TOMB_OP_QNE:
			{
				wt_i32 tmp;
				wt_bool res;

				CHECK_POP(ctx, sizeof(tmp));
				WTR_CTX_DS_POP(ctx, wt_i32, tmp);

				switch(opcode)
				{
				case WT_TOMB_OP_QLT: res = (tmp<0); break;
				case WT_TOMB_OP_QLE: res = (tmp<=0); break;
				case WT_TOMB_OP_QGT: res = (tmp>0); break;
				case WT_TOMB_OP_QGE: res = (tmp>=0); break;
				case WT_TOMB_OP_QEQ: res = (tmp==0); break;
				case WT_TOMB_OP_QNE: res = (tmp!=0); break;
				default: break;
				}

				WTR_CTX_DS_PUSH(ctx, wt_bool, res);
			}
			break;

		case WT_TOMB_OP_OSDR:
			{
				if(WTRCF_CSOS & ctx->frame.flags)
				{
					if(0<ctx->frame.csos_sp)
					{
						ctx->frame.csos_sp -= sizeof(struct wtr_objentry);
					}
					else
					{
						goto stack_underflow_exception;
					}
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_MSKBTM:
			{
				CHECK_POP(ctx, sizeof(wt_u32));

				if(32>opdata)
				{
					wt_u32 tmp;
					WTR_CTX_DS_POP(ctx, wt_u32, tmp);
					tmp &= (((wt_u32)1) << opdata) - 1;
					WTR_CTX_DS_PUSH(ctx, wt_u32, tmp);
				}
				else
				if(32==opdata)
				{
					/* nop */
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		case WT_TOMB_OP_LDB:
			{
				wt_vmptr addr;
				wt_u8    b;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);

				if(0!=wtr_vma_read(vm, addr, &b, sizeof(b)))
				{
					goto memory_access_violation_exception;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, b);
			}
			break;

		case WT_TOMB_OP_STB:
			{
				wt_vmptr addr;
				wt_u32   word;
				wt_u8    b;

				CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);
				WTR_CTX_DS_POP(ctx, wt_u32, word);

				b = (wt_u8)word;

				if(0!=wtr_vma_write(vm, addr, &b, sizeof(b)))
				{
					goto memory_access_violation_exception;
				}
			}
			break;

		case WT_TOMB_OP_LDH:
			{
				wt_vmptr addr;
				wt_u16   half;

				CHECK_POP(ctx, sizeof(wt_vmptr));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);

				if(0!=wtr_vma_read(vm, addr, &half, sizeof(half)))
				{
					goto memory_access_violation_exception;
				}

				WTR_CTX_DS_PUSH(ctx, wt_u32, half);
			}
			break;

		case WT_TOMB_OP_STH:
			{
				wt_vmptr addr;
				wt_u32   word;
				wt_u16   half;

				CHECK_POP(ctx, sizeof(wt_vmptr)+sizeof(wt_u32));
				WTR_CTX_DS_POP(ctx, wt_vmptr, addr);
				WTR_CTX_DS_POP(ctx, wt_u32, word);

				half = (wt_u16)word;

				if(0!=wtr_vma_write(vm, addr, &half, sizeof(half)))
				{
					goto memory_access_violation_exception;
				}
			}
			break;

		case WT_TOMB_OP_SGNXD:
			{
				CHECK_POP(ctx, sizeof(wt_u32));

				if(32>opdata && 0!=opdata)
				{
					const wt_u32 signbit = ((wt_u32)1) << (opdata - 1);

					wt_u32 tmp;

					WTR_CTX_DS_POP(ctx, wt_u32, tmp);

					if(signbit & tmp)
					{
						tmp |= ~(signbit - 1);
					}

					WTR_CTX_DS_PUSH(ctx, wt_u32, tmp);
				}
				else
				if(32==opdata)
				{
					/* nop */
				}
				else
				{
					goto bad_program_access_exception;
				}
			}
			break;

		default:
			{
				/* undefined instruction */

				exc = UndefinedInstructionException;
				goto raise_exception;

handle_rr:
				switch(ctx->run)
				{
				case WTRR_RETURNED:
				case WTRR_EXCEPTION:
					return ctx->run;

				case WTRR_RAISE:
					exc = ctx->last_exception;
					goto raise_exception;

				default:
					break;
				}
				break;

io_exception:
				exc = IOException;
				goto raise_exception;

memory_access_violation_exception:
				exc = MemoryAccessViolationException;
				goto raise_exception;

bad_program_access_exception:
				exc = BadProgramAccessException;
				goto raise_exception;

stack_overflow_exception:
stack_underflow_exception:
				exc = StackOverflowException;
				goto raise_exception;

division_by_zero_exception:
				exc = DivisionByZeroException;
				goto raise_exception;

index_out_of_range_exception:
				exc = IndexOutOfRangeException;
				goto raise_exception;

out_of_memory_exception:
				exc = OutOfMemoryException;
				goto raise_exception;

raise_exception:
				WTR_CTX_DS_PUSH(ctx, wt_exception, exc); /* if the stack is full, there is still a spare space for the exception */
unwind_frames_until_ehandler:
				ctx->run = wtr_processor_unwind_frames(vm, ctx);
				goto handle_rr;

#ifdef WTRUNTIME_DEBUG
				opcode = opcode;
#endif /* WTRUNTIME_DEBUG */
			}
			break;
		}
	}

	return ctx->run;
}
